Package: httr
Title: Tools for Working with URLs and HTTP
Version: 1.4.8
Authors@R: c(
    person("Hadley", "Wickham", , "hadley@posit.co", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: Useful tools for working with HTTP organised by HTTP verbs
    (GET(), POST(), etc). Configuration functions make it easy to control
    additional request components (authenticate(), add_headers() and so
    on).
License: MIT + file LICENSE
URL: https://httr.r-lib.org/, https://github.com/r-lib/httr
BugReports: https://github.com/r-lib/httr/issues
Depends: R (>= 3.6)
Imports: curl (>= 5.1.0), jsonlite, mime, openssl (>= 0.8), R6
Suggests: covr, httpuv, jpeg, knitr, png, readr, rmarkdown, testthat
        (>= 0.8.0), xml2
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-02-13 13:25:45 UTC; hadleywickham
Author: Hadley Wickham [aut, cre],
  Posit Software, PBC [cph, fnd]
Maintainer: Hadley Wickham <hadley@posit.co>
Repository: CRAN
Date/Publication: 2026-02-13 16:20:15 UTC
Built: R 4.4.0; ; 2026-07-17 19:04:07 UTC; unix

Package: httr
Title: Tools for Working with URLs and HTTP
Version: 1.4.8
Authors@R: c(
    person("Hadley", "Wickham", , "hadley@posit.co", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: Useful tools for working with HTTP organised by HTTP verbs
    (GET(), POST(), etc). Configuration functions make it easy to control
    additional request components (authenticate(), add_headers() and so
    on).
License: MIT + file LICENSE
URL: https://httr.r-lib.org/, https://github.com/r-lib/httr
BugReports: https://github.com/r-lib/httr/issues
Depends: R (>= 3.6)
Imports: curl (>= 5.1.0), jsonlite, mime, openssl (>= 0.8), R6
Suggests: covr, httpuv, jpeg, knitr, png, readr, rmarkdown, testthat
        (>= 0.8.0), xml2
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-02-13 13:25:45 UTC; hadleywickham
Author: Hadley Wickham [aut, cre],
  Posit Software, PBC [cph, fnd]
Maintainer: Hadley Wickham <hadley@posit.co>
Repository: CRAN
Date/Publication: 2026-02-13 16:20:15 UTC
Built: R 4.6.0; ; 2026-04-26 05:16:18 UTC; unix

Package: httr
Title: Tools for Working with URLs and HTTP
Version: 1.4.7
Authors@R: c(
    person("Hadley", "Wickham", , "hadley@posit.co", role = c("aut", "cre")),
    person("Posit, PBC", role = c("cph", "fnd"))
  )
Description: Useful tools for working with HTTP organised by HTTP verbs
    (GET(), POST(), etc). Configuration functions make it easy to control
    additional request components (authenticate(), add_headers() and so
    on).
License: MIT + file LICENSE
URL: https://httr.r-lib.org/, https://github.com/r-lib/httr
BugReports: https://github.com/r-lib/httr/issues
Depends: R (>= 3.5)
Imports: curl (>= 5.0.2), jsonlite, mime, openssl (>= 0.8), R6
Suggests: covr, httpuv, jpeg, knitr, png, readr, rmarkdown, testthat
        (>= 0.8.0), xml2
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Encoding: UTF-8
RoxygenNote: 7.2.3
NeedsCompilation: no
Packaged: 2023-08-15 02:56:56 UTC; hadleywickham
Author: Hadley Wickham [aut, cre],
  Posit, PBC [cph, fnd]
Maintainer: Hadley Wickham <hadley@posit.co>
Repository: CRAN
Date/Publication: 2023-08-15 09:00:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("$",insensitive)
S3method("[",insensitive)
S3method("[",url)
S3method("[[",insensitive)
S3method(as.character,form_file)
S3method(as.character,response)
S3method(as.request,"NULL")
S3method(as.request,Token)
S3method(as.request,list)
S3method(as.request,request)
S3method(as_field,default)
S3method(as_field,logical)
S3method(as_field,numeric)
S3method(c,request)
S3method(cookies,handle)
S3method(cookies,response)
S3method(headers,response)
S3method(http_error,character)
S3method(http_error,integer)
S3method(http_error,response)
S3method(print,cache_info)
S3method(print,handle)
S3method(print,oauth_app)
S3method(print,oauth_endpoint)
S3method(print,opts_list)
S3method(print,request)
S3method(print,response)
S3method(print,url)
S3method(request_fetch,write_disk)
S3method(request_fetch,write_memory)
S3method(request_fetch,write_stream)
S3method(status_code,numeric)
S3method(status_code,response)
export(BROWSE)
export(DELETE)
export(GET)
export(HEAD)
export(PATCH)
export(POST)
export(PUT)
export(RETRY)
export(Token)
export(Token1.0)
export(Token2.0)
export(TokenServiceAccount)
export(VERB)
export(accept)
export(accept_json)
export(accept_xml)
export(add_headers)
export(authenticate)
export(build_url)
export(cache_info)
export(config)
export(content)
export(content_type)
export(content_type_json)
export(content_type_xml)
export(cookies)
export(curl_docs)
export(get_callback)
export(guess_media)
export(handle)
export(handle_find)
export(handle_reset)
export(has_content)
export(headers)
export(hmac_sha1)
export(http_condition)
export(http_date)
export(http_error)
export(http_status)
export(http_type)
export(httr_dr)
export(httr_options)
export(init_oauth1.0)
export(init_oauth2.0)
export(insensitive)
export(message_for_status)
export(modify_url)
export(oauth1.0_token)
export(oauth2.0_access_token)
export(oauth2.0_authorize_url)
export(oauth2.0_token)
export(oauth_app)
export(oauth_callback)
export(oauth_endpoint)
export(oauth_endpoints)
export(oauth_exchanger)
export(oauth_header)
export(oauth_listener)
export(oauth_service_token)
export(oauth_signature)
export(parse_http_date)
export(parse_media)
export(parse_url)
export(parsed_content)
export(progress)
export(rerequest)
export(reset_config)
export(revoke_all)
export(set_callback)
export(set_config)
export(set_cookies)
export(sha1_hash)
export(sign_oauth1.0)
export(sign_oauth2.0)
export(status_code)
export(stop_for_status)
export(text_content)
export(timeout)
export(upload_file)
export(url_ok)
export(url_success)
export(use_proxy)
export(user_agent)
export(verbose)
export(warn_for_status)
export(with_config)
export(with_verbose)
export(write_disk)
export(write_function)
export(write_memory)
export(write_stream)
importFrom(R6,R6Class)
importFrom(utils,modifyList)

# httr

<!-- badges: start -->
[![CRAN status](https://www.r-pkg.org/badges/version/httr)](https://cran.r-project.org/package=httr)
[![R build status](https://github.com/r-lib/httr/workflows/R-CMD-check/badge.svg)](https://github.com/r-lib/httr/actions)
[![Codecov test coverage](https://codecov.io/gh/r-lib/httr/branch/master/graph/badge.svg)](https://codecov.io/gh/r-lib/httr?branch=master)
<!-- badges: end -->

The aim of httr is to provide a wrapper for the [curl](https://cran.r-project.org/package=curl) package, customised to the demands of modern web APIs.

Key features:

* Functions for the most important http verbs: `GET()`, `HEAD()`, `PATCH()`, 
  `PUT()`, `DELETE()` and `POST()`.

* Automatic connection sharing across requests to the same website (by
  default, curl handles are managed automatically), cookies are maintained
  across requests, and a up-to-date root-level SSL certificate store is used.

* Requests return a standard reponse object that captures the http status line,
  headers and body, along with other useful information.

  * Response content is available with `content()` as a raw vector (`as =
    "raw"`), a character vector (`as = "text"`), or parsed into an R object
    (`as = "parsed"`), currently for html, xml, json, png and jpeg.

  * You can convert http errors into R errors with `stop_for_status()`.

* Config functions make it easier to modify the request in common ways:
  `set_cookies()`, `add_headers()`, `authenticate()`, `use_proxy()`, 
  `verbose()`, `timeout()`, `content_type()`, `accept()`, `progress()`.

* Support for OAuth 1.0 and 2.0 with `oauth1.0_token()` and `oauth2.0_token()`.
  The demo directory has eight OAuth demos: four for 1.0 (twitter, vimeo,
  withings and yahoo) and four for 2.0 (facebook, github, google, linkedin). 
  OAuth credentials are automatically cached within a project. 

httr wouldn't be possible without the hard work of the authors of [curl](https://cran.r-project.org/package=curl) and [libcurl](http://curl.haxx.se/). Thanks! httr is inspired by http libraries in other languages, such as [Resty](http://beders.github.com/Resty/Resty/Examples.html), [Requests](https://requests.readthedocs.io/en/master/) and [httparty](http://github.com/jnunemaker/httparty/tree/master).

## Installation

To get the current released version from CRAN:

```R
install.packages("httr")
```

To get the current development version from github:

```R
# install.packages("devtools")
devtools::install_github("r-lib/httr")
```

## Code of Conduct

Please note that the httr project is released with a [Contributor Code of Conduct](https://httr.r-lib.org/CODE_OF_CONDUCT.html). By contributing to this project, you agree to abide by its terms.

# httr

## Status

[![Lifecycle: superseded](https://img.shields.io/badge/lifecycle-superseded-blue.svg)](https://lifecycle.r-lib.org/articles/stages.html#superseded)

httr is superseded: only changes necessary to keep it on CRAN will be made. We recommend using [httr2](https://httr2.r-lib.org) instead.

---

<!-- badges: start -->
[![CRAN status](https://www.r-pkg.org/badges/version/httr)](https://cran.r-project.org/package=httr)
[![R-CMD-check](https://github.com/r-lib/httr/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/httr/actions/workflows/R-CMD-check.yaml)
[![Codecov test coverage](https://codecov.io/gh/r-lib/httr/graph/badge.svg)](https://app.codecov.io/gh/r-lib/httr)
<!-- badges: end -->

The aim of httr is to provide a wrapper for the [curl](https://cran.r-project.org/package=curl) package, customised to the demands of modern web APIs.

Key features:

* Functions for the most important http verbs: `GET()`, `HEAD()`, `PATCH()`, 
  `PUT()`, `DELETE()` and `POST()`.

* Automatic connection sharing across requests to the same website (by
  default, curl handles are managed automatically), cookies are maintained
  across requests, and a up-to-date root-level SSL certificate store is used.

* Requests return a standard reponse object that captures the http status line,
  headers and body, along with other useful information.

  * Response content is available with `content()` as a raw vector (`as =
    "raw"`), a character vector (`as = "text"`), or parsed into an R object
    (`as = "parsed"`), currently for html, xml, json, png and jpeg.

  * You can convert http errors into R errors with `stop_for_status()`.

* Config functions make it easier to modify the request in common ways:
  `set_cookies()`, `add_headers()`, `authenticate()`, `use_proxy()`, 
  `verbose()`, `timeout()`, `content_type()`, `accept()`, `progress()`.

* Support for OAuth 1.0 and 2.0 with `oauth1.0_token()` and `oauth2.0_token()`.
  The demo directory has eight OAuth demos: four for 1.0 (twitter, vimeo,
  withings and yahoo) and four for 2.0 (facebook, github, google, linkedin). 
  OAuth credentials are automatically cached within a project. 

httr wouldn't be possible without the hard work of the authors of [curl](https://cran.r-project.org/package=curl) and [libcurl](https://curl.se/). Thanks! httr is inspired by http libraries in other languages, such as [Resty](http://beders.github.io/Resty/Resty/Examples.html), [Requests](https://requests.readthedocs.io/en/latest/) and [httparty](https://github.com/jnunemaker/httparty/tree/master).

## Installation

To get the current released version from CRAN:

```R
install.packages("httr")
```

To get the current development version from github:

```R
# install.packages("pak")
pak::pak("r-lib/httr")
```

## Code of Conduct

Please note that the httr project is released with a [Contributor Code of Conduct](https://httr.r-lib.org/CODE_OF_CONDUCT.html). By contributing to this project, you agree to abide by its terms.

#' Use a proxy to connect to the internet.
#'
#' @param url,port location of proxy
#' @param username,password login details for proxy, if needed
#' @param auth type of HTTP authentication to use. Should be one of the
#'   following: basic, digest, digest_ie, gssnegotiate, ntlm, any.
#' @family config
#' @export
#' @examples
#' # See http://www.hidemyass.com/proxy-list for a list of public proxies
#' # to test with
#' # GET("http://had.co.nz", use_proxy("64.251.21.73", 8080), verbose())
use_proxy <- function(url, port = NULL, username = NULL, password = NULL,
                      auth = "basic") {
  if (!is.null(username) || !is.null(password)) {
    proxyuserpwd <- paste0(username, ":", password)
  } else {
    proxyuserpwd <- NULL
  }

  if (!is.null(port)) stopifnot(is.numeric(port))

  config(
    proxy = url,
    proxyuserpwd = proxyuserpwd,
    proxyport = port,
    proxyauth = auth_flags(auth)
  )
}

body_config <- function(body = NULL,
                        encode = c("form", "json", "multipart", "raw"),
                        type = NULL) {
  # No body
  if (identical(body, FALSE)) {
    return(config(post = TRUE, nobody = TRUE))
  }

  # Send single file lazily
  if (inherits(body, "form_file")) {
    con <- file(body$path, "rb")
    size <- file.info(body$path)$size

    return(c(
      config(
        post = TRUE,
        readfunction = function(nbytes, ...) {
          if (is.null(con)) {
            return(raw())
          }
          bin <- readBin(con, "raw", nbytes)
          if (length(bin) < nbytes) {
            close(con)
            con <<- NULL
          }
          bin
        },
        seekfunction = function(offset, ...) {
          if (is.null(con)) {
            con <<- file(body$path, "rb")
          }
          seek(con, where = offset)
        },
        postfieldsize_large = size
      ),
      content_type(body$type)
    ))
  }

  # For character/raw, send raw bytes
  if (is.character(body) || is.raw(body)) {
    return(body_raw(body, type = type))
  }

  # Post with empty body
  if (is.null(body)) {
    return(body_raw(raw()))
  }

  if (!is.list(body)) {
    stop("Unknown type of `body`: must be NULL, FALSE, character, raw or list",
      call. = FALSE
    )
  }

  body <- compact(body)

  # Deal with three ways to encode: form, multipart & json
  encode <- match.arg(encode)
  if (encode == "raw") {
    body_raw(body)
  } else if (encode == "form") {
    body_raw(compose_query(body), "application/x-www-form-urlencoded")
  } else if (encode == "json") {
    body_raw(
      jsonlite::toJSON(body, auto_unbox = TRUE, digits = 22),
      "application/json"
    )
  } else if (encode == "multipart") {
    if (!all(has_name(body))) {
      stop("All components of body must be named", call. = FALSE)
    }
    request(fields = lapply(body, as_field))
  }
}

as_field <- function(x) UseMethod("as_field")
#' @export
as_field.numeric <- function(x) as.character(x)
#' @export
as_field.logical <- function(x) as.character(x)
#' @export
as_field.default <- function(x) x # assume curl will handle

body_raw <- function(body, type = NULL) {
  if (is.character(body)) {
    body <- charToRaw(paste(body, collapse = "\n"))
  }
  stopifnot(is.raw(body))

  c(
    config(
      post = TRUE,
      postfieldsize = length(body),
      postfields = body
    ),
    # For raw bodies, override default POST content-type
    content_type(type %||% "")
  )
}

#' Retry a request until it succeeds.
#'
#' Safely retry a request until it succeeds, as defined by the `terminate_on`
#' parameter, which by default means a response for which [http_error()]
#' is `FALSE`. Will also retry on error conditions raised by the underlying curl code,
#' but if the last retry still raises one, `RETRY` will raise it again with
#' [stop()].
#' It is designed to be kind to the server: after each failure
#' randomly waits up to twice as long. (Technically it uses exponential
#' backoff with jitter, using the approach outlined in
#' <https://www.awsarchitectureblog.com/2015/03/backoff.html>.)
#' If the server returns status code 429 and specifies a `retry-after` value, that
#' value will be used instead, unless it's smaller than `pause_min`.
#'
#' @inheritParams VERB
#' @inherit GET params return
#' @inheritParams POST
#' @param times Maximum number of requests to attempt.
#' @param pause_base,pause_cap This method uses exponential back-off with
#'   full jitter - this means that each request will randomly wait between 0
#'   and `pause_base * 2 ^ attempt` seconds, up to a maximum of
#'   `pause_cap` seconds.
#' @param pause_min Minimum time to wait in the backoff; generally
#'   only necessary if you need pauses less than one second (which may
#'   not be kind to the server, use with caution!).
#' @param quiet If `FALSE`, will print a message displaying how long
#'   until the next request.
#' @param terminate_on Optional vector of numeric HTTP status codes that if found
#'   on the response will terminate the retry process. If `NULL`, will keep
#'   retrying while [http_error()] is `TRUE` for the response.
#' @param terminate_on_success If `TRUE`, the default, this will
#'   automatically terminate when the request is successful, regardless of the
#'   value of `terminate_on`.
#' @return The last response. Note that if the request doesn't succeed after
#'   `times` times this will be a failed request, i.e. you still need
#'   to use [stop_for_status()].
#' @export
#' @examples
#' # Succeeds straight away
#' RETRY("GET", "http://httpbin.org/status/200")
#' # Never succeeds
#' RETRY("GET", "http://httpbin.org/status/500")
#' \dontrun{
#' # Invalid hostname generates curl error condition and is retried but eventually
#' # raises an error condition.
#' RETRY("GET", "http://invalidhostname/")
#' }
RETRY <- function(verb, url = NULL, config = list(), ...,
                  body = NULL, encode = c("multipart", "form", "json", "raw"),
                  times = 3, pause_base = 1, pause_cap = 60, pause_min = 1,
                  handle = NULL, quiet = FALSE,
                  terminate_on = NULL,
                  terminate_on_success = TRUE) {
  stopifnot(is.numeric(times), length(times) == 1L)
  stopifnot(is.numeric(pause_base), length(pause_base) == 1L)
  stopifnot(is.numeric(pause_cap), length(pause_cap) == 1L)
  stopifnot(is.numeric(terminate_on) || is.null(terminate_on))
  stopifnot(is.logical(terminate_on_success), length(terminate_on_success) == 1)

  hu <- handle_url(handle, url, ...)
  req <- request_build(verb, hu$url, body_config(body, match.arg(encode)), config, ...)
  resp <- tryCatch(request_perform(req, hu$handle$handle), error = function(e) e)

  i <- 1
  while (!retry_should_terminate(i, times, resp, terminate_on, terminate_on_success)) {
    backoff_full_jitter(i, resp, pause_base, pause_cap, pause_min, quiet = quiet)

    i <- i + 1
    resp <- tryCatch(request_perform(req, hu$handle$handle), error = function(e) e)
  }

  if (inherits(resp, "error")) {
    stop(resp)
  }

  resp
}

retry_should_terminate <- function(i, times, resp, terminate_on, terminate_on_success) {
  if (i >= times) {
    TRUE
  } else if (inherits(resp, "error")) {
    FALSE
  } else if (terminate_on_success && !http_error(resp)) {
    TRUE
  } else if (!is.null(terminate_on)) {
    status_code(resp) %in% terminate_on
  } else {
    !http_error(resp)
  }
}

backoff_full_jitter <- function(i, resp, pause_base = 1, pause_cap = 60,
                                pause_min = 1, quiet = FALSE) {
  length <- max(pause_min, stats::runif(1, max = min(pause_cap, pause_base * (2^i))))
  if (!quiet) {
    if (inherits(resp, "error")) {
      error_description <- gsub("[\n\r]*$", "\n", as.character(resp))
      status <- "ERROR"
    } else {
      error_description <- ""
      status <- status_code(resp)
    }
    if (status == 429) {
      retry_after <- resp$headers[["retry-after"]]
      if (!is.null(retry_after)) {
        length <- max(pause_min, as.numeric(retry_after))
      }
    }
    message(error_description, "Request failed [", status, "]. Retrying in ", round(length, 1), " seconds...")
  }
  Sys.sleep(length)
}

context("test-http-head")

test_that("HEAD succeeds regardless of wrapper", {
  skip_httpbin()
  expect_false(http_error(HEAD("https://httpbin.org")))
  expect_false(http_error(VERB("HEAD", "https://httpbin.org")))
  expect_false(http_error(RETRY("HEAD", "https://httpbin.org")))
})

context("test-retry")

test_that("sucessful requests terminate when terminate_on_success is true", {
  should_terminate <- function(status_code) {
    resp <- response(status_code = status_code)
    retry_should_terminate(1, 10,
      resp = resp,
      terminate_on = 500L,
      terminate_on_success = TRUE
    )
  }

  expect_true(should_terminate(200L))
  expect_false(should_terminate(400L))
})

test_that("if request_perform() throws an error, RETRY passes it on", {
  skip_on_cran()
  expect_error(
    RETRY("POST", "http://98d90a2a254647889e2e4c236fb576cd.com", times = 1),
    regexp = "resolve host"
  )
})

context("test-http-head")

test_that("HEAD succeeds regardless of wrapper", {
  expect_false(http_error(HEAD("https://httpbin.org")))
  expect_false(http_error(VERB("HEAD", "https://httpbin.org")))
  expect_false(http_error(RETRY("HEAD", "https://httpbin.org")))
})

#' Parse a media type.
#'
#' Parsed according to RFC 2616, as at
#' <https://www.rfc-editor.org/rfc/rfc2616#section-3.7>.
#'
#' A simplified minimal EBNF is:
#'
#' * media-type     = `type "/" subtype *( ";" parameter )`
#' * type           = `token`
#' * subtype        = `token`
#' * parameter      = `attribute "=" value`
#' * attribute      = `token`
#' * value          = `token | quoted-string`
#' * token          = `1*<any CHAR except CTLs or ()<>@@,;:\"/[]?=\{\}`
#' @export
#' @param x String to parse
#' @keywords internal
#' @examples
#' parse_media("text/plain")
#' parse_media("text/plain; charset=utf-8")
#' parse_media("text/plain; charset=\"utf-8\"")
#' parse_media("text/plain; randomparam=\";=;=\"")
parse_media <- function(x) {
  stopifnot(is.character(x), length(x) == 1)

  # Use scan to deal with quoted strings. It loses the quotes, but it's
  # ok because the attribute can't be a quoted string so there's no ambiguity
  # about who the = belongs to.
  parse <- function(x, sep) {
    scan(text = x, what = character(), sep = sep, quiet = TRUE, quote = '"')
  }
  pieces <- str_trim(parse(tolower(x), ";"))

  types <- str_split_fixed(pieces[1], "/", 2)[1, ]
  type <- tolower(types[1])
  subtype <- tolower(types[2])

  if (length(pieces) > 1) {
    param_pieces <- lapply(pieces[-1], str_split_fixed, "=", 2)
    names <- vapply(param_pieces, "[", 1, FUN.VALUE = character(1))
    values <- vapply(param_pieces, "[", 2, FUN.VALUE = character(1))

    params <- stats::setNames(as.list(values), names)
  } else {
    params <- list()
  }

  list(
    complete = paste(type, "/", subtype, sep = ""),
    type = type,
    subtype = subtype,
    params = params
  )
}

context("parse_media")

test_that("can parse bare media type", {
  out <- parse_media("text/plain")

  expect_equal(out$type, "text")
  expect_equal(out$subtype, "plain")
  expect_equal(out$params, list())
})

test_that("can parse params", {
  out1 <- parse_media("text/plain; a=1")
  out2 <- parse_media("text/plain; a=1; b=2")

  expect_equal(out1$params, list(a = "1"))
  expect_equal(out2$params, list(a = "1", b = "2"))
})

# Refresh an OAuth 2.0 credential.
#
# Refreshes the given token, and returns a new credential with a
# valid access_token. Based on:
# https://developers.google.com/accounts/docs/OAuth2InstalledApp#refresh
refresh_oauth2.0 <- function(endpoint, app, credentials, user_params = NULL,
                             use_basic_auth = FALSE) {
  if (is.null(credentials$refresh_token)) {
    stop("Refresh token not available", call. = FALSE)
  }

  refresh_url <- endpoint$access
  req_params <- list(
    refresh_token = credentials$refresh_token,
    client_id = app$key,
    grant_type = "refresh_token"
  )

  if (!is.null(user_params)) {
    req_params <- utils::modifyList(user_params, req_params)
  }

  if (isTRUE(use_basic_auth)) {
    response <- POST(refresh_url,
      body = req_params, encode = "form",
      authenticate(app$key, app$secret, type = "basic")
    )
  } else {
    req_params$client_secret <- app$secret
    response <- POST(refresh_url, body = req_params, encode = "form")
  }

  err <- find_oauth2.0_error(response)
  if (!is.null(err)) {
    lines <- c(
      paste0("Unable to refresh token: ", err$error),
      err$error_description,
      err$error_uri
    )
    warning(paste(lines, collapse = "\n"), call. = FALSE)
    return(NULL)
  }

  stop_for_status(response)
  refresh_data <- content(response)
  utils::modifyList(credentials, refresh_data)
}

# !!! The special redirect URI "urn:ietf:wg:oauth:2.0:oob used
# !!! by httr in case httuv is not installed is currently not
# !!! supported by Azure Active Directory (AAD).
# !!! Therefore it is required to install httpuv to make this work.

# 1. Register an app app in AAD, e.g. as a "Native app", with
#    redirect URI <http://localhost:1410/>.
# 2. Insert the App name:
app_name <- "myapp" # not important for authorization grant flow
# 3. Insert the created apps client ID which was issued after app creation:
client_id <- "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee"
# In case your app was registered as a web app instead of a native app,
# you might have to add your secret key string here:
client_secret <- NULL
# API resource ID to request access for, e.g. Power BI:
resource_uri <- "https://analysis.windows.net/powerbi/api"

# Obtain OAuth2 endpoint settings for azure:
#    This uses the "common" endpoint.
#    To use a tenant url, create an
#    oauth_endpoint(authorize = "https://login.windows.net/<tenant_id>/oauth2/authorize",
#                   access = "https://login.windows.net/<tenant_id>/oauth2/token")
#    with <tenant_id> replaced by your endpoint ID.
azure_endpoint <- oauth_endpoints("azure")

# Create the app instance.
myapp <- oauth_app(
  appname = app_name,
  key = client_id,
  secret = client_secret
)

# Step through the authorization chain:
#    1. You will be redirected to you authorization endpoint via web browser.
#    2. Once you responded to the request, the endpoint will redirect you to
#       the local address specified by httr.
#    3. httr will acquire the authorization code (or error) from the data
#       posted to the redirect URI.
#    4. If a code was acquired, httr will contact your authorized token access
#       endpoint to obtain the token.
mytoken <- oauth2.0_token(azure_endpoint, myapp,
  user_params = list(resource = resource_uri),
  use_oob = FALSE
)
if (("error" %in% names(mytoken$credentials)) && (nchar(mytoken$credentials$error) > 0)) {
  errorMsg <- paste("Error while acquiring token.",
    paste("Error message:", mytoken$credentials$error),
    paste("Error description:", mytoken$credentials$error_description),
    paste("Error code:", mytoken$credentials$error_codes),
    sep = "\n"
  )
  stop(errorMsg)
}

# Resource API can be accessed through "mytoken" at this point.

#' Use http authentication.
#'
#' It's not obvious how to turn authentication off after using it, so
#' I recommend using custom handles with authentication.
#'
#' @param user user name
#' @param password password
#' @param type type of HTTP authentication.  Should be one of the following
#'   types supported by Curl: basic, digest, digest_ie, gssnegotiate,
#'   ntlm, any.  It defaults to "basic", the most common type.
#' @export
#' @family config
#' @examples
#' \dontrun{
#' GET("http://httpbin.org/basic-auth/user/passwd")
#' GET(
#'   "http://httpbin.org/basic-auth/user/passwd",
#'   authenticate("user", "passwd")
#' )
#' }
authenticate <- function(user, password, type = "basic") {
  stopifnot(is.character(user), length(user) == 1)
  stopifnot(is.character(password), length(password) == 1)
  stopifnot(is.character(type), length(type) == 1)

  config(httpauth = auth_flags(type), userpwd = paste0(user, ":", password))
}

auth_flags <- function(x = "basic") {
  constants <- c(
    basic = 1,
    digest = 2,
    gssnegotiate = 4,
    ntlm = 8,
    digest_ie = 16,
    any = -17
  )
  x <- match.arg(x, names(constants))

  constants[[x]]
}

context("Config")


test_that("basic authentication works", {
  skip_httpbin()

  h <- handle("http://httpbin.org")
  path <- "basic-auth/user/passwd"

  r <- GET(path = path, handle = h)
  expect_equal(r$status_code, 401)

  r <- GET(
    path = path, handle = h,
    config = authenticate("user", "passwd", "basic")
  )
  expect_equal(r$status_code, 200)

  # Authentication shouldn't persist
  r <- GET(path = path, handle = h)
  expect_equal(r$status_code, 401)
})

test_that("digest authentication works", {
  skip_httpbin()

  h <- handle("http://httpbin.org")
  path <- "digest-auth/qop/user/passwd"

  r <- GET(path = path, handle = h)
  expect_equal(r$status_code, 401)

  r <- GET(
    path = path, handle = h,
    config = authenticate("user", "passwd", "digest")
  )
  expect_equal(r$status_code, 200)
})

test_that("timeout enforced", {
  skip_httpbin()
  expect_error(
    GET("http://httpbin.org/delay/1", timeout(0.5)),
    "(Timeout was reached)|(timed out)"
  )
})

context("Config")


test_that("basic authentication works", {
  h <- handle("http://httpbin.org")
  path <- "basic-auth/user/passwd"

  r <- GET(path = path, handle = h)
  expect_equal(r$status_code, 401)

  r <- GET(
    path = path, handle = h,
    config = authenticate("user", "passwd", "basic")
  )
  expect_equal(r$status_code, 200)

  # Authentication shouldn't persist
  r <- GET(path = path, handle = h)
  expect_equal(r$status_code, 401)
})

test_that("digest authentication works", {
  h <- handle("http://httpbin.org")
  path <- "digest-auth/qop/user/passwd"

  r <- GET(path = path, handle = h)
  expect_equal(r$status_code, 401)

  r <- GET(
    path = path, handle = h,
    config = authenticate("user", "passwd", "digest")
  )
  expect_equal(r$status_code, 200)
})

test_that("timeout enforced", {
  skip_on_cran()
  expect_error(
    GET("http://httpbin.org/delay/1", timeout(0.5)),
    "(Timeout was reached)|(timed out)"
  )
})

.onLoad <- function(libname, pkgname) {
  op <- options()
  op.dplyr <- list(
    httr_oob_default = FALSE,
    httr_oauth_cache = NA
  )
  toset <- !(names(op.dplyr) %in% names(op))
  if (any(toset)) options(op.dplyr[toset])

  invisible()
}

release_questions <- function() {
  c(
    "Have you run all the OAuth demos?"
  )
}

#' S3 object to define response writer.
#'
#' This S3 object allows you to control how the response body is saved.
#'
#' @param subclass,... Class name and fields. Used in class constructors.
#' @keywords internal
#' @export
write_function <- function(subclass, ...) {
  structure(list(...), class = c(subclass, "write_function"))
}

#' Control where the response body is written.
#'
#' The default behaviour is to use `write_memory()`, which caches
#' the response locally in memory. This is useful when talking to APIs as
#' it avoids a round-trip to disk. If you want to save a file that's bigger
#' than memory, use `write_disk()` to save it to a known path.
#'
#' @param path Path to content to.
#' @param overwrite Will only overwrite existing `path` if TRUE.
#' @export
#' @examples
#' tmp <- tempfile()
#' r1 <- GET("https://www.google.com", write_disk(tmp))
#' readLines(tmp)
#'
#' # The default
#' r2 <- GET("https://www.google.com", write_memory())
#'
#' # Save a very large file
#' \dontrun{
#' GET(
#'   "http://www2.census.gov/acs2011_5yr/pums/csv_pus.zip",
#'   write_disk("csv_pus.zip"), progress()
#' )
#' }
write_disk <- function(path, overwrite = FALSE) {
  if (!overwrite && file.exists(path)) {
    stop("Path exists and overwrite is FALSE", call. = FALSE)
  }
  request(output = write_function("write_disk", path = path, file = NULL))
}

#' @rdname write_disk
#' @export
write_memory <- function() {
  request(output = write_function("write_memory"))
}

# Streaming -----------------------------------------------------------------------

#' Process output in a streaming manner.
#'
#' This is the most general way of processing the response from the server -
#' you receive the raw bytes as they come in, and you can do whatever you want
#' with them.
#'
#' @param f Callback function. It should have a single argument, a raw
#'   vector containing the bytes recieved from the server. This will usually
#'   be 16k or less. The return value of the function is ignored.
#' @examples
#' GET(
#'   "https://github.com/jeroen/data/raw/gh-pages/diamonds.json",
#'   write_stream(function(x) {
#'     print(length(x))
#'     length(x)
#'   })
#' )
#' @export
write_stream <- function(f) {
  stopifnot(is.function(f), length(formals(f)) == 1)
  request(output = write_function("write_stream", f = f))
}


request_fetch <- function(x, url, handle) UseMethod("request_fetch")
request_fetch.write_memory <- function(x, url, handle) {
  curl::curl_fetch_memory(url, handle = handle)
}
request_fetch.write_disk <- function(x, url, handle) {
  resp <- curl::curl_fetch_disk(url, x$path, handle = handle)
  resp$content <- path(resp$content)
  resp
}
request_fetch.write_stream <- function(x, url, handle) {
  curl::curl_fetch_stream(url, x$f, handle = handle)
}

path <- function(x) structure(x, class = "path")
length.path <- function(x) file.info(x)$size
is.path <- function(x) inherits(x, "path")

check_encoding <- function(x) {
  if ((tolower(x) %in% tolower(iconvlist()))) {
    return(x)
  }

  message("Invalid encoding ", x, ": defaulting to UTF-8.")
  "UTF-8"
}

guess_encoding <- function(encoding = NULL, type = NULL) {
  # Try extracting from media type
  if (is.null(encoding)) {
    encoding <- tryCatch(
      error = function(err) NULL,
      parse_media(type)$params$charset
    )
  }

  # Fall back to UTF-8
  if (is.null(encoding)) {
    message("No encoding supplied: defaulting to UTF-8.")
    return("UTF-8")
  }

  check_encoding(encoding)
}

parse_text <- function(content, type = NULL, encoding = NULL) {
  encoding <- guess_encoding(encoding, type)
  iconv(readBin(content, character()), from = encoding, to = "UTF-8")
}

parse_auto <- function(content, type = NULL, encoding = NULL, ...) {
  if (length(content) == 0) {
    return(NULL)
  }

  if (is.null(type)) {
    stop("Unknown mime type: can't parse automatically. Use the `type` ",
      "argument to specify manually.",
      call. = FALSE
    )
  }

  mt <- parse_media(type)
  parser <- parsers[[mt$complete]]
  if (is.null(parser)) {
    stop("No automatic parser available for ", mt$complete, ".",
      call. = FALSE
    )
  }

  parser(x = content, type = type, encoding = encoding, ...)
}

parseability <- function(type) {
  if (is.null(type) || type == "") return("raw")
  mt <- parse_media(type)

  if (exists(mt$complete, parsers)) {
    "parsed"
  } else if (mt$type == "text") {
    "text"
  } else {
    "raw"
  }
}

parsers <- new.env(parent = emptyenv())

# Binary formats ---------------------------------------------------------------

# http://www.ietf.org/rfc/rfc4627.txt - section 3. (encoding)
parsers$`application/json` <- function(x, type = NULL, encoding = NULL,
                                       simplifyVector = FALSE, ...) {
  jsonlite::fromJSON(parse_text(x, encoding = "UTF-8"),
    simplifyVector = simplifyVector, ...
  )
}
parsers$`application/x-www-form-urlencoded` <- function(x, encoding = NULL,
                                                        type = NULL, ...) {
  parse_query(parse_text(x, encoding = "UTF-8"))
}

# Text formats -----------------------------------------------------------------
parsers$`image/jpeg` <- function(x, type = NULL, encoding = NULL, ...) {
  need_package("jpeg")
  jpeg::readJPEG(x)
}

parsers$`image/png` <- function(x, type = NULL, encoding = NULL, ...) {
  need_package("png")
  png::readPNG(x)
}

parsers$`text/plain` <- function(x, type = NULL, encoding = NULL, ...) {
  encoding <- guess_encoding(encoding, type)
  parse_text(x, type = type, encoding = encoding)
}

parsers$`text/html` <- function(x, type = NULL, encoding = NULL, ...) {
  need_package("xml2")

  encoding <- guess_encoding(encoding, type)
  xml2::read_html(x, encoding = encoding, ...)
}

parsers$`application/xml` <- function(x, type = NULL, encoding = NULL, ...) {
  need_package("xml2")

  encoding <- guess_encoding(encoding, type)
  xml2::read_xml(x, encoding = encoding, ...)
}

parsers$`text/xml` <- function(x, type = NULL, encoding = NULL, ...) {
  need_package("xml2")

  encoding <- guess_encoding(encoding, type)
  xml2::read_xml(x, encoding = encoding, ...)
}

parsers$`text/csv` <- function(x, type = NULL, encoding = NULL, ...) {
  need_package("readr")

  encoding <- guess_encoding(encoding, type)
  readr::read_csv(x, locale = readr::locale(encoding = encoding), ...)
}

parsers$`text/tab-separated-values` <- function(x, type = NULL, encoding = NULL, ...) {
  need_package("readr")

  encoding <- guess_encoding(encoding, type)
  readr::read_tsv(x, locale = readr::locale(encoding = encoding), ...)
}

context("test-content-parse")

test_that("default encoding is UTF-8", {
  expect_message(x <- guess_encoding(), "defaulting to UTF-8")
  expect_equal(x, "UTF-8")
})

test_that("invalid encoding/type gets UTF-8", {
  expect_message(x <- guess_encoding("abcd"), "defaulting to UTF-8")
  expect_equal(x, "UTF-8")

  expect_message(x <- guess_encoding(NULL, "abcd"), "defaulting to UTF-8")
  expect_equal(x, "UTF-8")
})

test_that("encoding guessed from type", {
  skip_if_not("latin1" %in% tolower(iconvlist()), "latin1 alias not present")
  expect_equal(guess_encoding(NULL, "text/plain; charset=latin1"), "latin1")
})

test_that("explicit type overrides charset", {
  expect_equal(guess_encoding("UTF-8", "text/plain; charset=latin1"), "UTF-8")
})

context("test-content-parse")

test_that("default encoding is UTF-8", {
  expect_message(x <- guess_encoding(), "defaulting to UTF-8")
  expect_equal(x, "UTF-8")
})

test_that("invalid encoding/type gets UTF-8", {
  expect_message(x <- guess_encoding("abcd"), "defaulting to UTF-8")
  expect_equal(x, "UTF-8")

  expect_message(x <- guess_encoding(NULL, "abcd"), "defaulting to UTF-8")
  expect_equal(x, "UTF-8")
})

test_that("encoding guessed from type", {
  expect_equal(guess_encoding(NULL, "text/plain; charset=latin1"), "latin1")

  # But overridden by type
  expect_equal(guess_encoding("UTF-8", "text/plain; charset=latin1"), "UTF-8")
})

#' Maintain a pool of handles.
#'
#' The handle pool is used to automatically reuse Curl handles for the same
#' scheme/host/port combination. This ensures that the http session is
#' automatically reused, and cookies are maintained across requests to a site
#' without user intervention.
#'
#' @format An environment.
#' @keywords internal
handle_pool <- new.env(hash = TRUE, parent = emptyenv())

handle_name <- function(url) {
  build_url(parse_url(url)[c("scheme", "hostname", "port")])
}

#' @export
#' @rdname handle_pool
handle_find <- function(url) {
  name <- handle_name(url)
  if (exists(name, handle_pool)) {
    handle <- handle_pool[[name]]
  } else {
    handle <- handle(name)
    handle_pool[[name]] <- handle
  }

  handle
}

#' @export
#' @rdname handle_pool
handle_reset <- function(url) {
  name <- handle_name(url)
  if (exists(name, envir = handle_pool)) {
    rm(list = name, envir = handle_pool)
  }
}

#' Set user agent.
#'
#' Override the default RCurl user agent of `NULL`
#'
#' @param agent string giving user agent
#' @export
#' @family config
#' @examples
#' GET("http://httpbin.org/user-agent")
#' GET("http://httpbin.org/user-agent", user_agent("httr"))
user_agent <- function(agent) {
  stopifnot(is.character(agent), length(agent) == 1)

  config(useragent = agent)
}

library(httr)

# Noun Project has an API secured with OAuth 1.0a One-legged. A client key and a
# secret must be used to sign requests when accessing the API. This is an
# exemple using OAuth 1.0a One legged auth mechanism
# http://oauthbible.com/#oauth-10a-one-legged

# 1. Register an application to get required keys:
# https://thenounproject.com/accounts/login/?next=/developers/apps/
nouns_app <- oauth_app("noun_project",
  key = rstudioapi::askForPassword(),
  secret = rstudioapi::askForPassword()
)

# 2. Each request must be signed using the app key and secret
#    see ?oauth_signature for more information on signature
url <- "http://api.thenounproject.com/icon/15"
signature <- oauth_signature(url, method = "GET", app = nouns_app)
res <- GET(url, oauth_header(signature))
stop_for_status(res)
content(res)

# 3. Create a wrapper function to sign each request more easily
get_nouns_api <- function(endpoint,
                          baseurl = "http://api.thenounproject.com/",
                          app = nouns_app,
                          ...) {
  url <- modify_url(baseurl, path = endpoint)
  info <- oauth_signature(url, app = app)
  header_oauth <- oauth_header(info)
  GET(url, header_oauth, ...)
}
res <- get_nouns_api("collections")
stop_for_status(res)
content(res)

# 4. Signing request requires the METHOD used. If the API has a POST for
# example, you must take it into account.
url <- "http://api.thenounproject.com/notify/publish?test=1"
signature <- oauth_signature(url, method = "POST", app = nouns_app)
res <- POST(url, oauth_header(signature), body = list(icons = 15), encode = "json")
stop_for_status(res)
content(res)

#' Diagnose common configuration problems
#'
#' Currently one check: that curl uses nss.
#'
#' @export
httr_dr <- function() {
  check_for_nss()
}

check_for_nss <- function() {
  if (!grepl("^NSS", curl::curl_version()$ssl_version)) return()

  warning('
  ------------------------------------------------------------------------
  Your installed RCurl is linked to the NSS library (`libcurl4-nss-dev`)
  which is likely to cause issues.

  To resolve the problem:

  1. Quit R.

  2. Install OpenSSL (`apt-get install libcurl4-openssl-dev`) or
     GnuTLS (`apt-get install libcurl4-gnutls-dev`) variants of libCurl.

  3. Restart R.

  4. Reinstall RCurl: `install.packages("RCurl")`.
  ------------------------------------------------------------------------
  ', call. = FALSE)
}

#' Walk the user through the OAuth2 dance without a local webserver.
#'
#' This performs a similar function to [oauth_listener()],
#' but without running a local webserver.  This manual process can be useful
#' in situations where the user is remotely accessing the machine outside a
#' browser (say via ssh) or when it's not possible to successfully receive a
#' callback (such as when behind a firewall).
#'
#' This function should generally not be called directly by the user.
#'
#' @param request_url the url to provide to the user
#' @export
#' @keywords internal
oauth_exchanger <- function(request_url) {
  BROWSE(request_url)

  authorization_code <- str_trim(readline("Enter authorization code: "))
  list(code = authorization_code)
}

#' Extract the content type of a response
#'
#' @param x A response
#' @return A string giving the complete mime type, with all parameters
#'   stripped off.
#' @export
#' @examples
#' r1 <- GET("http://httpbin.org/image/png")
#' http_type(r1)
#' headers(r1)[["Content-Type"]]
#' 
#' r2 <- GET("http://httpbin.org/ip")
#' http_type(r2)
#' headers(r2)[["Content-Type"]]
http_type <- function(x) {
  stopifnot(is.response(x))

  type <- x$headers[["Content-Type"]] %||%
    mime::guess_type(x$url, empty = "application/octet-stream")

  parse_media(type)$complete
}

#' POST file to a server.
#'
#' @inherit GET params return
#' @param body One of the following:
#'
#'   * `FALSE`: No body. This is typically not used with `POST`,
#'     `PUT`, or `PATCH`, but can be useful if you need to send a
#'     bodyless request (like `GET`) with `VERB()`.
#'   * `NULL`: An empty body
#'   * `""`: A length 0 body
#'   * `upload_file("path/")`: The contents of a file.  The mime
#'     type will be guessed from the extension, or can be supplied explicitly
#'     as the second argument to `upload_file()`
#'   * A character or raw vector: sent as is in body. Use
#'     [content_type()] to tell the server what sort of data
#'     you are sending.
#'   * A named list: See details for encode.
#' @param encode If the body is a named list, how should it be encoded? Can be
#'   one of form (application/x-www-form-urlencoded), multipart,
#'   (multipart/form-data), or json (application/json).
#'
#'   For "multipart", list elements can be strings or objects created by
#'   [upload_file()]. For "form", elements are coerced to strings
#'   and escaped, use `I()` to prevent double-escaping. For "json",
#'   parameters are automatically "unboxed" (i.e. length 1 vectors are
#'   converted to scalars). To preserve a length 1 vector as a vector,
#'   wrap in `I()`. For "raw", either a character or raw vector. You'll
#'   need to make sure to set the [content_type()] yourself.
#' @export
#' @family http methods
#' @examples
#' \dontrun{
#' b2 <- "http://httpbin.org/post"
#' POST(b2, body = "A simple text string")
#' POST(b2, body = list(x = "A simple text string"))
#' POST(b2, body = list(y = upload_file(system.file("CITATION"))))
#' POST(b2, body = list(x = "A simple text string"), encode = "json")
#'
#' # body can also be provided as a json string directly to deal
#' # with specific case, like an empty element in the json string.
#' # passing as string directly
#' POST(b2, body = '{"a":1,"b":{}}', encode = "raw")
#' # or building the json string before
#' json_body <- jsonlite::toJSON(list(a = 1, b = NULL), auto_unbox = TRUE)
#' POST(b2, body = json_body, encode = "raw")
#'
#' # Various types of empty body:
#' POST(b2, body = NULL, verbose())
#' POST(b2, body = FALSE, verbose())
#' POST(b2, body = "", verbose())
#' }
POST <- function(url = NULL, config = list(), ..., body = NULL,
                 encode = c("multipart", "form", "json", "raw"),
                 handle = NULL) {
  encode <- match.arg(encode)

  hu <- handle_url(handle, url, ...)
  req <- request_build("POST", hu$url, body_config(body, match.arg(encode)), as.request(config), ...)
  request_perform(req, hu$handle$handle)
}

context("Response")


test_that("status codes returned as expected", {
  skip_httpbin()
  expect_equal(GET("http://httpbin.org/status/320")$status_code, 320)
  expect_equal(GET("http://httpbin.org/status/404")$status_code, 404)
  expect_equal(GET("http://httpbin.org/status/418")$status_code, 418)
})

test_that("DELETE deletes", {
  skip_httpbin()
  expect_equal(GET("http://httpbin.org/delete")$status_code, 405)
  expect_equal(DELETE("http://httpbin.org/delete")$status_code, 200)
})

test_that("POST posts", {
  skip_httpbin()
  expect_equal(GET("http://httpbin.org/post")$status_code, 405)
  expect_equal(POST("http://httpbin.org/post")$status_code, 200)
})

test_that("PATCH patches", {
  skip_httpbin()
  expect_equal(GET("http://httpbin.org/patch")$status_code, 405)
  expect_equal(PATCH("http://httpbin.org/patch")$status_code, 200)
})

test_that("PUT puts", {
  skip_httpbin()
  expect_equal(GET("http://httpbin.org/put")$status_code, 405)
  expect_equal(PUT("http://httpbin.org/put")$status_code, 200)
})

test_that("headers returned as expected", {
  skip_httpbin()
  round_trip <- function(...) {
    req <- GET("http://httpbin.org/headers", add_headers(...))
    headers <- content(req)$headers
    names(headers) <- tolower(names(headers))
    headers
  }

  expect_equal(round_trip(a = 1)$a, "1")
  expect_equal(round_trip(a = "a + b")$a, "a + b")
})

test_that("application/json responses parsed as lists", {
  skip_httpbin()
  test_user_agent <- function(user_agent = NULL) {
    response <- GET(
      "http://httpbin.org/user-agent",
      add_headers("user-agent" = user_agent)
    )

    expect_equal(response$status_code, 200)
    expected_reply <- list("user-agent" = user_agent %||% default_ua())
    expect_equal(expected_reply, content(response))
  }

  test_user_agent()
  test_user_agent("gunicorn-client/0.1")
})

#' The response object.
#'
#' @description
#' The response object captures all information from a request.  It includes
#' fields:
#'
#' * `url` the url the request was actually sent to (after redirects)
#' * `handle` the handle associated with the url
#' * `status_code` the http status code
#' * `header` a named list of headers returned by the server
#' * `cookies` a named list of cookies returned by the server
#' * `content` the body of the response, as raw vector. See [content()] for various ways to access the content.
#' * `time` request timing information
#' * `config` configuration for the request
#'
#' @details For non-http(s) responses, some parts including the status and
#'   header may not be interpretable the same way as http responses.
#'
#' @name response
#' @family response methods
NULL

response <- function(...) {
  structure(list(...), class = "response")
}

is.response <- function(x) {
  inherits(x, "response")
}

#' @export
print.response <- function(x, ..., max.lines = 10, width = getOption("width")) {
  content_type <- x$headers$`content-type`

  cat("Response [", x$url, "]\n", sep = "")
  cat("  Date: ", format(x$date, "%Y-%m-%d %H:%M"), "\n", sep = "")
  cat("  Status: ", x$status_code, "\n", sep = "")
  cat("  Content-Type: ", content_type %||% "<unknown>", "\n", sep = "")

  size <- length(x$content)
  if (size == 0) {
    cat("<EMPTY BODY>\n")
    return()
  }

  cat("  Size: ", bytes(size), "\n", sep = "")
  if (is.path(x$content)) {
    cat("<ON DISK> ", x$content)
    return()
  }
  if (!is_text(content_type)) {
    cat("<BINARY BODY>\n")
    return()
  }

  # Content is text, so print up to `max.lines` lines, truncating each line to
  # at most `width` characters wide
  suppressMessages(text <- content(x, "text"))

  breaks <- gregexpr("\n", text, fixed = TRUE)[[1]]
  last_line <- breaks[min(length(breaks), max.lines)]
  lines <- strsplit(substr(text, 1, last_line), "\n")[[1]]

  too_wide <- nchar(lines) > width
  lines[too_wide] <- paste0(substr(lines[too_wide], 1, width - 3), "...")

  cat(lines, sep = "\n")
  if (max.lines < length(breaks)) cat("...\n")

  invisible(x)
}

is_text <- function(type) {
  if (is.null(type)) return(FALSE)

  media <- parse_media(type)
  if (media$type == "text") return(TRUE)
  if (media$type == "application" && media$subtype == "json") return(TRUE)

  FALSE
}

#' @export
as.character.response <- function(x, ...) {
  content(x, "text")
}

#' Compute caching information for a response.
#'
#' `cache_info()` gives details of cacheability of a response,
#' `rerequest()` re-performs the original request doing as little work
#' as possible (if not expired, returns response as is, or performs
#' revalidation if Etag or Last-Modified headers are present).
#'
#' @param r A response
#' @export
#' @examples
#' # Never cached, always causes redownload
#' r1 <- GET("https://www.google.com")
#' cache_info(r1)
#' r1$date
#' rerequest(r1)$date
#'
#' # Expires in a year
#' r2 <- GET("https://www.google.com/images/srpr/logo11w.png")
#' cache_info(r2)
#' r2$date
#' rerequest(r2)$date
#'
#' \dontrun{
#' # Has last-modified and etag, so does revalidation
#' r3 <- GET("http://httpbin.org/cache")
#' cache_info(r3)
#' r3$date
#' rerequest(r3)$date
#'
#' # Expires after 5 seconds
#' r4 <- GET("http://httpbin.org/cache/5")
#' cache_info(r4)
#' r4$date
#' rerequest(r4)$date
#' Sys.sleep(5)
#' cache_info(r4)
#' rerequest(r4)$date
#' }
cache_info <- function(r) {
  stopifnot(is.response(r))

  # If parsing fails using -Inf; If missing use NA
  expires <- parse_http_date(r$headers$expires, Inf) %||% NULL

  # Parse cache control
  control <- parse_cache_control(r$headers$`cache-control`)
  max_age <- as.integer(control$`max-age`) %||% NULL

  # Compute expiry
  if (!is.null(max_age)) {
    expires <- r$date + max_age
  } else if (!is.null(r$headers$expires)) {
    expires <- parse_http_date(r$headers$expires, -Inf)
  } else {
    expires <- NULL
  }

  # Is this cacheable?
  cacheable <- r$request$method %in% c("GET", "HEAD") &&
    status_code(r) %in% c(200L, 203L, 300L, 301L, 410L) &&
    (!is.null(expires) || !is.null(r$headers$etag) ||
      !is.null(r$headers$`last-modified`)) &&
    !any(c("no-store", "no-cache") %in% control$flags)
  # What impact should any(c("public", "private") %in% control$flags) have?

  structure(
    list(
      method = r$method,
      url = r$url,

      cacheable = cacheable,
      expires = expires,
      etag = r$headers$etag %||% NULL,
      modified = parse_http_date(r$headers$`last-modified`, NULL)
    ),
    class = "cache_info"
  )
}

is.cache_info <- function(x) inherits(x, "cache_info")

#' @export
print.cache_info <- function(x, ...) {
  cat("<cache_info> ", x$method, " ", x$url, "\n", sep = "")
  cat("  Cacheable:     ", x$cacheable, "\n", sep = "")
  if (!is.null(x$expires)) {
    cat("  Expires:       ", http_date(x$expires), sep = "")
    if (x$expires < Sys.time()) cat(" <expired>")
    cat("\n")
  }
  cat("  Last-Modified: ", http_date(x$modified), "\n", sep = "")
  cat("  Etag:          ", x$etag, "\n", sep = "")
  invisible(x)
}

parse_cache_control <- function(x) {
  if (is.null(x)) return(list())

  pieces <- strsplit(x, ",")[[1]]
  pieces <- gsub("^\\s+|\\s+$", "", pieces)
  pieces <- tolower(pieces)

  is_value <- grepl("=", pieces)
  flags <- pieces[!is_value]

  keyvalues <- strsplit(pieces[is_value], "\\s*=\\s*")
  keys <- c(rep("flags", length(flags)), lapply(keyvalues, "[[", 1))
  values <- c(flags, lapply(keyvalues, "[[", 2))

  stats::setNames(values, keys)
}

#' @rdname cache_info
#' @export
rerequest <- function(r) {
  x <- cache_info(r)
  if (!x$cacheable) {
    return(reperform(r))
  }

  # Cacheable, and hasn't expired
  if (!is.null(x$expires) && x$expires >= Sys.time()) {
    return(r)
  }

  # Requires validation
  req <- c(
    r$request,
    request(headers = c(
      `If-Modified-Since` = http_date(x$modified),
      `If-None-Match` = x$etag
    ))
  )
  validated <- request_perform(req, r$handle)

  if (status_code(validated) == 304L) {
    r
  } else {
    validated
  }
}

reperform <- function(resp) {
  request_perform(resp$request, resp$handle)
}

context("Headers")

# Setting ---------------------------------------------------------------------

test_that("Only last duplicated header kept in add_headers", {
  expect_equal(add_headers(x = 1, x = 2)$headers, c(x = "2"))
})

test_that("Only last duplicated header kept when combined", {
  out <- c(add_headers(x = 1), add_headers(x = 2))
  expect_equal(out$headers, c(x = "2"))
})

# Getting ---------------------------------------------------------------------

test_that("All headers captures headers from redirects", {
  skip("Currently broken")
  r1 <- GET("https://httpbin.org/redirect/1")
  expect_equal(length(r1$all_headers), 1 + 1)

  r3 <- GET("https://httpbin.org/redirect/3")
  expect_equal(length(r3$all_headers), 3 + 1)
})

# Parsing ---------------------------------------------------------------------

test_that("Trailing line breaks removed", {
  header <- charToRaw("HTTP/1.1 200 OK\r\nA: B\r\n")
  expect_equal(parse_http_headers(header)[[1]]$headers$A, "B")
})

test_that("Invalid header raises error", {
  lines <- c(
    "HTTP/1.1 200 OK",
    "A: B",
    "Invalid"
  )
  header <- charToRaw(paste(lines, collapse = "\n"))
  expect_warning(parse_http_headers(header), "Failed to parse headers")
})

test_that("http status line only needs two components", {
  headers <- parse_http_headers(charToRaw("HTTP/1.1 200"))
  expect_equal(headers[[1]]$status, 200L)
})

test_that("Key/value parsing tolerates multiple ':'", {
  lines <- c(
    "HTTP/1.1 200 OK",
    "A: B:C",
    "D:E:F"
  )
  header <- charToRaw(paste(lines, collapse = "\n"))

  expect_equal(parse_http_headers(header)[[1]]$headers$A, "B:C")
  expect_equal(parse_http_headers(header)[[1]]$headers$D, "E:F")
})

test_that("cache_info() handles flagless cache control", {
  response <- list(flags = "private", `max-age` = "0", `max-stale` = "0")
  expect_equal(response[1], parse_cache_control("private"))
  expect_equal(response[2], parse_cache_control("max-age=0"))
  expect_equal(response[1:2], parse_cache_control("private, max-age=0"))
  expect_equal(
    response, parse_cache_control("private, max-age=0, max-stale=0")
  )
})

#' SHA1 hash
#'
#' Creates a SHA1 hash of data using either HMAC or RSA.
#'
#' @param key The key to create the hash with
#' @param string data to securely hash
#' @param method The method to use, either HMAC-SHA1 or RSA-SHA1
#' @keywords internal
#' @export
sha1_hash <- function(key, string, method = "HMAC-SHA1") {
  if (is.character(string)) {
    string <- charToRaw(paste(string, collapse = "\n"))
  }
  if (is.character(key)) {
    key <- charToRaw(paste(key, collapse = "\n"))
  }

  if (!method %in% c("HMAC-SHA1", "RSA-SHA1")) {
    stop(paste0("Unsupported hashing method: ", method), call. = FALSE)
  }

  if (method == "HMAC-SHA1") {
    hash <- openssl::sha1(string, key = key)
  } else {
    hash <- openssl::signature_create(string, openssl::sha1, key = key)
  }

  openssl::base64_encode(hash)
}

#' HMAC SHA1
#'
#' As described in <http://datatracker.ietf.org/doc/rfc2104/>.
#'
#' @param key secret key
#' @param string data to securely hash
#' @keywords internal
#' @export
hmac_sha1 <- function(key, string) {
  sha1_hash(key, string, "HMAC-SHA1")
}

#' Add a progress bar.
#'
#' @param type Type of progress to display: either number of bytes uploaded
#'   or downloaded.
#' @param con Connection to send output too. Usually `stdout()` or
#'    `stderr`.
#' @export
#' @examples
#' cap_speed <- config(max_recv_speed_large = 10000)
#' \donttest{
#' # If file size is known, you get a progress bar:
#' x <- GET("http://httpbin.org/bytes/102400", progress(), cap_speed)
#' # Otherwise you get the number of bytes downloaded:
#' x <- GET("http://httpbin.org/stream-bytes/102400", progress(), cap_speed)
#' }
progress <- function(type = c("down", "up"), con = stdout()) {
  type <- match.arg(type)

  request(options = list(
    noprogress = FALSE,
    progressfunction = progress_bar(type, con)
  ))
}

progress_bar <- function(type, con) {
  bar <- NULL

  show_progress <- function(down, up) {
    if (type == "down") {
      total <- down[[1]]
      now <- down[[2]]
    } else {
      total <- up[[1]]
      now <- up[[2]]
    }

    if (total == 0 && now == 0) {
      # Reset progress bar when seeing first byte
      bar <<- NULL
    } else if (total == 0) {
      cat("\rDownloading: ", bytes(now, digits = 2), "     ", sep = "", file = con)
      utils::flush.console()
      # Can't automatically add newline on completion because there's no
      # way to tell when then the file has finished downloading
    } else {
      if (is.null(bar)) {
        bar <<- utils::txtProgressBar(max = total, style = 3, file = con)
      }
      utils::setTxtProgressBar(bar, now)
      if (now == total) close(bar)
    }

    TRUE
  }

  show_progress
}


bytes <- function(x, digits = 3, ...) {
  power <- min(floor(log(abs(x), 1000)), 4)
  if (power < 1) {
    unit <- "B"
  } else {
    unit <- c("kB", "MB", "GB", "TB")[[power]]
    x <- x / (1000^power)
  }

  formatted <- format(signif(x, digits = digits),
    big.mark = ",",
    scientific = FALSE
  )

  paste0(formatted, " ", unit)
}

"%||%" <- function(a, b) {
  if (length(a) > 0) a else b
}

timestamp <- function(x = Sys.time()) {
  format(x, "%Y-%m-%dT%H:%M:%SZ", tz = "UTC")
}

sort_names <- function(x) x[order(names(x))]

nonce <- function(length = 10) {
  paste(sample(c(letters, LETTERS, 0:9), length, replace = TRUE),
    collapse = ""
  )
}

has_env_var <- function(x) !identical(Sys.getenv(x), "")

named <- function(x) x[has_name(x)]
unnamed <- function(x) x[!has_name(x)]

has_name <- function(x) {
  nms <- names(x)
  if (is.null(nms)) {
    return(rep(FALSE, length(x)))
  }

  !is.na(nms) & nms != ""
}

travis_encrypt <- function(vars) {
  values <- Sys.getenv(vars)
  cat(paste0("travis encrypt ", paste0(vars, "=", values, collapse = " ")))
}

is_installed <- function(pkg) {
  system.file(package = pkg) != ""
}

need_package <- function(pkg) {
  if (is_installed(pkg)) return(invisible())

  stop("Please install ", pkg, " package", call. = FALSE)
}

last <- function(x) {
  if (length(x) < 1) return(x)
  x[[length(x)]]
}

compact <- function(x) {
  empty <- vapply(x, is_empty, logical(1))
  x[!empty]
}

is_empty <- function(x) length(x) == 0

keep_last <- function(...) {
  x <- c(...)
  x[!duplicated(names(x), fromLast = TRUE)]
}

named_vector <- function(title, x) {
  if (length(x) == 0) return()

  cat(title, ":\n", sep = "")
  bullets <- paste0("* ", names(x), ": ", as.character(x))
  cat(bullets, sep = "\n")
}

keep_last <- function(...) {
  x <- c(...)
  x[!duplicated(names(x), fromLast = TRUE)]
}

find_cert_bundle <- function() {
  if (.Platform$OS.type != "windows") {
    return()
  }

  env <- Sys.getenv("CURL_CA_BUNDLE")
  if (!identical(env, "")) {
    return(env)
  }

  bundled <- file.path(R.home("etc"), "curl-ca-bundle.crt")
  if (file.exists(bundled)) {
    return(bundled)
  }

  # Fall back to certificate bundle in openssl
  system.file("cacert.pem", package = "openssl")
}

isFALSE <- function(x) identical(x, FALSE)

#' Set curl options.
#'
#' Generally you should only need to use this function to set CURL options
#' directly if there isn't already a helpful wrapper function, like
#' [set_cookies()], [add_headers()] or
#' [authenticate()]. To use this function effectively requires
#' some knowledge of CURL, and CURL options. Use [httr_options()] to
#' see a complete list of available options. To see the libcurl documentation
#' for a given option, use [curl_docs()].
#'
#' Unlike Curl (and RCurl), all configuration options are per request, not
#' per handle.
#'
#' @seealso [set_config()] to set global config defaults, and
#'  [with_config()] to temporarily run code with set options.
#' @param token An OAuth token (1.0 or 2.0)
#' @family config
#' @family ways to set configuration
#' @seealso All known available options are listed in [httr_options()]
#' @param ... named Curl options.
#' @export
#' @examples
#' # There are a number of ways to modify the configuration of a request
#' # * you can add directly to a request
#' HEAD("https://www.google.com", verbose())
#' 
#' # * you can wrap with with_config()
#' with_config(verbose(), HEAD("https://www.google.com"))
#' 
#' # * you can set global with set_config()
#' old <- set_config(verbose())
#' HEAD("https://www.google.com")
#' # and re-establish the previous settings with
#' set_config(old, override = TRUE)
#' HEAD("https://www.google.com")
#' # or
#' reset_config()
#' HEAD("https://www.google.com")
#' 
#' # If available, you should use a friendly httr wrapper over RCurl
#' # options. But you can pass Curl options (as listed in httr_options())
#' # in config
#' HEAD("https://www.google.com/", config(verbose = TRUE))
config <- function(..., token = NULL) {
  request(options = list(...), auth_token = token)
}

#' List available options.
#'
#' This function lists all available options for [config()].
#' It provides both the short R name which you use with httr, and the longer
#' Curl name, which is useful when searching the documentation. `curl_doc`
#' opens a link to the libcurl documentation for an option in your browser.
#'
#' RCurl and httr use slightly different names to libcurl: the initial
#' `CURLOPT_` is removed, all underscores are converted to periods and
#' the option is given in lower case.  Thus "CURLOPT_SSLENGINE_DEFAULT"
#' becomes "sslengine.default".
#'
#' @param x An option name (either short or full).
#' @param matches If not missing, this restricts the output so that either
#'   the httr or curl option matches this regular expression.
#' @return A data frame with three columns:
#' \item{httr}{The short name used in httr}
#' \item{libcurl}{The full name used by libcurl}
#' \item{type}{The type of R object that the option accepts}
#' @export
#' @examples
#' httr_options()
#' httr_options("post")
#' 
#' # Use curl_docs to read the curl documentation for each option.
#' # You can use either the httr or curl option name.
#' curl_docs("userpwd")
#' curl_docs("CURLOPT_USERPWD")
httr_options <- function(matches) {
  constants <- curl::curl_options()
  constants <- constants[order(names(constants))]

  rcurl <- tolower(names(constants))

  opts <- data.frame(
    httr = rcurl,
    libcurl = translate_curl(rcurl),
    type = curl_option_types(constants),
    stringsAsFactors = FALSE
  )

  if (!missing(matches)) {
    sel <- grepl(matches, opts$httr, ignore.case = TRUE) |
      grepl(matches, opts$libcurl, ignore.case = TRUE)
    opts <- opts[sel, , drop = FALSE]
  }

  opts
}

curl_option_types <- function(opts = curl::curl_options()) {
  type_name <- c("integer", "string", "function", "number")
  type <- floor(opts / 10000)

  type_name[type + 1]
}

#' @export
print.opts_list <- function(x, ...) {
  cat(paste0(format(names(x)), ": ", x, collapse = "\n"), "\n", sep = "")
  invisible(x)
}

translate_curl <- function(x) {
  paste0("CURLOPT_", gsub(".", "_", toupper(x), fixed = TRUE))
}

#' @export
#' @rdname httr_options
curl_docs <- function(x) {
  stopifnot(is.character(x), length(x) == 1)

  opts <- httr_options()
  if (x %in% opts$httr) {
    x <- opts$libcurl[match(x, opts$httr)]
  }
  if (!(x %in% opts$libcurl)) {
    stop(x, " is not a known curl option", call. = FALSE)
  }

  url <- paste0("http://curl.haxx.se/libcurl/c/", x, ".html")
  BROWSE(url)
}

cache <- new.env(parent = emptyenv())
cache$default_ua <- NULL

default_ua <- function() {
  if (is.null(cache$default_ua)) {
    versions <- c(
      libcurl = curl::curl_version()$version,
      `r-curl` = as.character(utils::packageVersion("curl")),
      httr = as.character(utils::packageVersion("httr"))
    )
    cache$default_ua <- paste0(names(versions), "/", versions, collapse = " ")
  }
  cache$default_ua
}

#' Set (and reset) global httr configuration.
#'
#' @param config Settings as generated by [add_headers()],
#'   [set_cookies()] or [authenticate()].
#' @param override if `TRUE`, ignore existing settings, if `FALSE`,
#'   combine new config with old.
#' @return invisibility, the old global config.
#' @family ways to set configuration
#' @export
#' @examples
#' GET("http://google.com")
#' set_config(verbose())
#' GET("http://google.com")
#' reset_config()
#' GET("http://google.com")
set_config <- function(config, override = FALSE) {
  stopifnot(is.request(config))

  old <- getOption("httr_config") %||% config()
  if (!override) config <- c(old, config)
  options(httr_config = config)
  invisible(old)
}

#' @export
#' @rdname set_config
reset_config <- function() set_config(config(), TRUE)

#' Execute code with configuration set.
#'
#' @family ways to set configuration
#' @inheritParams set_config
#' @param expr code to execute under specified configuration
#' @export
#' @examples
#' with_config(verbose(), {
#'   GET("http://had.co.nz")
#'   GET("http://google.com")
#' })
#' 
#' # Or even easier:
#' with_verbose(GET("http://google.com"))
with_config <- function(config = config(), expr, override = FALSE) {
  stopifnot(is.request(config))

  old <- set_config(config, override)
  on.exit(set_config(old, override = TRUE))
  force(expr)
}

#' @export
#' @param ... Other arguments passed on to [verbose()]
#' @rdname with_config
with_verbose <- function(expr, ...) {
  with_config(verbose(...), expr)
}

#' \pkg{httr} makes http easy.
#'
#' `httr` is organised around the six most common http verbs:
#' [GET()], [PATCH()],
#' [POST()], [HEAD()],
#' [PUT()], and [DELETE()].
#'
#' Each request returns a [response()] object which provides
#' easy access to status code, cookies, headers, timings, and other useful
#' info.  The content of the request is available as a raw vector
#' ([content()]), character vector ([text_content()]),
#' or parsed into an R object ([parsed_content()]), currently for
#' html, xml, json, png and jpeg).
#'
#' Requests can be modified by various config options like
#' [set_cookies()], [add_headers()],
#' [authenticate()], [use_proxy()],
#' [verbose()], and [timeout()]
#'
#' httr supports OAuth 1.0 and 2.0. Use [oauth1.0_token()] and
#' [oauth2.0_token()] to get user tokens, and
#' [sign_oauth1.0()] and [sign_oauth2.0()] to sign
#' requests. The demos directory has twelve demos of using OAuth: four for 1.0
#' (linkedin, twitter, vimeo, and yahoo) and eight for 2.0 (azure, facebook,
#' github, google, linkedin, reddit, yahoo, and yelp).
#'
#' @keywords internal
"_PACKAGE"

#' Extract the headers from a response
#'
#' @param x A request object
#' @seealso [add_headers()] to send additional headers in a
#'   request
#' @export
#' @examples
#' r <- GET("http://httpbin.org/get")
#' headers(r)
headers <- function(x) UseMethod("headers")

#' @export
headers.response <- function(x) {
  x$headers
}

#' Add additional headers to a request.
#'
#' Wikipedia provides a useful list of common http headers:
#' <http://en.wikipedia.org/wiki/List_of_HTTP_header_fields>.
#'
#' @param ... named header values.  To stop an existing header from being
#'   set, pass an empty string: `""`.
#' @param .headers a named character vector
#' @export
#' @seealso [accept()] and [content_type()] for
#'   convenience functions for setting accept and content-type headers.
#' @family config
#' @examples
#' add_headers(a = 1, b = 2)
#' add_headers(.headers = c(a = "1", b = "2"))
#' 
#' GET("http://httpbin.org/headers")
#' 
#' # Add arbitrary headers
#' GET(
#'   "http://httpbin.org/headers",
#'   add_headers(version = version$version.string)
#' )
#' 
#' # Override default headers with empty strings
#' GET("http://httpbin.org/headers", add_headers(Accept = ""))
add_headers <- function(..., .headers = character()) {
  request(headers = c(..., .headers))
}


#' Set content-type and accept headers.
#'
#' These are convenient wrappers aroud [add_headers()].
#'
#' `accept_json`/`accept_xml` and
#' `content_type_json`/`content_type_xml` are useful shortcuts to
#' ask for json or xml responses or tell the server you are sending json/xml.
#'
#' @param type A mime type or a file extension. If a file extension (i.e. starts
#'   with `.`) will guess the mime type using [mime::guess_type()].
#' @export
#' @examples
#' GET("http://httpbin.org/headers")
#' 
#' GET("http://httpbin.org/headers", accept_json())
#' GET("http://httpbin.org/headers", accept("text/csv"))
#' GET("http://httpbin.org/headers", accept(".doc"))
#' 
#' GET("http://httpbin.org/headers", content_type_xml())
#' GET("http://httpbin.org/headers", content_type("text/csv"))
#' GET("http://httpbin.org/headers", content_type(".xml"))
content_type <- function(type) {
  if (is.null(type)) return()

  if (substr(type, 1, 1) == ".") {
    type <- mime::guess_type(type, empty = NULL)
  }

  add_headers("Content-Type" = type)
}

#' @export
#' @rdname content_type
content_type_json <- function() content_type("application/json")

#' @export
#' @rdname content_type
content_type_xml <- function() content_type("application/xml")



#' @export
#' @rdname content_type
accept <- function(type) {
  if (substr(type, 1, 1) == ".") {
    type <- mime::guess_type(type, empty = NULL)
  }
  add_headers("Accept" = type)
}

#' @export
#' @rdname content_type
accept_json <- function() accept("application/json")

#' @export
#' @rdname content_type
accept_xml <- function() accept("application/xml")

# Parses a header lines as recieved from libcurl. Multiple responses
# will be intermingled, each separated by an http status line.
parse_http_headers <- function(raw) {
  lines <- strsplit(rawToChar(raw), "\r?\n")[[1]]

  new_response <- grepl("^HTTP", lines)
  grps <- cumsum(new_response)

  lapply(unname(split(lines, grps)), parse_single_header)
}

parse_single_header <- function(lines) {
  status <- parse_http_status(lines[[1]])

  # Parse headers into name-value pairs
  header_lines <- lines[lines != ""][-1]
  pos <- regexec("^([^:]*):\\s*(.*)$", header_lines)
  pieces <- regmatches(header_lines, pos)

  n <- vapply(pieces, length, integer(1))
  if (any(n != 3)) {
    bad <- header_lines[n != 3]
    pieces <- pieces[n == 3]

    warning("Failed to parse headers:\n", paste0(bad, "\n"), call. = FALSE)
  }

  names <- vapply(pieces, "[[", 2, FUN.VALUE = character(1))
  values <- lapply(pieces, "[[", 3)
  headers <- insensitive(stats::setNames(values, names))

  list(status = status$status, version = status$version, headers = headers)
}

parse_http_status <- function(x) {
  status <- as.list(strsplit(x, "\\s+")[[1]])
  names(status) <- c("version", "status", "message")[seq_along(status)]
  status$status <- as.integer(status$status)


  status
}

context("Body")

round_trip <- function(...) {
  skip_httpbin()
  content(POST("http://httpbin.org/post", ...))
}

data_path <- upload_file("data.txt")
data <- readLines("data.txt")

test_that("NULL body gives empty data element", {
  out <- round_trip(body = NULL)
  expect_equal(out$data, "")
})

test_that("FALSE body doesn't send body", {
  out <- round_trip(body = FALSE)
  expect_equal(out$data, NULL)
})

test_that("string/raw in body gives same string in data element", {
  out <- round_trip(body = "test")
  expect_equal(out$data, "test")

  out <- round_trip(body = charToRaw("test"))
  expect_equal(out$data, "test")
})

test_that("string/raw in body doesn't lose content type", {
  skip_httpbin()

  body <- charToRaw("test")
  content_type <- "application/awesome"
  response <- content(POST("http://httpbin.org/post",
    body = body,
    add_headers("Content-Type" = content_type)
  ))
  expect_equal(response$headers$`Content-Type`, content_type)
})

test_that("empty list gives empty body", {
  out <- round_trip(body = list(), encode = "form")
  expect_equal(out$data, "")

  out <- round_trip(body = list(), encode = "multipart")
  expect_equal(out$data, "")
})

test_that("named list matches form results (encode = 'form')", {
  out <- round_trip(body = list(a = 1, b = 2), encode = "form")
  expect_equal(out$form$a, "1")
  expect_equal(out$form$b, "2")
})

test_that("named list matches form results (encode = 'multipart')", {
  out <- round_trip(body = list(a = 1, b = 2), encode = "multipart")
  expect_equal(out$form$a, "1")
  expect_equal(out$form$b, "2")
})

test_that("named list matches form results (encode = 'json')", {
  out <- round_trip(body = list(a = 1, b = 2), encode = "json")
  expect_equal(out$json$a[[1]], 1)
  expect_equal(out$json$b[[1]], 2)
})

test_that("decimal precision is preserved (encode = 'json')", {
  out <- round_trip(
    body = list(a = 0.8675309867530986753098675309),
    encode = "json"
  )
  expect_equal(out$data, "{\"a\":0.867530986753099}")
})

test_that("can do own encoding", {
  out <- round_trip(body = '{"a":1,"b":2}', content_type_json(), encode = "raw")
  expect_equal(out$json, list(a = 1, b = 2))
})

test_that("NULL elements are automatically dropped", {
  out <- round_trip(body = list(x = 1, y = NULL), encode = "form")
  expect_equal(out$form, list(x = "1"))

  out <- round_trip(body = list(x = 1, y = NULL), encode = "multipart")
  expect_equal(out$form, list(x = "1"))
})

test_that("file and form vals mixed give form and data elements", {
  out <- round_trip(body = list(y = data_path, a = 1))
  expect_equal(out$form$a, "1")
  expect_equal(strsplit(out$files$y, "\r?\n")[[1]], data)
})

test_that("single file matches contents on disk", {
  out <- round_trip(body = data_path)
  expect_equal(strsplit(out$data, "\r?\n")[[1]], data)
  expect_equal(out$headers$`Content-Type`, "text/plain")
})

test_that("explicit content type overrides defaults", {
  out <- round_trip(
    body = jsonlite::toJSON(list(a = 1, b = 2)),
    content_type_json()
  )

  expect_equal(out$headers$`Content-Type`, "application/json")
})

#' VERB a url.
#'
#' Use an arbitrary verb.
#'
#' @inherit GET params return
#' @inheritParams POST
#' @param verb Name of verb to use.
#' @family http methods
#' @export
#' @examples
#' r <- VERB(
#'   "PROPFIND", "http://svn.r-project.org/R/tags/",
#'   add_headers(depth = 1), verbose()
#' )
#' stop_for_status(r)
#' content(r)
#'
#' \dontrun{
#' VERB("POST", url = "http://httpbin.org/post")
#' VERB("POST", url = "http://httpbin.org/post", body = "foobar")
#' }
VERB <- function(verb, url = NULL, config = list(), ...,
                 body = NULL, encode = c("multipart", "form", "json", "raw"),
                 handle = NULL) {
  hu <- handle_url(handle, url, ...)
  req <- request_build(verb, hu$url, body_config(body, match.arg(encode)), config, ...)
  request_perform(req, hu$handle$handle)
}

#' Get and set environment variables.
#'
#' This is a bad idea because it will leave it on the history
#'
#' @param Name of environment variable.
#' @param Value of environment variable. Use `NA` to unset.
#' @param Scope of change. If "session", the change is ephemeral and will
#'   disappear when you restart R. If "user", modifies your `~/.Renviron`
#'   so that it affects every project. If "site", modifies the site .Renviron
#'   so that it affects every user.
#' @return Invisibly, the previous value of the environment variable.
#' @noRd
#' @examples
#' \dontrun{
#' # Set locally
#' set_envvar("HTTR", "true")
#' 
#' # Set for every new session (and this session)
#' set_envvar("HTTR", "false", "user")
#' # Update existing value
#' set_envvar("HTTR", "true", "user")
#' }
set_envvar <- function(name, value, scope = c("session", "user", "site")) {
  stopifnot(length(name) == 1, length(value) == 1)
  scope <- match.arg(scope)

  old <- get_envvar(value)

  path <- switch(scope,
    session = NULL,
    user = Sys.getenv("R_ENVIRON_USER", "~/.Renviron"),
    site = Sys.getenv("R_ENVIRON", file.path(R.home("etc"), "Renviron.site"))
  )
  set_envvar_local(name, value)
  if (scope != "session") {
    set_envvar_renviron(name, value, path)
  }

  invisible(old)
}


#' @rdname set_envvar
#' @noRd
get_envvar <- function(name) {
  stopifnot(is.character(name))
  Sys.getenv(name, unset = NA_character_)
}

set_envvar_local <- function(name, value) {
  if (is.na(value)) {
    Sys.unsetenv(name)
  } else {
    l <- stats::setNames(list(quote(value)), name)
    do.call("Sys.setenv", l)
  }
}

# @return \code{TRUE} if an existing env var was modified, \code{FALSE}
#   otherwise (invisibly).
set_envvar_renviron <- function(name, value, path) {
  ev <- paste0(name, '="', value, '"')
  if (is.na(value)) value <- ""

  if (!file.exists(path)) {
    # Create if it doesn't already exist
    cat(ev, "\n", sep = "", file = path)
    return(invisible(FALSE))
  }

  lines <- tryCatch(readLines(path), error = function(e) {
    stop("Failed to read ", path, " with error:\n", e$message, call. = FALSE)
  })
  re <- paste0("^", name, "=")
  matches <- which(grepl(re, lines))

  # No matches, so append to end of file
  if (length(matches) == 0) {
    message("Setting ", name, " in ", path)
    cat(ev, "\n", sep = "", file = path, append = TRUE)
    return(invisible(FALSE))
  }

  message("Updating ", name, " in ", path)
  if (length(matches) == 1) {
    lines[matches] <- ev
  } else {
    lines[matches[1]] <- ev
    lines <- lines[-matches[-1]]
  }

  lines <- tryCatch(writeLines(lines, path), error = function(e) {
    stop("Failed to write ", path, " with error:\n", e$message, call. = FALSE)
  })

  invisible(TRUE)
}

can_write <- function(x) file.access(x, 6)

#' Create a vector with case insensitive name matching.
#'
#' @param x vector to modify
#' @export
#' @keywords internal
#' @examples
#' x <- c("abc" = 1, "def" = 2)
#' x["ABC"]
#' y <- insensitive(x)
#' y["ABC"]
#' y[["ABC"]]
insensitive <- function(x) {
  names(x) <- tolower(names(x))
  structure(x, class = c("insensitive", class(x)))
}

#' @export
`[.insensitive` <- function(x, i, ...) {
  if (is.character(i)) {
    i <- tolower(i)
  }

  NextMethod()
}

#' @export
`[[.insensitive` <- `[.insensitive`


#' @export
"$.insensitive" <- function(x, name) {
  name <- tolower(name)
  x[[name]]
}

#' Get url HEADers.
#'
#' @section RFC2616:
#' The HEAD method is identical to GET except that the server MUST NOT return
#' a message-body in the response. The metainformation contained in the HTTP
#' headers in response to a HEAD request SHOULD be identical to the
#' information sent in response to a GET request. This method can be used for
#' obtaining metainformation about the entity implied by the request without
#' transferring the entity-body itself. This method is often used for testing
#' hypertext links for validity, accessibility, and recent modification.
#'
#' The response to a HEAD request MAY be cacheable in the sense that the
#' information contained in the response MAY be used to update a previously
#' cached entity from that resource. If the new field values indicate that the
#' cached entity differs from the current entity (as would be indicated by a
#' change in Content-Length, Content-MD5, ETag or Last-Modified), then the
#' cache MUST treat the cache entry as stale.
#'
#' @inherit GET params return
#' @family http methods
#' @export
#' @examples
#' HEAD("http://google.com")
#' headers(HEAD("http://google.com"))
HEAD <- function(url = NULL, config = list(), ..., handle = NULL) {
  hu <- handle_url(handle, url, ...)
  req <- request_build("HEAD", hu$url, config, ...)
  request_perform(req, hu$handle$handle)
}

#' Retrieve OAuth 1.0 access token.
#'
#' See demos for use.
#'
#' @param endpoint An OAuth endpoint, created by [oauth_endpoint()]
#' @param app An OAuth consumer application, created by
#'    [oauth_app()]
#' @param permission optional, a string of permissions to ask for.
#' @param is_interactive DEPRECATED
#' @param private_key Optional, a key provided by [openssl::read_key()].
#'   Used for signed OAuth 1.0.
#' @export
#' @keywords internal
init_oauth1.0 <- function(endpoint, app, permission = NULL,
                          is_interactive = interactive(),
                          private_key = NULL) {
  oauth_sig <- function(url, method, token = NULL, token_secret = NULL, private_key = NULL, ...) {
    oauth_header(oauth_signature(url, method, app, token, token_secret, private_key,
      other_params = c(list(...), oauth_callback = oauth_callback())
    ))
  }

  # 1. Get an unauthorized request token
  response <- POST(endpoint$request, oauth_sig(endpoint$request, "POST", private_key = private_key))
  stop_for_status(response)
  params <- content(response, type = "application/x-www-form-urlencoded")
  token <- params$oauth_token
  secret <- params$oauth_token_secret

  # 2. Authorize the token
  authorize_url <- modify_url(endpoint$authorize, query = list(
    oauth_token = token,
    permission = "read"
  ))
  verifier <- oauth_listener(authorize_url, is_interactive)
  verifier <- verifier$oauth_verifier %||% verifier[[1]]

  # 3. Request access token
  response <- POST(endpoint$access,
    oauth_sig(endpoint$access, "POST", token, secret, oauth_verifier = verifier, private_key = private_key),
    body = ""
  )
  stop_for_status(response)
  content(response, type = "application/x-www-form-urlencoded")
}

#' Retrieve OAuth 2.0 access token.
#'
#' See demos for use.
#'
#' @inheritParams init_oauth1.0
#' @param scope a character vector of scopes to request.
#' @param user_params Named list holding endpoint specific parameters to pass to
#'   the server when posting the request for obtaining or refreshing the
#'   access token.
#' @param type content type used to override incorrect server response
#' @param use_oob if FALSE, use a local webserver for the OAuth dance.
#'   Otherwise, provide a URL to the user and prompt for a validation
#'   code. Defaults to the of the `"httr_oob_default"` default,
#'   or `TRUE` if `httpuv` is not installed.
#' @param oob_value if provided, specifies the value to use for the redirect_uri
#'   parameter when retrieving an authorization URL. Defaults to "urn:ietf:wg:oauth:2.0:oob".
#'   Requires `use_oob = TRUE`.
#' @param use_basic_auth if `TRUE` use http basic authentication to
#'   retrieve the token. Some authorization servers require this.
#'   If `FALSE`, the default, retrieve the token by including the
#'   app key and secret in the request body.
#' @param config_init Additional configuration settings sent to
#'   [POST()], e.g. [user_agent()].
#' @param client_credentials Default to `FALSE`. Set to `TRUE` to use
#'   *Client Credentials Grant* instead of *Authorization
#'   Code Grant*. See <https://www.rfc-editor.org/rfc/rfc6749#section-4.4>.
#' @param query_authorize_extra Default to `list()`. Set to named list
#'   holding query parameters to append to initial auth page query. Useful for
#'   some APIs.
#' @export
#' @keywords internal
init_oauth2.0 <- function(endpoint, app, scope = NULL,
                          user_params = NULL,
                          type = NULL,
                          use_oob = getOption("httr_oob_default"),
                          oob_value = NULL,
                          is_interactive = interactive(),
                          use_basic_auth = FALSE,
                          config_init = list(),
                          client_credentials = FALSE,
                          query_authorize_extra = list()) {
  scope <- check_scope(scope)
  use_oob <- check_oob(use_oob, oob_value)
  if (use_oob) {
    redirect_uri <- if (!is.null(oob_value)) oob_value else "urn:ietf:wg:oauth:2.0:oob"
    state <- NULL
  } else {
    redirect_uri <- app$redirect_uri
    state <- nonce()
  }

  # Some Oauth2 grant type not required an authorization request and code
  # (see https://tools.ietf.org/html/rfc6749#section-4.4)
  if (client_credentials) {
    code <- NULL
  } else {
    authorize_url <- oauth2.0_authorize_url(
      endpoint,
      app,
      scope = scope,
      redirect_uri = redirect_uri,
      state = state,
      query_extra = query_authorize_extra
    )
    code <- oauth_authorize(authorize_url, use_oob)
  }

  # Use authorisation code to get (temporary) access token
  oauth2.0_access_token(
    endpoint,
    app,
    code = code,
    user_params = user_params,
    type = type,
    redirect_uri = redirect_uri,
    client_credentials = client_credentials,
    use_basic_auth = use_basic_auth,
    config = config_init
  )
}

#' @export
#' @importFrom utils modifyList
#' @rdname init_oauth2.0
#' @param query_extra See `query_authorize_extra`
oauth2.0_authorize_url <- function(endpoint, app, scope,
                                   redirect_uri = app$redirect_uri,
                                   state = nonce(),
                                   query_extra = list()) {
  # TODO might need to put some params before and some after...

  query_extra <- query_extra %||% list() # i.e. make list if query_extra is null

  default_query <- list(
    client_id = app$key,
    scope = scope,
    redirect_uri = redirect_uri,
    response_type = "code",
    state = state
  )

  query <- compact(modifyList(default_query, query_extra))

  modify_url(
    endpoint$authorize,
    query = query
  )
}

#' @export
#' @rdname init_oauth2.0
oauth2.0_access_token <- function(endpoint,
                                  app,
                                  code,
                                  user_params = NULL,
                                  type = NULL,
                                  use_basic_auth = FALSE,
                                  redirect_uri = app$redirect_uri,
                                  client_credentials = FALSE,
                                  config = list()) {
  req_params <- compact(list(
    client_id = app$key,
    redirect_uri = if (client_credentials) NULL else redirect_uri,
    grant_type = if (client_credentials) "client_credentials" else "authorization_code",
    code = code
  ))

  if (!is.null(user_params)) {
    req_params <- utils::modifyList(user_params, req_params)
  }

  # Send credentials using HTTP Basic or as parameters in the request body
  # See https://tools.ietf.org/html/rfc6749#section-2.3 (Client Authentication)
  if (isTRUE(use_basic_auth)) {
    req <- POST(endpoint$access,
      encode = "form",
      body = req_params,
      authenticate(app$key, app$secret, type = "basic"),
      config = config
    )
  } else {
    req_params$client_secret <- app$secret
    req <- POST(endpoint$access,
      encode = "form",
      body = req_params,
      config = config
    )
  }

  stop_for_status(req, task = "get an access token")
  content(req, type = type)
}


# Parameter checking ------------------------------------------------------

check_scope <- function(x) {
  if (is.null(x)) {
    return(NULL)
  }

  if (!is.character(x)) {
    stop("`scope` must be a character vector", call. = FALSE)
  }
  paste(x, collapse = " ")
}

# gargle needs to access (pseudo-)OOB flow from Google Colab, so we need to
# be able to use the "rlang_interactive" option to signal that we are in
# an environment that is interactive (enough) to complete this flow.
is_interactive <- function() {
  getOption("rlang_interactive") %||% interactive()
}

check_oob <- function(use_oob, oob_value = NULL) {
  if (!is.logical(use_oob) || length(use_oob) != 1) {
    stop("`use_oob` must be a length-1 logical vector", call. = FALSE)
  }

  if (!use_oob && !is_installed("httpuv")) {
    message("httpuv not installed, defaulting to out-of-band authentication")
    use_oob <- TRUE
  }

  if (use_oob) {
    if (!is_interactive()) {
      stop(
        "Can only use oob authentication in an interactive session",
        call. = FALSE
      )
    }
  }

  if (!is.null(oob_value)) {
    if (!use_oob) {
      stop(
        "Can only use custom oob value if use_oob is enabled",
        call. = FALSE
      )
    }
  }

  use_oob
}


oauth_authorize <- function(url, oob = FALSE) {
  if (oob) {
    oauth_exchanger(url)$code
  } else {
    oauth_listener(url)$code
  }
}

context("OAuth")

test_that("oauth2.0 signing works", {
  skip_httpbin()
  request_url <- "http://httpbin.org/headers"

  token <- Token2.0$new(
    app = oauth_app("x", "y", "z"),
    endpoint = oauth_endpoints("google"),
    credentials = list(access_token = "ofNoArms")
  )

  token$params$as_header <- TRUE
  header_response <- GET(request_url, config(token = token))
  response_content <- content(header_response)$headers
  expect_equal("Bearer ofNoArms", response_content$Authorization)
  expect_equal(request_url, header_response$url)

  token$params$as_header <- FALSE
  url_response <- GET(request_url, config(token = token))
  response_content <- content(url_response)$headers
  expect_equal(NULL, response_content$Authorization)
  expect_equal(
    parse_url(url_response$url)$query,
    list(access_token = "ofNoArms")
  )
})

test_that("partial OAuth1 flow works", {
  skip_on_cran()
  # From rfigshare

  endpoint <- oauth_endpoint(
    base_url = "http://api.figshare.com/v1/pbl/oauth",
    "request_token", "authorize", "access_token"
  )
  myapp <- oauth_app("rfigshare",
    key = "Kazwg91wCdBB9ggypFVVJg",
    secret = "izgO06p1ymfgZTsdsZQbcA"
  )
  sig <- sign_oauth1.0(myapp,
    token = "xdBjcKOiunwjiovwkfTF2QjGhROeLMw0y0nSCSgvg3YQxdBjcKOiunwjiovwkfTF2Q",
    token_secret = "4mdM3pfekNGO16X4hsvZdg"
  )

  r <- GET("http://api.figshare.com/v1/my_data/articles", sig)
  expect_equal(status_code(r), 200)
})

test_that("oauth_encode1 works", {
  # chinese characters for "xaringan"
  orig_string <- "\u5199\u8f6e\u773c"
  restored_string <- URLdecode(oauth_encode1(orig_string))
  Encoding(restored_string) <- "UTF-8"

  expect_equal(orig_string, restored_string)
})

test_that("oauth2.0 authorize url appends query params", {
  app <- oauth_app("x", "y", "z")
  scope <- NULL
  query_extra <- list(
    foo = "bar"
  )
  authURL <- oauth2.0_authorize_url(
    endpoint = oauth_endpoints("google"),
    app = app,
    scope = scope,
    query_extra = query_extra
  )

  url <- parse_url(authURL)
  expect_equal(url$query$foo, "bar")
})

test_that("oauth2.0 authorize url handles empty query_extra input", {
  # common constructor
  authorize_url_extra_params <- function(extra_params) {
    app <- oauth_app("x", "y", "z")
    scope <- NULL

    url_with_empty_input <- oauth2.0_authorize_url(
      endpoint = oauth_endpoints("google"),
      app = app,
      scope = scope,
      state = "testing-nonce",
      query_extra = extra_params
    )
    parse_url(url_with_empty_input)$query
  }

  # expect NA (i.e. no) error
  expect_error(authorize_url_extra_params(list()), NA) # with empty list
  expect_error(authorize_url_extra_params(NULL), NA) # with NULL list
})

# Parameter checking ------------------------------------------------------

test_that("scope must be character or NULL", {
  expect_equal(check_scope("a"), "a")
  expect_equal(check_scope(NULL), NULL)
  expect_error(check_scope(1), "character vector")
})

test_that("multiple scopes collapsed with space", {
  expect_equal(check_scope(c("a", "b")), "a b")
})

test_that("oob must be a flag", {
  expect_error(check_oob("a"), "logical vector")
  expect_error(check_oob(c(FALSE, FALSE)), "logical vector")
})

test_that("can not use oob in non-interactive session", {
  old <- options()
  on.exit(options(old))
  options(rlang_interactive = FALSE)

  expect_error(check_oob(TRUE), "interactive")
})

test_that("can not use custom oob value without enabling oob", {
  testthat::skip_if_not_installed("httpuv")

  old <- options()
  on.exit(options(old))
  options(rlang_interactive = TRUE)

  expect_error(check_oob(FALSE, "custom_value"), "custom oob value")
})

context("OAuth")

test_that("oauth2.0 signing works", {
  request_url <- "http://httpbin.org/headers"

  token <- Token2.0$new(
    app = oauth_app("x", "y", "z"),
    endpoint = oauth_endpoints("google"),
    credentials = list(access_token = "ofNoArms")
  )

  token$params$as_header <- TRUE
  header_response <- GET(request_url, config(token = token))
  response_content <- content(header_response)$headers
  expect_equal("Bearer ofNoArms", response_content$Authorization)
  expect_equal(request_url, header_response$url)

  token$params$as_header <- FALSE
  url_response <- GET(request_url, config(token = token))
  response_content <- content(url_response)$headers
  expect_equal(NULL, response_content$Authorization)
  expect_equal(
    parse_url(url_response$url)$query,
    list(access_token = "ofNoArms")
  )
})

test_that("partial OAuth1 flow works", {
  skip_on_cran()
  # From rfigshare

  endpoint <- oauth_endpoint(
    base_url = "http://api.figshare.com/v1/pbl/oauth",
    "request_token", "authorize", "access_token"
  )
  myapp <- oauth_app("rfigshare",
    key = "Kazwg91wCdBB9ggypFVVJg",
    secret = "izgO06p1ymfgZTsdsZQbcA"
  )
  sig <- sign_oauth1.0(myapp,
    token = "xdBjcKOiunwjiovwkfTF2QjGhROeLMw0y0nSCSgvg3YQxdBjcKOiunwjiovwkfTF2Q",
    token_secret = "4mdM3pfekNGO16X4hsvZdg"
  )

  r <- GET("http://api.figshare.com/v1/my_data/articles", sig)
  expect_equal(status_code(r), 200)
})

test_that("oauth_encode1 works", {
  # chinese characters for "xaringan"
  orig_string <- "\u5199\u8f6e\u773c"
  restored_string <- URLdecode(oauth_encode1(orig_string))
  Encoding(restored_string) <- "UTF-8"

  expect_equal(orig_string, restored_string)
})

test_that("oauth2.0 authorize url appends query params", {
  app <- oauth_app("x", "y", "z")
  scope <- NULL
  query_extra <- list(
    foo = "bar"
  )
  authURL <- oauth2.0_authorize_url(
    endpoint = oauth_endpoints("google"),
    app = app,
    scope = scope,
    query_extra = query_extra
  )

  url <- parse_url(authURL)
  expect_equal(url$query$foo, "bar")
})

test_that("oauth2.0 authorize url handles empty query_extra input", {
  # common constructor
  authorize_url_extra_params <- function(extra_params) {
    app <- oauth_app("x", "y", "z")
    scope <- NULL

    url_with_empty_input <- oauth2.0_authorize_url(
      endpoint = oauth_endpoints("google"),
      app = app,
      scope = scope,
      state = "testing-nonce",
      query_extra = extra_params
    )
    parse_url(url_with_empty_input)$query
  }

  # expect NA (i.e. no) error
  expect_error(authorize_url_extra_params(list()), NA) # with empty list
  expect_error(authorize_url_extra_params(NULL), NA) # with NULL list
})

# Parameter checking ------------------------------------------------------

test_that("scope must be character or NULL", {
  expect_equal(check_scope("a"), "a")
  expect_equal(check_scope(NULL), NULL)
  expect_error(check_scope(1), "character vector")
})

test_that("multiple scopes collapsed with space", {
  expect_equal(check_scope(c("a", "b")), "a b")
})

test_that("oob must be a flag", {
  expect_error(check_oob("a"), "logical vector")
  expect_error(check_oob(c(FALSE, FALSE)), "logical vector")
})

test_that("can not use oob in non-interactive session", {
  expect_error(check_oob(TRUE), "interactive")
})

test_that("can not use custom oob value without enabling oob", {
  testthat::skip_if_not_installed("httpuv")
  with_mock(
    `httr:::is_interactive` = function() TRUE,
    expect_error(check_oob(FALSE, "custom_value"), "custom oob value")
  )
})

#' Set maximum request time.
#'
#' @param seconds number of seconds to wait for a response until giving up.
#'   Can not be less than 1 ms.
#' @family config
#' @details This timeout is passed on to [curl::handle_setopt()].
#'   See there and [curl::curl_options()] for more details.
#' @export
#' @examples
#' \dontrun{
#' GET("http://httpbin.org/delay/3", timeout(1))
#' GET("http://httpbin.org/delay/1", timeout(2))
#' }
timeout <- function(seconds) {
  if (seconds < 0.001) {
    stop("Timeout cannot be less than 1 ms", call. = FALSE)
  }

  config(timeout_ms = seconds * 1000)
}

#' OAuth token objects.
#'
#' These objects represent the complete set of data needed for OAuth access:
#' an app, an endpoint, cached credentials and parameters. They should be
#' created through their constructor functions [oauth1.0_token()]
#' and [oauth2.0_token()].
#'
#' @section Methods:
#' * `cache()`: caches token to disk
#' * `sign(method, url)`: returns list of url and config
#' * `refresh()`: refresh access token (if possible)
#' * `validate()`: TRUE if the token is still valid, FALSE otherwise
#'
#' @section Caching:
#' OAuth tokens are cached on disk in a file called `.httr-oauth`
#' saved in the current working directory.  Caching is enabled if:
#'
#' * The session is interactive, and the user agrees to it, OR
#' * The `.httr-oauth` file is already present, OR
#' * `getOption("httr_oauth_cache")` is `TRUE`
#'
#' You can suppress caching by setting the `httr_oauth_cache` option to
#' `FALSE`.
#'
#' Tokens are cached based on their endpoint and parameters.
#'
#' The cache file should not be included in source code control or R packages
#' (because it contains private information), so httr will automatically add
#' the appropriate entries to `.gitignore` and `.Rbuildignore` if needed.
#'
#' @docType class
#' @keywords internal
#' @format An R6 class object.
#' @importFrom R6 R6Class
#' @export
#' @name Token-class
Token <- R6::R6Class("Token", list(
  endpoint = NULL,
  app = NULL,
  credentials = NULL,
  params = NULL,
  cache_path = FALSE,
  private_key = NULL,

  initialize = function(app, endpoint, params = list(), credentials = NULL,
                        private_key = NULL,
                        cache_path = getOption("httr_oauth_cache")) {
    stopifnot(
      is.oauth_endpoint(endpoint) || !is.null(credentials),
      is.oauth_app(app),
      is.list(params)
    )

    self$app <- app
    self$endpoint <- endpoint
    self$params <- params
    self$cache_path <- use_cache(cache_path)
    self$private_key <- private_key

    if (!is.null(credentials)) {
      # Use credentials created elsewhere - usually for tests
      self$credentials <- credentials
      return(self)
    }

    # Are credentials cache already?
    if (self$load_from_cache()) {
      self
    } else {
      self$init_credentials()
      self$cache()
    }
  },

  print = function(...) {
    cat("<Token>\n", sep = "")
    print(self$endpoint)
    print(self$app)
    cat("<credentials> ", paste0(names(self$credentials), collapse = ", "), "\n",
      sep = ""
    )
    cat("---\n")
  },
  cache = function(path = self$cache_path) {
    cache_token(self, path)
    self
  },
  load_from_cache = function() {
    if (is.null(self$cache_path)) return(FALSE)

    cached <- fetch_cached_token(self$hash(), self$cache_path)
    if (is.null(cached)) return(FALSE)

    self$endpoint <- cached$endpoint
    self$app <- cached$app
    self$credentials <- cached$credentials
    self$params <- cached$params
    TRUE
  },
  hash = function() {
    # endpoint = which site
    # app = client identification
    # params = scope
    msg <- serialize(
      list(self$endpoint, self$app, normalize_scopes(self$params$scope)),
      NULL
    )

    # for compatibility with digest::digest()
    paste(openssl::md5(msg[-(1:14)]), collapse = "")
  },
  sign = function() {
    stop("Must be implemented by subclass", call. = FALSE)
  },
  refresh = function() {
    stop("Must be implemented by subclass", call. = FALSE)
  },
  init_credentials = function() {
    stop("Must be implemented by subclass", call. = FALSE)
  }
))

#' Generate an oauth1.0 token.
#'
#' This is the final object in the OAuth dance - it encapsulates the app,
#' the endpoint, other parameters and the received credentials.
#'
#' See [Token()] for full details about the token object, and the
#' caching policies used to store credentials across sessions.
#'
#' @inheritParams init_oauth1.0
#' @param as_header If `TRUE`, the default, sends oauth in header.
#'   If `FALSE`, adds as parameter to url.
#' @param cache A logical value or a string. `TRUE` means to cache
#'   using the default cache file `.httr-oauth`, `FALSE` means
#'   don't cache, and `NA` means to guess using some sensible heuristics.
#'   A string means use the specified path as the cache file.
#' @return A `Token1.0` reference class (RC) object.
#' @family OAuth
#' @export
oauth1.0_token <- function(endpoint, app, permission = NULL,
                           as_header = TRUE,
                           private_key = NULL,
                           cache = getOption("httr_oauth_cache")) {
  params <- list(
    permission = permission,
    as_header = as_header
  )

  Token1.0$new(
    app = app,
    endpoint = endpoint,
    params = params,
    private_key = private_key,
    cache_path = cache
  )
}

#' @export
#' @rdname Token-class
Token1.0 <- R6::R6Class("Token1.0", inherit = Token, list(
  init_credentials = function(force = FALSE) {
    self$credentials <- init_oauth1.0(
      self$endpoint, self$app,
      permission = self$params$permission,
      private_key = self$private_key
    )
  },
  can_refresh = function() {
    FALSE
  },
  refresh = function() {
    stop("Not implemented")
  },
  sign = function(method, url) {
    oauth <- oauth_signature(
      url, method, self$app, self$credentials$oauth_token,
      self$credentials$oauth_token_secret, self$private_key
    )
    if (isTRUE(self$params$as_header)) {
      c(request(url = url), oauth_header(oauth))
    } else {
      url <- parse_url(url)
      url$query <- c(url$query, oauth)
      request(url = build_url(url))
    }
  }
))

#' Generate an oauth2.0 token.
#'
#' This is the final object in the OAuth dance - it encapsulates the app,
#' the endpoint, other parameters and the received credentials. It is a
#' reference class so that it can be seamlessly updated (e.g. using
#' `$refresh()`) when access expires.
#'
#' See [Token()] for full details about the token object, and the
#' caching policies used to store credentials across sessions.
#'
#' @inheritParams init_oauth2.0
#' @param as_header If `TRUE`, the default, configures the token to add
#'   itself to the bearer header of subsequent requests. If `FALSE`,
#'   configures the token to add itself as a url parameter of subsequent
#'   requests.
#' @param credentials Advanced use only: allows you to completely customise
#'   token generation.
#' @inheritParams oauth1.0_token
#' @return A `Token2.0` reference class (RC) object.
#' @family OAuth
#' @export
oauth2.0_token <- function(endpoint, app, scope = NULL, user_params = NULL,
                           type = NULL,
                           use_oob = getOption("httr_oob_default"),
                           oob_value = NULL,
                           as_header = TRUE,
                           use_basic_auth = FALSE,
                           cache = getOption("httr_oauth_cache"),
                           config_init = list(),
                           client_credentials = FALSE,
                           credentials = NULL,
                           query_authorize_extra = list()) {
  params <- list(
    scope = scope,
    user_params = user_params,
    type = type,
    use_oob = use_oob,
    oob_value = oob_value,
    as_header = as_header,
    use_basic_auth = use_basic_auth,
    config_init = config_init,
    client_credentials = client_credentials,
    query_authorize_extra = query_authorize_extra
  )

  Token2.0$new(
    app = app,
    endpoint = endpoint,
    params = params,
    credentials = credentials,
    cache_path = if (is.null(credentials)) cache else FALSE
  )
}

#' @export
#' @rdname Token-class
Token2.0 <- R6::R6Class("Token2.0", inherit = Token, list(
  init_credentials = function() {
    self$credentials <- init_oauth2.0(self$endpoint, self$app,
      scope = self$params$scope, user_params = self$params$user_params,
      type = self$params$type, use_oob = self$params$use_oob,
      oob_value = self$params$oob_value,
      use_basic_auth = self$params$use_basic_auth,
      config_init = self$params$config_init,
      client_credentials = self$params$client_credentials,
      query_authorize_extra = self$params$query_authorize_extra
    )
  },
  can_refresh = function() {
    !is.null(self$credentials$refresh_token)
  },
  refresh = function() {
    cred <- refresh_oauth2.0(
      self$endpoint, self$app,
      self$credentials, self$params$user_params, self$params$use_basic_auth
    )
    if (is.null(cred)) {
      remove_cached_token(self)
    } else {
      self$credentials <- cred
      self$cache()
    }
    self
  },
  sign = function(method, url) {
    if (self$params$as_header) {
      request(url = url, headers = c(
        Authorization = paste("Bearer", self$credentials$access_token)
      ))
    } else {
      url <- parse_url(url)
      url$query$access_token <- self$credentials$access_token
      request(url = build_url(url))
    }
  },
  validate = function() {
    validate_oauth2.0(self$endpoint, self$credentials)
  },
  revoke = function() {
    revoke_oauth2.0(self$endpoint, self$credentials)
  }
))


#' Generate OAuth token for service accounts.
#'
#' Service accounts provide a way of using OAuth2 without user intervention.
#' They instead assume that the server has access to a private key used
#' to sign requests. The OAuth app is not needed for service accounts:
#' that information is embedded in the account itself.
#'
#' @inheritParams oauth2.0_token
#' @inheritParams jwt_signature
#' @param secrets Secrets loaded from JSON file, downloaded from console.
#' @family OAuth
#' @export
#' @examples
#' \dontrun{
#' endpoint <- oauth_endpoints("google")
#' secrets <- jsonlite::fromJSON("~/Desktop/httrtest-45693cbfac92.json")
#' scope <- "https://www.googleapis.com/auth/bigquery.readonly"
#'
#' token <- oauth_service_token(endpoint, secrets, scope)
#' }
oauth_service_token <- function(endpoint, secrets, scope = NULL, sub = NULL) {
  if (!is.oauth_endpoint(endpoint)) {
    stop("`endpoint` must be an OAuth endpoint", call. = FALSE)
  }
  if (!is.list(secrets)) {
    stop("`secrets` must be a list.", call. = FALSE)
  }

  scope <- check_scope(scope)

  TokenServiceAccount$new(
    endpoint = endpoint,
    secrets = secrets,
    params = list(scope = scope, sub = sub)
  )
}

#' @export
#' @rdname Token-class
TokenServiceAccount <- R6::R6Class("TokenServiceAccount", inherit = Token2.0, list(
  secrets = NULL,
  initialize = function(endpoint, secrets, params) {
    self$endpoint <- endpoint
    self$secrets <- secrets
    self$params <- params

    self$refresh()
  },
  can_refresh = function() {
    TRUE
  },
  refresh = function() {
    self$credentials <- init_oauth_service_account(
      self$secrets,
      scope = self$params$scope,
      sub = self$params$sub
    )
    self
  },
  sign = function(method, url) {
    config <- add_headers(
      Authorization = paste("Bearer", self$credentials$access_token)
    )
    request_build(method = method, url = url, config)
  },

  # Never cache
  cache = function(path) self,
  load_from_cache = function() self
))

normalize_scopes <- function(x) {
  stats::setNames(sort(unique(x)), NULL)
}

context("OAuth cache")

tmp_dir <- function() {
  x <- tempfile()
  dir.create(x)
  x
}

test_that("use_cache() returns NULL or filepath", {
  old <- options()
  on.exit(options(old))
  owd <- setwd(tmp_dir())
  on.exit(setwd(owd), add = TRUE)

  expect_equal(use_cache(FALSE), NULL)
  expect_equal(use_cache(TRUE), ".httr-oauth")
  expect_equal(use_cache("xyz"), "xyz")
})

test_that("use_cache() respects options", {
  old <- options()
  on.exit(options(old))
  owd <- setwd(tmp_dir())
  on.exit(setwd(owd), add = TRUE)

  options(httr_oauth_cache = FALSE)
  expect_equal(use_cache(), NULL)

  options(httr_oauth_cache = TRUE)
  expect_equal(use_cache(), ".httr-oauth")
})

test_that("token saved to and restored from cache, scopes normalized in hash", {
  owd <- setwd(tmp_dir())
  on.exit(setwd(owd))

  token_a <- Token2.0$new(
    app = oauth_app("x", "y", "z"),
    endpoint = oauth_endpoints("google"),
    params = list(scope = c(a = "a", "b")),
    cache_path = ".tmp-cache",
    credentials = list(a = 1)
  )
  token_a$cache()

  token_b <- Token2.0$new(
    app = oauth_app("x", "y", "z"),
    endpoint = oauth_endpoints("google"),
    params = list(scope = c(b = "b", "a")),
    cache_path = ".tmp-cache"
  )

  expect_true(token_b$load_from_cache())
  expect_equal(token_b$app, token_a$app)
  expect_equal(token_b$endpoint, token_a$endpoint)
  expect_equal(token_b$credentials, token_a$credentials)
})

test_that("new caches are Rbuildignored and gitignored", {
  owd <- setwd(tmp_dir())
  on.exit(setwd(owd))
  file.create("DESCRIPTION")

  ## default: options("httr_oauth_cache" = NA)
  ## not tested
  ## if cache does not exist, will not be created if !interactive()

  old <- options("httr_oauth_cache" = TRUE)
  on.exit(options(old), add = TRUE)
  use_cache()
  expect_true(file.exists(".Rbuildignore"))
  expect_identical(readLines(".Rbuildignore"), "^\\.httr-oauth$")
  expect_true(file.exists(".gitignore"))
  expect_identical(readLines(".gitignore"), ".httr-oauth")

  unlink(c(".httr-oauth", ".Rbuildignore", ".gitignore"))

  use_cache("xyz")
  expect_true(file.exists(".Rbuildignore"))
  expect_identical(readLines(".Rbuildignore"), "^xyz$")
  expect_true(file.exists(".gitignore"))
  expect_identical(readLines(".gitignore"), "xyz")
})

#' Send PATCH request to a server.
#'
#' @inherit GET params return
#' @inheritParams POST
#' @family http methods
#' @export
PATCH <- function(url = NULL, config = list(), ...,
                  body = NULL, encode = c("multipart", "form", "json", "raw"),
                  handle = NULL) {
  encode <- match.arg(encode)

  hu <- handle_url(handle, url, ...)
  req <- request_build("PATCH", hu$url, body_config(body, match.arg(encode)), config, ...)
  request_perform(req, hu$handle$handle)
}

handle_url <- function(handle = NULL, url = NULL, ...) {
  if (is.null(url) && is.null(handle)) {
    stop("Must specify at least one of url or handle")
  }

  if (is.null(handle)) handle <- handle_find(url)
  if (is.null(url)) url <- handle$url

  # If named components in ..., modify the url
  new <- named(list(...))
  if (length(new) > 0 || is.url(url)) {
    old <- parse_url(url)
    url <- build_url(utils::modifyList(old, new))
  }

  list(handle = handle, url = url)
}

callback_env <- new.env(parent = emptyenv())
callback_env$request <- NULL
callback_env$response <- NULL

#' Install or uninstall a callback function
#'
#' Supported callback functions: \describe{
#' \item{\sQuote{request}}{This callback is called before an HTTP request
#'   is performed, with the `request` object as an argument.
#'   If the callback returns a value other than `NULL`, the HTTP
#'   request is not performed at all, and the return value of the callback
#'   is returned. This mechanism can be used to replay previously
#'   recorded HTTP responses.
#' }
#' \item{\sQuote{response}}{This callback is called after an HTTP request
#'   is performed. The callback is called with two arguments: the
#'   `request` object and the `response` object of the HTTP
#'   request. If this callback returns a value other than `NULL`,
#'   then this value is returned by `httr`.}
#' }
#'
#' Note that it is not possible to install multiple callbacks of the same
#' type. The installed callback overwrites the previously intalled one.
#' To uninstall a callback function, set it to `NULL` with
#' `set_callback()`.
#'
#' See the `httrmock` package for a proper example that uses
#' callbacks.
#'
#' @param name Character scalar, name of the callback to query or set.
#' @param new_callback The callback function to install, a function object;
#'   or `NULL` to remove the currently installed callback (if any).
#'
#' @return `get_callback` returns the currently installed
#'   callback, or `NULL` if none is installed.
#'
#'   `set_callback` returns the previously installed callback,
#'   or `NULL` if none was installed.
#'
#' @export
#' @examples
#' \dontrun{
#' ## Log all HTTP requests to the screeen
#' req_logger <- function(req) {
#'   cat("HTTP request to", sQuote(req$url), "\n")
#' }
#' 
#' old <- set_callback("request", req_logger)
#' g1 <- GET("https://httpbin.org")
#' g2 <- GET("https://httpbin.org/ip")
#' set_callback("request", old)
#' 
#' ## Log all HTTP requests and response status codes as well
#' req_logger2 <- function(req) {
#'   cat("HTTP request to", sQuote(req$url), "... ")
#' }
#' res_logger <- function(req, res) {
#'   cat(res$status_code, "\n")
#' }
#' 
#' old_req <- set_callback("request", req_logger2)
#' old_res <- set_callback("response", res_logger)
#' g3 <- GET("https://httpbin.org")
#' g4 <- GET("https://httpbin.org/ip")
#' set_callback("request", old_req)
#' set_callback("response", old_res)
#' 
#' ## Return a recorded response, without performing the HTTP request
#' replay <- function(req) {
#'   if (req$url == "https://httpbin.org") g3
#' }
#' old_req <- set_callback("request", replay)
#' grec <- GET("https://httpbin.org")
#' grec$date == g3$date
#' set_callback("request", old_req)
#' }
#' 
get_callback <- function(name) {
  stopifnot(is.character(name), length(name) == 1, !is.na(name))
  if (!name %in% ls(callback_env)) stop("Unknown httr callback: ", name)
  callback_env[[name]]
}

#' @export
#' @rdname get_callback

set_callback <- function(name, new_callback = NULL) {
  stopifnot(is.character(name), length(name) == 1, !is.na(name))
  if (!name %in% ls(callback_env)) stop("Unknown httr callback: ", name)

  old <- callback_env[[name]]
  stopifnot(is.null(new_callback) || is.function(new_callback))
  callback_env[[name]] <- new_callback
  invisible(old)
}

perform_callback <- function(name, ...) {
  if (!is.null(callback <- get_callback(name))) {
    callback(...)
  }
}

context("Callback")

test_that("request callback", {
  f <- function(req) req$url
  old <- set_callback("request", f)
  on.exit(set_callback("request", old))

  expect_identical(GET("foo.bar"), "foo.bar")

  expect_identical(f, get_callback("request"))
  expect_identical(f, set_callback("request", old))
  expect_identical(old, get_callback("request"))
})

#' Extract status code from response.
#'
#' @param x A response
#' @export
status_code <- function(x) UseMethod("status_code")
#' @export
status_code.response <- function(x) x$status_code
#' @export
status_code.numeric <- function(x) x

#' Give information on the status of a request.
#'
#' Extract the http status code and convert it into a human readable message.
#'
#' http servers send a status code with the response to each request. This code
#' gives information regarding the outcome of the execution of the request
#' on the server. Roughly speaking, codes in the 100s and 200s mean the request
#' was successfully executed; codes in the 300s mean the page was redirected;
#' codes in the 400s mean there was a mistake in the way the client sent the
#' request; codes in the 500s mean the server failed to fulfill
#' an apparently valid request. More details on the codes can be found at
#' `http://en.wikipedia.org/wiki/Http_error_codes`.
#'
#' @param x a request object or a number.
#' @return If the status code does not match a known status, an error.
#'  Otherwise, a list with components
#'  \item{category}{the broad category of the status}
#'  \item{message}{the meaning of the status code}
#' @family response methods
#' @examples
#' http_status(100)
#' http_status(404)
#'
#' \dontrun{
#' x <- GET("http://httpbin.org/status/200")
#' http_status(x)
#'
#' http_status(GET("http://httpbin.org/status/300"))
#' http_status(GET("http://httpbin.org/status/301"))
#' http_status(GET("http://httpbin.org/status/404"))
#'
#' # errors out on unknown status
#' http_status(GET("http://httpbin.org/status/320"))
#' }
#' @export
http_status <- function(x) {
  status <- status_code(x)

  if (!status %in% names(http_statuses)) {
    stop("Unknown http status code: ", status, call. = FALSE)
  }

  status_desc <- http_statuses[[as.character(status)]]

  status_types <- c(
    "Information", "Success", "Redirection", "Client error",
    "Server error"
  )
  status_type <- status_types[[status %/% 100]]

  # create the final information message
  message <- paste(status_type, ": (", status, ") ", status_desc, sep = "")

  list(
    category = status_type,
    reason = status_desc,
    message = message
  )
}

http_statuses <- c(
  "100" = "Continue",
  "101" = "Switching Protocols",
  "102" = "Processing (WebDAV; RFC 2518)",
  "200" = "OK",
  "201" = "Created",
  "202" = "Accepted",
  "203" = "Non-Authoritative Information",
  "204" = "No Content",
  "205" = "Reset Content",
  "206" = "Partial Content",
  "207" = "Multi-Status (WebDAV; RFC 4918)",
  "208" = "Already Reported (WebDAV; RFC 5842)",
  "226" = "IM Used (RFC 3229)",
  "300" = "Multiple Choices",
  "301" = "Moved Permanently",
  "302" = "Found",
  "303" = "See Other",
  "304" = "Not Modified",
  "305" = "Use Proxy",
  "306" = "Switch Proxy",
  "307" = "Temporary Redirect",
  "308" = "Permanent Redirect (experimental Internet-Draft)",
  "400" = "Bad Request",
  "401" = "Unauthorized",
  "402" = "Payment Required",
  "403" = "Forbidden",
  "404" = "Not Found",
  "405" = "Method Not Allowed",
  "406" = "Not Acceptable",
  "407" = "Proxy Authentication Required",
  "408" = "Request Timeout",
  "409" = "Conflict",
  "410" = "Gone",
  "411" = "Length Required",
  "412" = "Precondition Failed",
  "413" = "Request Entity Too Large",
  "414" = "Request-URI Too Long",
  "415" = "Unsupported Media Type",
  "416" = "Requested Range Not Satisfiable",
  "417" = "Expectation Failed",
  "418" = "I'm a teapot (RFC 2324)",
  "420" = "Enhance Your Calm (Twitter)",
  "422" = "Unprocessable Entity (WebDAV; RFC 4918)",
  "423" = "Locked (WebDAV; RFC 4918)",
  "424" = "Failed Dependency (WebDAV; RFC 4918)",
  "424" = "Method Failure (WebDAV)",
  "425" = "Unordered Collection (Internet draft)",
  "426" = "Upgrade Required (RFC 2817)",
  "428" = "Precondition Required (RFC 6585)",
  "429" = "Too Many Requests (RFC 6585)",
  "431" = "Request Header Fields Too Large (RFC 6585)",
  "444" = "No Response (Nginx)",
  "449" = "Retry With (Microsoft)",
  "450" = "Blocked by Windows Parental Controls (Microsoft)",
  "451" = "Unavailable For Legal Reasons (Internet draft)",
  "499" = "Client Closed Request (Nginx)",
  "500" = "Internal Server Error",
  "501" = "Not Implemented",
  "502" = "Bad Gateway",
  "503" = "Service Unavailable",
  "504" = "Gateway Timeout",
  "505" = "HTTP Version Not Supported",
  "506" = "Variant Also Negotiates (RFC 2295)",
  "507" = "Insufficient Storage (WebDAV; RFC 4918)",
  "508" = "Loop Detected (WebDAV; RFC 5842)",
  "509" = "Bandwidth Limit Exceeded (Apache bw/limited extension)",
  "510" = "Not Extended (RFC 2774)",
  "511" = "Network Authentication Required (RFC 6585)",
  "598" = "Network read timeout error (Unknown)",
  "599" = "Network connect timeout error (Unknown)"
)

#' Check for an http error.
#'
#' @param x Object to check. Default methods are provided for strings
#'   (which perform an [HEAD()] request), responses, and
#'   integer status codes.
#' @param ... Other arguments passed on to methods.
#' @return  `TRUE` if the request fails (status code 400 or above),
#'   otherwise `FALSE`.
#' @export
#' @family response methods
#' @examples
#' \dontrun{
#' # You can pass a url:
#' http_error("http://www.google.com")
#' http_error("http://httpbin.org/status/404")
#'
#' # Or a request
#' r <- GET("http://httpbin.org/status/201")
#' http_error(r)
#' }
#'
#' # Or an (integer) status code
#' http_error(200L)
#' http_error(404L)
http_error <- function(x, ...) {
  UseMethod("http_error")
}
#' @export
http_error.character <- function(x, ...) {
  http_error(HEAD(x, ...))
}
#' @export
http_error.response <- function(x, ...) {
  http_error(status_code(x))
}

#' @export
http_error.integer <- function(x, ...) {
  x >= 400L
}

#' @export
#' @rdname http_error
#' @usage NULL
url_success <- function(x, ...) {
  warning(
    "`url_success(x)` is deprecated; please use `!http_error(x)` instead.",
    call. = FALSE
  )
  !http_error(x, ...)
}

#' @export
#' @rdname http_error
#' @usage NULL
url_ok <- function(x, ...) {
  warning(
    "`url_ok(x)` is deprecated; ",
    "please use `identical(status_code(x), 200L)` instead.",
    call. = FALSE
  )
  identical(status_code(HEAD(x, ...)), 200L)
}

#' Take action on http error.
#'
#' Converts http errors to R errors or warnings - these should always
#' be used whenever you're creating requests inside a function, so
#' that the user knows why a request has failed.
#'
#' @return If request was successful, the response (invisibly). Otherwise,
#'   raised a classed http error or warning, as generated by
#'   [http_condition()]
#' @inheritParams http_condition
#' @seealso [http_status()] and
#'   `http://en.wikipedia.org/wiki/Http_status_codes` for more information
#'   on http status codes.
#' @family response methods
#' @examples
#' \dontrun{
#' x <- GET("http://httpbin.org/status/200")
#' stop_for_status(x) # nothing happens
#' warn_for_status(x)
#' message_for_status(x)
#'
#' x <- GET("http://httpbin.org/status/300")
#' stop_for_status(x)
#' warn_for_status(x)
#' message_for_status(x)
#'
#' x <- GET("http://httpbin.org/status/404")
#' stop_for_status(x)
#' warn_for_status(x)
#' message_for_status(x)
#'
#' # You can provide more information with the task argument
#' warn_for_status(x, "download spreadsheet")
#' message_for_status(x, "download spreadsheet")
#' }
#' @export
stop_for_status <- function(x, task = NULL) {
  if (status_code(x) < 300) {
    return(invisible(x))
  }

  call <- sys.call(-1)
  stop(http_condition(x, "error", task = task, call = call))
}

#' @rdname stop_for_status
#' @export
warn_for_status <- function(x, task = NULL) {
  if (status_code(x) < 300) {
    return(invisible(x))
  }

  call <- sys.call(-1)
  warning(http_condition(x, "warning", task = task, call = call))
}

#' @rdname stop_for_status
#' @export
message_for_status <- function(x, task = NULL) {
  call <- sys.call(-1)
  message(http_condition(x, "message", task = task, call = call))
}


#' Generate a classed http condition.
#'
#' This function generate S3 condition objects which are passed to
#' [stop()] or [warning()] to generate classes warnings
#' and error. These can be used in conjunction with [tryCatch()]
#' to respond differently to different type of failure.
#'
#' @keywords internal
#' @return An S3 object that inherits from (e.g.) condition, `type`,
#'   http_error, http_400 and http_404.
#' @param x a response, or numeric http code (or other object with
#'   `status_code` method)
#' @param type type of condition to generate. Must be one of error,
#'   warning or message.
#' @param task The text of the message: either `NULL` or a
#'   character vector. If non-`NULL`, the error message will finish with
#'   "Failed to `task`".
#' @param call The call stored in the condition object.
#' @seealso
#'   <http://adv-r.had.co.nz/Exceptions-Debugging.html#condition-handling>
#'   for more details about R's condition handling model
#' @export
#' @examples
#' \dontrun{
#' # You can use tryCatch to take different actions based on the type
#' # of error. Note that tryCatch will call the first handler that
#' # matches any classes of the condition, not the best matching, so
#' # always list handlers from most specific to least specific
#' f <- function(url) {
#'   tryCatch(stop_for_status(GET(url)),
#'     http_404 = function(c) "That url doesn't exist",
#'     http_403 = function(c) "You need to authenticate!",
#'     http_400 = function(c) "You made a mistake!",
#'     http_500 = function(c) "The server screwed up"
#'   )
#' }
#' f("http://httpbin.org/status/404")
#' f("http://httpbin.org/status/403")
#' f("http://httpbin.org/status/505")
#' }
http_condition <- function(x, type, task = NULL, call = sys.call(-1)) {
  type <- match.arg(type, c("error", "warning", "message"))

  if (is.null(task)) {
    task <- ""
  } else if (is.character(task)) {
    task <- paste0(" Failed to ", task, ".")
  } else {
    stop("`task` must be NULL or a character vector", call. = FALSE)
  }

  status <- status_code(x)
  reason <- http_status(status)$reason

  message <- sprintf("%s (HTTP %d).%s", reason, status, task)

  status_type <- (status %/% 100) * 100
  http_class <- paste0("http_", unique(c(status, status_type, "error")))

  structure(
    list(message = message, call = call),
    class = c(http_class, type, "condition")
  )
}

context("SSL")

test_that("SSL certificates verified as expected", {
  skip_on_cran()
  expect_equal(status_code(GET("https://google.com")), 200L)
  expect_equal(status_code(GET("https://firefox.com")), 200L)
})

context("http_error")

test_that("http_error works with urls", {
  skip_httpbin()
  expect_false(http_error("http://httpbin.org/status/200"))
  expect_true(http_error("http://httpbin.org/status/404"))
})

test_that("http_error works with responses", {
  skip_httpbin()
  r200 <- GET("http://httpbin.org/status/200")
  expect_false(http_error(r200))

  r404 <- GET("http://httpbin.org/status/404")
  expect_true(http_error(r404))
})

test_that("http_error works with integers", {
  expect_false(http_error(200L))
  expect_true(http_error(404L))
})

context("http_condition")

test_that("non failures passed through as is", {
  expect_equal(stop_for_status(200), 200)
})

test_that("status converted to errors", {
  expect_error(stop_for_status(300), "Multiple Choices (HTTP 300)", fixed = TRUE)
  expect_error(stop_for_status(404), "Not Found (HTTP 404)", fixed = TRUE)
  expect_error(stop_for_status(500), "Internal Server Error (HTTP 500)", fixed = TRUE)
})

test_that("task adds informative message", {
  expect_error(stop_for_status(300, "download"), "Failed to download.")
})

test_that("unknown status converted to error", {
  expect_error(stop_for_status(10000), "Unknown http status code: 10000", fixed = TRUE)
})

context("http_error")

test_that("http_error works with urls", {
  expect_false(http_error("http://httpbin.org/status/200"))
  expect_true(http_error("http://httpbin.org/status/404"))
})

test_that("http_error works with responses", {
  r200 <- GET("http://httpbin.org/status/200")
  expect_false(http_error(r200))

  r404 <- GET("http://httpbin.org/status/404")
  expect_true(http_error(r404))
})

test_that("http_error works with integers", {
  expect_false(http_error(200L))
  expect_true(http_error(404L))
})

use_cache <- function(cache = getOption("httr_oauth_cache")) {
  if (length(cache) != 1) {
    stop("cache should be length 1 vector", call. = FALSE)
  }
  if (!is.logical(cache) && !is.character(cache)) {
    stop("Cache must either be logical or string (file path)")
  }

  # If missing, see if it's ok to use one, and cache the results of
  # that check in a global option.
  if (is.na(cache)) {
    cache <- can_use_cache()
    options("httr_oauth_cache" = cache)
  }
  ## cache is now TRUE, FALSE or path

  if (isFALSE(cache)) {
    return(NULL)
  }

  if (isTRUE(cache)) {
    cache <- ".httr-oauth"
  }

  if (!file.exists(cache)) {
    create_cache(cache)
  }
  return(cache)
}

can_use_cache <- function(path = ".httr-oauth") {
  file.exists(path) || should_cache(path)
}

should_cache <- function(path = ".httr-oauth") {
  if (!interactive()) return(FALSE)

  cat("Use a local file ('", path, "'), to cache OAuth access credentials ",
    "between R sessions?\n",
    sep = ""
  )
  utils::menu(c("Yes", "No")) == 1
}

create_cache <- function(path = ".httr-oauth") {
  file.create(path, showWarnings = FALSE)
  if (!file.exists(path)) {
    stop("Failed to create local cache ('", path, "')", call. = FALSE)
  }

  # Protect cache as much as possible
  Sys.chmod(path, "0600")

  if (file.exists("DESCRIPTION")) {
    add_line(".Rbuildignore", paste0("^", gsub("\\.", "\\\\.", path), "$"))
  }
  add_line(".gitignore", path)

  TRUE
}


add_line <- function(path, line, quiet = FALSE) {
  if (file.exists(path)) {
    lines <- readLines(path, warn = FALSE)
    lines <- lines[lines != ""]
  } else {
    lines <- character()
  }

  if (line %in% lines) return(TRUE)
  if (!quiet) message("Adding ", line, " to ", path)

  lines <- c(lines, line)
  try(writeLines(lines, path))
  TRUE
}

cache_token <- function(token, cache_path) {
  if (is.null(cache_path)) return()

  tokens <- load_cache(cache_path)
  tokens[[token$hash()]] <- token
  saveRDS(tokens, cache_path)
}

fetch_cached_token <- function(hash, cache_path) {
  if (is.null(cache_path)) return()

  load_cache(cache_path)[[hash]]
}

remove_cached_token <- function(token) {
  if (is.null(token$cache_path)) return()

  tokens <- load_cache(token$cache_path)
  tokens[[token$hash()]] <- NULL
  saveRDS(tokens, token$cache_path)
}

load_cache <- function(cache_path) {
  if (!file.exists(cache_path) || file_size(cache_path) == 0) {
    list()
  } else {
    readRDS(cache_path)
  }
}

file_size <- function(x) file.info(x, extra_cols = FALSE)$size

library(httr)

# Facebook requires a https redirect url, which there's no general way
# to generate for a localhost url. Instead we'll use device flow:
# https://developers.facebook.com/docs/facebook-login/for-devices/

# 1. Register an application at https://developers.facebook.com/apps/
#    Make sure Product > Facebook login > Settings > Login from Devices = Yes.
#    Record your app_id and client token below.
app_id <- "314363036052185"
client_token <- "14b30b3ca83057c5a77ade469b496a26"
access_token <- paste0(app_id, "|", client_token)

# 2. Request a user code:
device_url <- "https://graph.facebook.com/v2.6/device/login"
resp <- POST(device_url, body = list(
  access_token = access_token,
  scope = "public_profile" # change this as needed
))
stop_for_status(resp)
request <- content(resp, type = "application/json")
message("Go to <", request$verification_uri, "> and enter ", request$user_code)

poll <- function(code) {
  poll_url <- "https://graph.facebook.com/v2.6/device/login_status"
  POST(poll_url, body = list(
    access_token = access_token,
    code = code
  ))
}
keep_waiting <- function(resp) {
  content <- content(resp, type = "application/json")
  if (!"error" %in% names(content)) {
    return(FALSE)
  }

  content$error$error_subcode == 1349174
}

cat("Waiting for confirmation")
while (keep_waiting(resp <- poll(request$code))) {
  cat(".")
  Sys.sleep(request$interval)
}
cat("\n")
stop_for_status(resp)
token <- content(resp)$access_token

# 3. Use API
req <- GET("https://graph.facebook.com/v2.3/me", query = list(
  fields = "name,picture",
  access_token = token
))
stop_for_status(req)
str(content(req))

library(httr)

# 1. Find OAuth settings for linkedin:
#    https://developer.linkedin.com/documents/linkedins-oauth-details
endpoints <- oauth_endpoints("linkedin")

# 2. Register an application at https://www.linkedin.com/secure/developer
#    Make sure to register http://localhost:1410/ as an "OAuth 2.0 Redirect URL".
#    (the trailing slash is important!)
#
#    Replace key and secret below.
myapp <- oauth_app("linkedin",
  key = "outmkw3859gy",
  secret = "n7vBr3lokGOCDKCd"
)

# 3. Get OAuth credentials and specify a scope your app has permission for
token <- oauth2.0_token(endpoints, myapp, scope = "r_liteprofile")

# 4. Use API
req <- GET("https://api.linkedin.com/v2/me", config(token = token))
stop_for_status(req)
content(req)

library(httr)

# 1. Find OAuth settings for vimeo:
#    http://vimeo.com/api/docs/authentication
oauth_endpoints("vimeo")

# 2. Register an application at https://developer.vimeo.com/apps
#    Replace key and secret below.
myapp <- oauth_app("vimeo",
  key = "bd535bc38ed5caccd79330ff33075eb9",
  secret = "51ab8cb2cbb8b7eb"
)

# 3. Get OAuth credentials
vimeo_token <- oauth2.0_token(
  oauth_endpoints("vimeo"), myapp,
  as_header = TRUE,
  use_basic_auth = TRUE
)

# 4. Use API
req <- GET(
  "https://api.vimeo.com/me/videos",
  config(token = vimeo_token)
)
stop_for_status(req)
str(jsonlite::fromJSON(content(req, "text")))

#' Upload a file with [POST()] or [PUT()].
#'
#' @param path path to file
#' @param type mime type of path. If not supplied, will be guess by
#'   [mime::guess_type()] when needed.
#' @export
#' @examples
#' citation <- upload_file(system.file("CITATION"))
#' POST("http://httpbin.org/post", body = citation)
#' POST("http://httpbin.org/post", body = list(y = citation))
upload_file <- function(path, type = NULL) {
  stopifnot(is.character(path), length(path) == 1, file.exists(path))

  if (is.null(type)) {
    type <- mime::guess_type(path)
  }

  curl::form_file(path, type)
}

#' @export
as.character.form_file <- function(x, ...) x

test_that("can upload with redirect", {
  skip_on_cran()

  str <- paste(letters, collapse = '')
  tmp <- tempfile()
  writeBin(str, tmp)

  resp <- PUT(
    "https://hb.cran.dev/redirect-to?url=/put&status_code=307",
    body = upload_file(tmp)
  )

  expect_equal(resp$status_code, 200)
  expect_equal(resp$url, "https://hb.cran.dev/put")
  json <- content(resp)
  expect_equal(json$data, str)
})

#' Create an OAuth application.
#'
#' See the demos for instructions on how to create an OAuth app for linkedin,
#' twitter, vimeo, facebook, github and google. When wrapping an API from a
#' package, the author may want to include a default app to facilitate early and
#' casual use and then provide a method for heavy or advanced users to supply
#' their own app or key and secret.
#'
#' @param appname name of the application.  This is not used for OAuth, but is
#'   used to make it easier to identify different applications.
#' @param key consumer key, also sometimes called the client ID
#' @param secret consumer secret, also sometimes called the client secret.
#'   Despite its name, this does not necessarily need to be protected like a
#'   password, i.e. the user still has to authenticate themselves and grant the
#'   app permission to access resources on their behalf. For example, see
#'   Google's docs for
#'   \href{https://developers.google.com/identity/protocols/OAuth2#installed}{OAuth2 for installed applications}.
#' @param redirect_uri The URL that user will be redirected to after
#'   authorisation is complete. You should generally leave this as the default
#'   unless you're using a non-standard auth flow (like with shiny).
#' @export
#' @family OAuth
#' @examples
#' \dontrun{
#' google_app <- oauth_app(
#'   "google",
#'   key = "123456789.apps.googleusercontent.com",
#'   secret = "abcdefghijklmnopqrstuvwxyz"
#' )
#' }
oauth_app <- function(appname,
                      key,
                      secret = NULL,
                      redirect_uri = oauth_callback()) {
  if (missing(secret)) {
    env_name <- paste0(toupper(appname), "_CONSUMER_SECRET")
    secret <- Sys.getenv(env_name)
    if (secret == "") {
      warning("Couldn't find secret in environment variable ", env_name,
        call. = FALSE
      )
      secret <- NULL
    } else {
      message("Using secret stored in environment variable ", env_name)
    }
  }
  structure(
    list(
      appname = appname,
      secret = secret,
      key = key,
      redirect_uri = redirect_uri
    ),
    class = "oauth_app"
  )
}

is.oauth_app <- function(x) inherits(x, "oauth_app")

#' @export
print.oauth_app <- function(x, ...) {
  cat("<oauth_app> ", x$appname, "\n", sep = "")
  cat("  key:    ", x$key, "\n", sep = "")
  secret <- if (is.null(x$secret)) "<not supplied>" else "<hidden>"
  cat("  secret: ", secret, "\n", sep = "")
  invisible(x)
}

request <- function(method = NULL, url = NULL, headers = NULL,
                    fields = NULL, options = NULL, auth_token = NULL,
                    output = NULL) {
  if (!is.null(method)) {
    stopifnot(is.character(method), length(method) == 1)
  }
  if (!is.null(url)) {
    stopifnot(is.character(url), length(url) == 1)
  }
  if (!is.null(headers)) {
    stopifnot(is.character(headers))
  }
  if (!is.null(fields)) {
    stopifnot(is.list(fields))
  }
  if (!is.null(output)) {
    stopifnot(inherits(output, "write_function"))
  }

  structure(
    list(
      method = method,
      url = url,
      headers = keep_last(headers),
      fields = fields,
      options = compact(keep_last(options)),
      auth_token = auth_token,
      output = output
    ),
    class = "request"
  )
}
is.request <- function(x) inherits(x, "request")

request_default <- function() {
  c(
    request(
      options = list(
        useragent = default_ua()
      ),
      headers = c(Accept = "application/json, text/xml, application/xml, */*"),
      output = write_function("write_memory")
    ),
    getOption("httr_config")
  )
}

#' @export
c.request <- function(...) {
  Reduce(request_combine, list(...))
}

as.request <- function(x) UseMethod("as.request")
as.request.list <- function(x) structure(x, class = "request")
as.request.request <- function(x) x
as.request.NULL <- function(x) request()
as.request.Token <- function(x) request(auth_token = x)

request_build <- function(method, url, ...) {
  extra <- list(...)
  extra[has_name(extra)] <- NULL

  if (method == "HEAD") {
    extra <- c(extra, list(config(nobody = TRUE)))
  }

  req <- Reduce(request_combine, extra, init = request())

  req$method <- toupper(method)
  req$url <- url

  req
}

request_combine <- function(x, y) {
  if (length(x) == 0 && length(y) == 0) return(request())
  if (length(x) == 0) return(y)
  if (length(y) == 0) return(x)
  stopifnot(is.request(x), is.request(y))

  request(
    method = y$method %||% x$method,
    url = y$url %||% x$url,
    headers = keep_last(x$headers, y$headers),
    fields = c(x$fields, y$fields),
    options = keep_last(x$options, y$options),
    auth_token = y$auth_token %||% x$auth_token,
    output = y$output %||% x$output
  )
}

#' @export
print.request <- function(x, ...) {
  cat("<request>\n")
  if (!is.null(x$method) && !is.null(x$url)) {
    cat(x$method, " ", x$url, "\n", sep = "")
  }
  if (!is.null(x$output)) {
    cat("Output: ", class(x$output)[[1]], "\n", sep = "")
  }
  named_vector("Options", x$options)
  named_vector("Headers", x$headers)
  named_vector("Fields", x$fields)
  if (!is.null(x$auth_token)) {
    cat("Auth token: ", class(x$auth_token)[[1]], "\n", sep = "")
  }

  invisible(x)
}

request_prepare <- function(req) {
  req <- request_combine(request_default(), req)

  # Use specific options for GET and POST; otherwise, perform a custom request.
  # The PUT/UPLOAD options don't appear to work, instead hanging forever.
  switch(req$method,
    GET = req$options$httpget <- TRUE,
    POST = req$options$post <- TRUE,
    req$options$customrequest <- req$method
  )

  # Sign request, if needed
  token <- req$auth_token
  if (!is.null(token)) {
    signed_req <- token$sign(req$method, req$url)
    stopifnot(is.request(signed_req))
    req <- c(req, signed_req)
  }

  req
}

request_perform <- function(req, handle, refresh = TRUE) {
  stopifnot(is.request(req), inherits(handle, "curl_handle"))

  req <- request_prepare(req)

  ## This callback can cancel the request
  if (!is.null(res <- perform_callback("request", req = req))) return(res)

  curl::handle_setopt(handle, .list = req$options)
  if (!is.null(req$fields)) {
    curl::handle_setform(handle, .list = req$fields)
  }
  curl::handle_setheaders(handle, .list = req$headers)
  on.exit(curl::handle_reset(handle), add = TRUE)

  resp <- request_fetch(req$output, req$url, handle)

  # If return 401 and have auth token, refresh it and then try again
  needs_refresh <- refresh && resp$status_code == 401L &&
    !is.null(req$auth_token) && req$auth_token$can_refresh()
  if (needs_refresh) {
    message("Auto-refreshing stale OAuth token.")
    req$auth_token$refresh()
    return(request_perform(req, handle, refresh = FALSE))
  }

  url_scheme <- parse_url(resp$url)$scheme
  is_http <- tolower(url_scheme) %in% c("http", "https")
  if (is_http) {
    all_headers <- parse_http_headers(resp$headers)
    headers <- last(all_headers)$headers
  } else {
    all_headers <- NULL
    headers <- resp$headers
  }

  if (is_http && !is.null(headers$date)) {
    date <- parse_http_date(headers$Date)
  } else {
    date <- Sys.time()
  }

  res <- response(
    url = resp$url,
    status_code = resp$status_code,
    headers = headers,
    all_headers = all_headers,
    cookies = curl::handle_cookies(handle),
    content = resp$content,
    date = date,
    times = resp$times,
    request = req,
    handle = handle
  )

  ## If the callback provides a result, we return that
  if (!is.null(cbres <- perform_callback("response", req, res))) {
    return(cbres)
  }

  res
}

context("Request")

test_that("c.request overwrites repeated options", {
  expect_equal(
    c(request(url = "a"), request(url = "b")),
    request(url = "b")
  )
})

test_that("c.request merges headers", {
  expect_equal(
    c(request(headers = c("a" = "a")), request(headers = c("b" = "b"))),
    request(headers = c("a" = "a", "b" = "b"))
  )
})

test_that('request_build upper cases verbs', {
  expect_equal(request_build('get', 'asdf.com')$method, "GET")
  expect_equal(request_build('post', 'asdf.com')$method, "POST")
})

test_that("non-http methods don't parse headers", {
  skip_on_cran()
  skip_on_ci()

  # Must not reuse or FTP connection is closed at the wrong time,
  # causing problems in final test
  r <- GET(
    "ftp://cran.r-project.org/incoming/",
    config(forbid_reuse = TRUE)
  )
  expect_type(r$headers, "raw")
})

context("Request")

test_that("c.request overwrites repeated options", {
  expect_equal(
    c(request(url = "a"), request(url = "b")),
    request(url = "b")
  )
})

test_that("c.request merges headers", {
  expect_equal(
    c(request(headers = c("a" = "a")), request(headers = c("b" = "b"))),
    request(headers = c("a" = "a", "b" = "b"))
  )
})

test_that('request_build upper cases verbs', {
  expect_equal(request_build('get', 'asdf.com')$method, "GET")
  expect_equal(request_build('post', 'asdf.com')$method, "POST")
})

test_that("non-http methods don't parse headers", {
  # skip on travis to avoid hammering the FTP server, which doesn't
  # seem to be able to handle multiple simultaneous requests
  skip_on_travis()

  # Must not reuse or FTP connection is closed at the wrong time,
  # causing problems in final test
  r <- GET(
    "ftp://cran.r-project.org/incoming/",
    config(forbid_reuse = TRUE)
  )
  expect_type(r$headers, "raw")
})

# These are drop in replacements for the stringr:: functions used in httr. They
# do not retain all functionality from stringr, only that which is used in
# httr. Notably they are generally not vectorized.
str_trim <- function(x) {
  gsub("(^\\s+)|(\\s+$)", "", x)
}

str_split_fixed <- function(string, pattern, n) {
  if (length(string) == 0) return(matrix(character(), nrow = 1, ncol = n))
  m <- gregexpr(pattern, string)[[1]]
  if (length(m) == 1 && m == -1) {
    res <- string
  } else {
    m_starts <- m
    m_ends <- m + attr(m, "match.length") - 1L
    starts <- c(1, m_ends + 1L)[seq_len(n)]
    ends <- c((m_starts - 1L)[seq_len(n - 1)], nchar(string))
    res <- lapply(string, function(x)
      unlist(Map(substr, x, starts, ends, USE.NAMES = FALSE))
    )
  }

  mat <- matrix("", nrow = length(res), ncol = n, byrow = TRUE)
  mat[seq_along(unlist(res))] <- unlist(res)
  mat[, seq_len(n), drop = FALSE]
}

str_split <- function(string, pattern, n = Inf) {
  res <- strsplit(string, pattern)
  if (is.finite(n)) {
    res[seq_len(n)]
  } else {
    res
  }
}

str_detect <- function(string, pattern) {
  grepl(pattern, string)
}

str_match <- function(string, pattern) {
  m <- regexpr(pattern, string, perl = TRUE)
  cbind(
    substr(string, m, attr(m, "match.length") + m - 1L),
    substr(
      string, attr(m, "capture.start"),
      attr(m, "capture.length") + attr(m, "capture.start") - 1L
    )
  )
}

str_replace <- function(string, pattern, replace) {
  m <- regexpr(pattern, string)
  regmatches(string, m) <- replace
  string
}

str_extract <- function(string, pattern) {
  m <- regexpr(pattern, string)
  res <- substr(string, m, attr(m, "match.length") + m - 1L)
  res[m == -1] <- NA_character_
  res
}

init_oauth_service_account <- function(secrets, scope = NULL, sub = NULL) {
  signature <- jwt_signature(
    secrets,
    aud = secrets$token_uri,
    scope = scope,
    sub = sub
  )

  res <- POST(
    secrets$token_uri,
    body = list(
      grant_type = "urn:ietf:params:oauth:grant-type:jwt-bearer",
      assertion = signature
    ),
    encode = "form"
  )
  stop_for_status(res)

  content(res, type = "application/json")
}

#' Generate a JWT signature given credentials.
#'
#' As described in
#' <https://developers.google.com/accounts/docs/OAuth2ServiceAccount>
#'
#' @param credentials Parsed contents of the credentials file.
#' @param scope A space-delimited list of the permissions that the application
#'    requests.
#' @param iss Email address of the client_id of the application making the
#'   access token request.
#' @param scope A space-delimited list of the permissions that the application
#'    requests.
#' @param sub The email address of the user for which the application is
#'    requesting delegated access.
#' @param aud A descriptor of the intended target of the assertion. This
#'    typically comes from the service auth file.
#' @param iat The time the assertion was issued, measured in seconds since
#'    00:00:00 UTC, January 1, 1970.
#' @param exp The expiration time of the assertion, measured in seconds since
#'   00:00:00 UTC, January 1, 1970. This value has a maximum of 1 hour from
#'   the issued time.
#' @param duration Duration of token, in seconds.
#' @keywords internal
#' @examples
#' \dontrun{
#' cred <- jsonlite::fromJSON("~/Desktop/httrtest-45693cbfac92.json")
#' jwt_signature(cred, "https://www.googleapis.com/auth/userinfo.profile")
#' }
jwt_signature <- function(credentials,
                          scope,
                          aud,
                          sub = NULL,
                          iat = as.integer(Sys.time()),
                          exp = iat + duration,
                          duration = 60L * 60L) {
  cs <- compact(list(
    iss = credentials$client_email,
    scope = scope,
    aud = aud,
    sub = sub,
    iat = iat,
    exp = exp
  ))

  jwt_sign(cs, credentials$private_key)
}

jwt_sign <- function(claimset, private_key, header = jwt_header()) {
  key <- openssl::read_key(private_key)
  to_sign_base64 <- paste0(jwt_base64(header), ".", jwt_base64(claimset))
  to_sign <- charToRaw(to_sign_base64)
  sig <- openssl::signature_create(to_sign, openssl::sha256, key)
  sig_base64 <- base64url(sig)

  paste0(to_sign_base64, ".", sig_base64)
}

jwt_header <- function() {
  list(
    alg = "RS256",
    typ = "JWT"
  )
}


jwt_base64 <- function(x) base64url(jwt_json(x))
jwt_json <- function(x) jsonlite::toJSON(x, auto_unbox = TRUE)
base64url <- function(x) {
  if (is.character(x)) {
    x <- charToRaw(x)
  }
  out <- chartr("+/", "-_", openssl::base64_encode(x))
  gsub("=+$", "", out)
}

context("OAuth: server side")

# Reference values from
# https://developers.google.com/accounts/docs/OAuth2ServiceAccount

test_that("reference header yields expected base64", {
  expect_equal(jwt_base64(jwt_header()), "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCJ9")
})

test_that("reference claimset yields expected base64", {
  cs <- list(
    iss = "761326798069-r5mljlln1rd4lrbhg75efgigp36m78j5@developer.gserviceaccount.com",
    scope = "https://www.googleapis.com/auth/prediction",
    aud = "https://accounts.google.com/o/oauth2/token",
    exp = 1328554385,
    iat = 1328550785
  )
  expected <- "eyJpc3MiOiI3NjEzMjY3OTgwNjktcjVtbGpsbG4xcmQ0bHJiaGc3NWVmZ2lncDM2bTc4ajVAZGV2ZWxvcGVyLmdzZXJ2aWNlYWNjb3VudC5jb20iLCJzY29wZSI6Imh0dHBzOi8vd3d3Lmdvb2dsZWFwaXMuY29tL2F1dGgvcHJlZGljdGlvbiIsImF1ZCI6Imh0dHBzOi8vYWNjb3VudHMuZ29vZ2xlLmNvbS9vL29hdXRoMi90b2tlbiIsImV4cCI6MTMyODU1NDM4NSwiaWF0IjoxMzI4NTUwNzg1fQ"

  expect_equal(jwt_base64(cs), expected)
})

oauth2.0_error_codes <- c(
  400,
  401
)

oauth2.0_errors <- c(
  "invalid_request",
  "invalid_client",
  "invalid_grant",
  "unauthorized_client",
  "unsupported_grant_type",
  "invalid_scope"
)

# This implements error checking according to the OAuth2.0
# specification: https://tools.ietf.org/html/rfc6749#section-5.2
find_oauth2.0_error <- function(response) {
  if (!status_code(response) %in% oauth2.0_error_codes) {
    return(NULL)
  }

  content <- content(response)
  if (!content$error %in% oauth2.0_errors) {
    return(NULL)
  }

  list(
    error = content$error,
    error_description = content$error_description,
    error_uri = content$error_uri
  )
}

#' Open specified url in browser.
#'
#' (This isn't really a http verb, but it seems to follow the same format).
#'
#' Only works in interactive sessions.
#'
#' @param config All configuration options are ignored because the request
#'   is handled by the browser, not \pkg{RCurl}.
#' @inherit GET params return
#' @family http methods
#' @export
#' @examples
#' BROWSE("http://google.com")
#' BROWSE("http://had.co.nz")
BROWSE <- function(url = NULL, config = list(), ..., handle = NULL) {
  hu <- handle_url(handle, url, ...)
  if (interactive()) {
    utils::browseURL(hu$url)
  } else {
    message("Please point your browser to the following url: ")
    message(hu$url)
  }
}

#' Set cookies.
#'
#' @param ... a named cookie values
#' @param .cookies a named character vector
#' @export
#' @family config
#' @seealso [cookies()] to see cookies in response.
#' @examples
#' set_cookies(a = 1, b = 2)
#' set_cookies(.cookies = c(a = "1", b = "2"))
#'
#' \dontrun{
#' GET("http://httpbin.org/cookies")
#' GET("http://httpbin.org/cookies", set_cookies(a = 1, b = 2))
#' }
set_cookies <- function(..., .cookies = character(0)) {
  cookies <- c(..., .cookies)
  stopifnot(is.character(cookies))

  cookies_str <- vapply(cookies, curl::curl_escape, FUN.VALUE = character(1))

  cookie <- paste(names(cookies), cookies_str, sep = "=", collapse = ";")

  config(cookie = cookie)
}

#' Access cookies in a response.
#'
#' @seealso [set_cookies()] to send cookies in request.
#' @param x A response.
#' @examples
#' \dontrun{
#' r <- GET("http://httpbin.org/cookies/set", query = list(a = 1, b = 2))
#' cookies(r)
#' }
#' @export
cookies <- function(x) UseMethod("cookies")

#' @export
cookies.response <- function(x) x$cookies

#' @export
cookies.handle <- function(x) {
  curl::handle_cookies(x$handle)
}

#' Send a DELETE request.
#'
#' @section RFC2616:
#'
#' The DELETE method requests that the origin server delete the resource
#' identified by the Request-URI. This method MAY be overridden by human
#' intervention (or other means) on the origin server. The client cannot be
#' guaranteed that the operation has been carried out, even if the status code
#' returned from the origin server indicates that the action has been
#' completed successfully. However, the server SHOULD NOT indicate success
#' unless, at the time the response is given, it intends to delete the
#' resource or move it to an inaccessible location.
#'
#' A successful response SHOULD be 200 (OK) if the response includes an entity
#' describing the status, 202 (Accepted) if the action has not yet been
#' enacted, or 204 (No Content) if the action has been enacted but the
#' response does not include an entity.
#'
#' If the request passes through a cache and the Request-URI identifies one or
#' more currently cached entities, those entries SHOULD be treated as stale.
#' Responses to this method are not cacheable.
#'
#' @inherit GET params return
#' @inheritParams POST
#' @family http methods
#' @export
#' @examples
#' DELETE("http://httpbin.org/delete")
#' POST("http://httpbin.org/delete")
DELETE <- function(url = NULL, config = list(), ...,
                   body = NULL, encode = c("multipart", "form", "json", "raw"),
                   handle = NULL) {
  hu <- handle_url(handle, url, ...)
  req <- request_build(
    "DELETE", hu$url, body_config(body, match.arg(encode)),
    as.request(config), ...
  )
  request_perform(req, hu$handle$handle)
}

# Good example for testing
# http://stevenlevithan.com/demo/parseuri/js/

#' Parse and build urls according to RFC3986.
#'
#' See <https://www.rfc-editor.org/rfc/rfc3986> for details of parsing
#' algorithm.
#'
#' @param url For `parse_url` a character vector (of length 1) to parse
#'   into components; for `build_url` a list of components to turn back
#'   into a string.
#' @return a list containing:
#' * scheme
#' * hostname
#' * port
#' * path
#' * params
#' * fragment
#' * query, a list
#' * username
#' * password
#' @export
#' @examples
#' parse_url("http://google.com/")
#' parse_url("http://google.com:80/")
#' parse_url("http://google.com:80/?a=1&b=2")
#'
#' url <- parse_url("http://google.com/")
#' url$scheme <- "https"
#' url$query <- list(q = "hello")
#' build_url(url)
parse_url <- function(url) {
  if (is.url(url)) return(url)

  url <- as.character(url)
  stopifnot(length(url) == 1)

  pull_off <- function(pattern) {
    if (!str_detect(url, pattern)) return(NULL)

    piece <- str_match(url, pattern)[, 2]
    url <<- str_replace(url, pattern, "")

    piece
  }

  fragment <- pull_off("#(.*)$")
  scheme <- pull_off("^([[:alpha:]][[:alpha:][:digit:]+.-]*):")
  netloc <- pull_off("^//([^/?]*)/?")

  if (identical(netloc, "")) { # corresponds to ///
    url <- paste0("/", url)
    port <- username <- password <- hostname <- NULL
  } else if (!is.null(netloc)) {
    pieces <- strsplit(netloc, "@")[[1]]
    if (length(pieces) == 1) {
      username <- NULL
      password <- NULL

      host <- pieces
    } else {
      user_pass <- strsplit(pieces[[1]], ":")[[1]]
      username <- user_pass[1]
      if (length(user_pass) == 1) {
        password <- NULL
      } else {
        password <- user_pass[2]
      }

      host <- pieces[2]
    }

    host_pieces <- str_split(host, ":")[[1]]
    hostname <- host_pieces[1]
    port <- if (length(host_pieces) > 1) host_pieces[2]
  } else {
    port <- username <- password <- hostname <- NULL
  }

  query <- pull_off("\\?(.*)$")
  if (!is.null(query)) {
    query <- parse_query(query)
  }
  params <- pull_off(";(.*)$")

  structure(list(
    scheme = scheme, hostname = hostname, port = port, path = url,
    query = query, params = params, fragment = fragment,
    username = username, password = password
  ),
  class = "url"
  )
}

is.url <- function(x) inherits(x, "url")
print.url <- function(x, ...) {
  cat("Url: ", build_url(x), "\n", sep = "")
  invisible(x)
}
"[.url" <- function(x, ...) {
  structure(NextMethod(), class = "url")
}

#' @export
#' @rdname parse_url
build_url <- function(url) {
  stopifnot(is.url(url))

  scheme <- url$scheme
  hostname <- url$hostname

  if (!is.null(url$port)) {
    port <- paste0(":", url$port)
  } else {
    port <- NULL
  }

  path <- paste(gsub("^/", "", url$path), collapse = "/")

  if (!is.null(url$params)) {
    params <- paste0(";", url$params)
  } else {
    params <- NULL
  }

  if (is.list(url$query)) {
    query <- compose_query(compact(url$query))
  } else {
    query <- url$query
  }
  if (!is.null(query) && nzchar(query)) {
    stopifnot(is.character(query), length(query) == 1)
    query <- paste0("?", query)
  }

  if (is.null(url$username) && !is.null(url$password)) {
    stop("Cannot set password without username")
  }

  paste0(
    scheme, "://",
    url$username, if (!is.null(url$password)) ":", url$password,
    if (!is.null(url$username)) "@",
    hostname, port, "/", path, params, query,
    if (!is.null(url$fragment)) "#", url$fragment
  )
}

#' Modify a url.
#'
#' Modify a url by first parsing it and then replacing components with
#' the non-NULL arguments of this function.
#'
#' @export
#' @param url the url to modify
#' @param scheme,hostname,port,path,query,params,fragment,username,password
#'   components of the url to change
modify_url <- function(url, scheme = NULL, hostname = NULL, port = NULL,
                       path = NULL, query = NULL, params = NULL, fragment = NULL,
                       username = NULL, password = NULL) {
  old <- parse_url(url)
  new <- compact(list(
    scheme = scheme, hostname = hostname, port = port, path = path,
    query = query, params = params, fragment = fragment,
    username = username, password = password
  ))

  build_url(utils::modifyList(old, new))
}

context("URL parsing and building")

test_that("parse_url works as expected", {
  urls <- list(
    "http://google.com/",
    "http://google.com/path",
    "http://google.com/path?a=1&b=2",
    "http://google.com/path;param?a=1&b=2",
    "http://google.com:80/path;param?a=1&b=2",
    "http://google.com:80/path;param?a=1&b=2#frag",
    "http://user@google.com:80/path;param?a=1&b=2",
    "http://user:pass@google.com:80/path;param?a=1&b=2",
    "svn+ssh://my.svn.server/repo/trunk"
  )

  expect_equal(
    lapply(urls, function(u) build_url(parse_url(u))),
    urls
  )
})


test_that("empty queries not converted to NA", {
  expect_equal(parse_url("http://x.com/?q=")$query, list(q = ""))
  expect_equal(parse_url("http://x.com/?q")$query, list(q = ""))

  expect_equal(parse_url("http://x.com/?a&q")$query, list(a = "", q = ""))
})

test_that("query strings escaped and unescaped correctly", {
  url <- "http://x.com/?x%20y=a%20b"
  parsed <- parse_url(url)
  expect_equal(parsed$query, list("x y" = "a b"))
  expect_equal(build_url(parsed), url)
})

test_that("password and no username is an error", {
  url <- "http://www.example.com/"
  parsed <- parse_url(url)
  expect_equal(build_url(parsed), url)
  parsed$password <- "secret"
  expect_error(build_url(parsed), "password without username")
})


test_that("handle_url modifies url with named components", {
  hu <- handle_url(NULL, "http://google.com")
  expect_equal(hu$url, "http://google.com")

  hu <- handle_url(NULL, "http://google.com", path = "abc")
  expect_equal(hu$url, "http://google.com/abc")
})

test_that("handle_url ignores unnamed arguments", {
  hu <- handle_url(NULL, "http://google.com", 1, 2, 3)
  expect_equal(hu$url, "http://google.com")
})

test_that("build_url collapse path", {
  url <- modify_url("http://google.com", path = c("one", "two"))
  expect_equal(url, "http://google.com/one/two")
})

test_that("build_url drops leading / in path", {
  url <- modify_url("http://google.com", path = "/one")
  expect_equal(url, "http://google.com/one")
})

test_that("build_url drops null or empty query", {
  url <- modify_url("http://google.com", query = list(a = 1, b = NULL))
  expect_equal(url, "http://google.com/?a=1")
  url <- modify_url("http://google.com", query = list(a = NULL))
  expect_equal(url, "http://google.com/")
  url <- modify_url("http://google.com", query = list())
  expect_equal(url, "http://google.com/")
})

test_that("parse_url pulls off domain correctly given query without trailing '/'", {
  url <- modify_url("http://google.com?a=1", query = list(b = 2))
  expect_equal(url, "http://google.com/?a=1&b=2")
})

test_that("parse_url preserves leading / in path", {
  url <- parse_url("file:///tmp/foobar")
  expect_equal(url$path, "/tmp/foobar")
})

test_that("scheme starts with alpha", {
  url <- parse_url("+ab://host/tmp/foobar")
  expect_equal(url$scheme, NULL)
})

test_that("scheme can contain digits", {
  url <- parse_url("ab1://host/tmp/foobar")
  expect_equal(url$scheme, "ab1")
})

test_that("scheme can contain plus", {
  url <- parse_url("a+b://host/tmp/foobar")
  expect_equal(url$scheme, "a+b")
})

test_that("scheme can contain period", {
  url <- parse_url("a.b://host/tmp/foobar")
  expect_equal(url$scheme, "a.b")
})

test_that("scheme can contain hyphen", {
  url <- parse_url("a-b://host/tmp/foobar")
  expect_equal(url$scheme, "a-b")
})

test_that("scheme can be a single character", {
  url <- parse_url("a://host/tmp/foobar")
  expect_equal(url$scheme, "a")
})

# compose_query -----------------------------------------------------------

test_that("I() prevents escaping", {
  expect_equal(compose_query(list(x = I("&"))), "x=&")
})

test_that("null elements are dropped", {
  expect_equal(compose_query(list(x = 1, y = NULL)), "x=1")
})

#' Sign an OAuth request
#'
#' Deprecated. Instead create a config object directly using
#' `config(token = my_token)`.
#'
#' @keywords internal
#' @name sign_oauth
NULL

#' @export
#' @rdname sign_oauth
sign_oauth1.0 <- function(app, token = NULL, token_secret = NULL,
                          as_header = TRUE, ...) {
  params <- list(as_header = as_header)

  credentials <- list(oauth_token = token, oauth_token_secret = token_secret)
  token <- Token1.0$new(
    endpoint = NULL, params = params, app = app,
    credentials = credentials
  )
  request(auth_token = token)
}

#' @export
#' @rdname sign_oauth
sign_oauth2.0 <- function(access_token, as_header = TRUE) {
  stop("Deprecated: supply token object to config directly", call. = FALSE)
}

#' Generate oauth signature.
#'
#' For advanced use only. Occassionally needed for sites that use some
#' components of the OAuth spec, but not all of them (e.g. 2-legged oauth)
#'
#' @param url,method Url and http method of request.
#' @param app [oauth_app()] object representing application.
#' @param token,token_secret OAuth token and secret.
#' @param other_params Named argument providing additional parameters
#'   (e.g. `oauth_callback` or `oauth_body_hash`).
#' @export
#' @keywords internal
#' @return A list of oauth parameters.
oauth_signature <- function(url, method = "GET", app, token = NULL,
                            token_secret = NULL,
                            private_key = NULL, other_params = NULL) {
  if (!is.null(private_key)) {
    signature_method <- "RSA-SHA1"
  } else {
    signature_method <- "HMAC-SHA1"
  }

  oauth_params <- compact(list(
    oauth_consumer_key = app$key,
    oauth_nonce = nonce(),
    oauth_signature_method = signature_method,
    oauth_timestamp = as.integer(Sys.time()),
    oauth_version = "1.0",
    oauth_token = token
  ))
  if (length(other_params) > 0) {
    oauth_params <- c(oauth_params, other_params)
  }

  norm_method <- toupper(method)
  norm_url <- oauth_normalise_url(url)
  norm_params <- oauth_normalize_params(url, oauth_params)
  norm_req <- paste0(
    norm_method, "&",
    oauth_encode(norm_url), "&",
    oauth_encode(norm_params)
  )

  # Generate signature
  if (signature_method == "HMAC-SHA1") {
    private_key <- paste0(oauth_encode(app$secret), "&", oauth_encode(token_secret))
  }
  oauth_params$oauth_signature <- sha1_hash(private_key, norm_req, signature_method)

  sort_names(oauth_params)
}

#' @rdname oauth_signature
#' @export
oauth_header <- function(info) {
  oauth <- paste0(
    "OAuth ",
    paste0(
      oauth_encode(names(info)), "=\"", oauth_encode(info), "\"",
      collapse = ", "
    )
  )
  add_headers(Authorization = oauth)
}

oauth_encode <- function(x) vapply(x, oauth_encode1, character(1))

# As described in http://tools.ietf.org/html/rfc5849#page-29
oauth_encode1 <- function(x) {
  encode <- function(x) paste0("%", toupper(as.character(charToRaw(x))), collapse = "")

  x <- as.character(x)
  chars <- strsplit(x, "")[[1]]
  ok <- !str_detect(chars, "[^A-Za-z0-9_.~-]")

  if (all(ok)) return(x)

  chars[!ok] <- unlist(lapply(chars[!ok], encode))
  paste0(chars, collapse = "")
}

oauth_normalise_url <- function(url) {
  url <- parse_url(url)

  # > Unless specified, URL scheme and authority MUST be lowercase and include
  # > the port number; http default port 80 and https default port 443 MUST be
  # > excluded.
  # > --- https://oauth.net/core/1.0/#anchor14
  url$scheme <- tolower(url$scheme)

  if (url$scheme == "http" && identical(url$port, "80")) {
    url$port <- NULL
  } else if (url$scheme == "https" && identical(url$port, "443")) {
    url$port <- NULL
  }

  build_url(url[c("scheme", "hostname", "port", "url", "path")])
}

oauth_normalize_params <- function(url, extra) {
  # Collect params, encode, sort and concatenate into a single string
  url <- parse_url(url)

  params <- c(url$query, extra)
  params_esc <- stats::setNames(oauth_encode(params), oauth_encode(names(params)))
  params_srt <- sort_names(params_esc)
  params_str <- paste0(names(params_srt), "=", params_srt, collapse = "&")

  params_str
}

context("oauth_signature")


# Url normalisation -------------------------------------------------------

test_that("protocol is lower-cased", {
  expect_equal(oauth_normalise_url("HTTP://google.com"), "http://google.com/")
})

test_that("default ports are stripped", {
  expect_equal(oauth_normalise_url("http://google.com:80"), "http://google.com/")
  expect_equal(oauth_normalise_url("https://google.com:443"), "https://google.com/")
})

#' GET a url.
#'
#' @section RFC2616:
#' The GET method means retrieve whatever information (in the form of an
#' entity) is identified by the Request-URI. If the Request-URI refers to a
#' data-producing process, it is the produced data which shall be returned as
#' the entity in the response and not the source text of the process, unless
#' that text happens to be the output of the process.
#'
#' The semantics of the GET method change to a "conditional GET" if the
#' request message includes an If-Modified-Since, If-Unmodified-Since,
#' If-Match, If-None-Match, or If-Range header field. A conditional GET method
#' requests that the entity be transferred only under the circumstances
#' described by the conditional header field(s). The conditional GET method is
#' intended to reduce unnecessary network usage by allowing cached entities to
#' be refreshed without requiring multiple requests or transferring data
#' already held by the client.
#'
#' The semantics of the GET method change to a "partial GET" if the request
#' message includes a Range header field. A partial GET requests that only
#' part of the entity be transferred, as described in <http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.35>
#' The partial GET method is intended to reduce unnecessary network usage by
#' allowing partially-retrieved entities to be completed without transferring
#' data already held by the client.
#'
#' @param url the url of the page to retrieve
#' @param ... Further named parameters, such as `query`, `path`, etc,
#'   passed on to [modify_url()]. Unnamed parameters will be combined
#'   with [config()].
#' @param config Additional configuration settings such as http
#'   authentication ([authenticate()]), additional headers
#'   ([add_headers()]), cookies ([set_cookies()]) etc.
#'   See [config()] for full details and list of helpers.
#' @param handle The handle to use with this request. If not
#'   supplied, will be retrieved and reused from the [handle_pool()]
#'   based on the scheme, hostname and port of the url. By default \pkg{httr}
#    automatically reuses the same http connection (aka handle) for mulitple
#'   requests to the same scheme/host/port combo. This substantially reduces
#'   connection time, and ensures that cookies are maintained over multiple
#'   requests to the same host. See [handle_pool()] for more
#'   details.
#'
#' @return A [response()] object.
#'
#' @family http methods
#' @export
#' @examples
#' GET("http://google.com/")
#' GET("http://google.com/", path = "search")
#' GET("http://google.com/", path = "search", query = list(q = "ham"))
#' 
#' # See what GET is doing with httpbin.org
#' url <- "http://httpbin.org/get"
#' GET(url)
#' GET(url, add_headers(a = 1, b = 2))
#' GET(url, set_cookies(a = 1, b = 2))
#' GET(url, add_headers(a = 1, b = 2), set_cookies(a = 1, b = 2))
#' GET(url, authenticate("username", "password"))
#' GET(url, verbose())
#' 
#' # You might want to manually specify the handle so you can have multiple
#' # independent logins to the same website.
#' google <- handle("http://google.com")
#' GET(handle = google, path = "/")
#' GET(handle = google, path = "search")
GET <- function(url = NULL, config = list(), ..., handle = NULL) {
  hu <- handle_url(handle, url, ...)
  req <- request_build("GET", hu$url, as.request(config), ...)
  request_perform(req, hu$handle$handle)
}

#' Create a handle tied to a particular host.
#'
#' This handle preserves settings and cookies across multiple requests. It is
#' the foundation of all requests performed through the httr package, although
#' it will mostly be hidden from the user.
#'
#' @param url full url to site
#' @param cookies DEPRECATED
#' @export
#' @note
#' Because of the way argument dispatch works in R, using handle() in the
#' http methods (See [GET()]) will cause problems when trying to
#' pass configuration arguments (See examples below). Directly specifying the
#' handle when using http methods is not recommended in general, since the
#' selection of the correct handle is taken care of when the user passes an url
#' (See [handle_pool()]).
#'
#' @examples
#' handle("http://google.com")
#' handle("https://google.com")
#' 
#' h <- handle("http://google.com")
#' GET(handle = h)
#' # Should see cookies sent back to server
#' GET(handle = h, config = verbose())
#' 
#' h <- handle("http://google.com", cookies = FALSE)
#' GET(handle = h)$cookies
#' \dontrun{
#' # Using the preferred way of configuring the http methods
#' # will not work when using handle():
#' GET(handle = h, timeout(10))
#' # Passing named arguments will work properly:
#' GET(handle = h, config = list(timeout(10), add_headers(Accept = "")))
#' }
#' 
handle <- function(url, cookies = TRUE) {
  stopifnot(is.character(url), length(url) == 1)

  if (!missing(cookies)) {
    warning("Cookies argument is deprecated", call. = FALSE)
  }

  h <- curl::new_handle()
  structure(list(handle = h, url = url), class = "handle")
}

#' @export
print.handle <- function(x, ...) {
  cat("Host: ", x$url, " <", ref(x$handle), ">\n", sep = "")
  invisible(x)
}

ref <- function(x) {
  str_extract(utils::capture.output(print(x))[1], "0x[0-9a-f]*")
}

is.handle <- function(x) inherits(x, "handle")

#' Revoke all OAuth tokens in the cache.
#'
#' Use this function if you think that your token may have been compromised,
#' e.g. you accidentally uploaded the cache file to github. It's not possible
#' to automatically revoke all tokens - this function will warn when it can't.
#'
#' @param cache_path Path to cache file. Defaults to `.httr-oauth` in
#'   current directory.
#' @export
revoke_all <- function(cache_path = NA) {
  cache_path <- use_cache(cache_path)

  if (is.null(cache_path)) {
    stop("Can't find cache")
  }

  tokens <- load_cache(cache_path)

  cant_revoke <- vapply(
    tokens, function(x) is.null(x$endpoint$revoke),
    logical(1)
  )
  if (any(cant_revoke)) {
    manual <- tokens[cant_revoke]
    apps <- vapply(manual, function(x) {
      paste0(x$app$appname, " (", x$app$key, ")")
    }, character(1), USE.NAMES = FALSE)
    warning("Can't revoke the following tokens automatically: ",
      paste0(apps, collapse = ", "),
      call. = FALSE
    )
  }

  lapply(tokens, function(x) try(revoke_oauth2.0(x)))
  invisible(TRUE)
}

revoke_oauth2.0 <- function(endpoint, credentials) {
  if (is.null(endpoint$revoke)) {
    stop("No revoke endpoint", call. = FALSE)
  }

  url <- modify_url(endpoint$revoke,
    query = list(token = credentials$access_token)
  )
  response <- GET(url, accept_json())
  stop_for_status(response)

  invisible(TRUE)
}

validate_oauth2.0 <- function(endpoint, credentials) {
  validate_url <- endpoint_validation_url(endpoint, credentials)
  response <- GET(validate_url, accept_json())
  status_code(response) == 200
}

get_token_scopes <- function(endpoint, credentials) {
  validate_url <- endpoint_validation_url(endpoint, credentials)
  response <- GET(validate_url, accept_json())
  if (response$status_code == 200) {
    strsplit(content(response)$scope, " ")[[1]]
  }
}

endpoint_validation_url <- function(endpoint, credentials) {
  if (is.null(endpoint$validate)) {
    stop("No validation endpoint", call. = FALSE)
  }
  base_url <- endpoint$validate
  url <- parse_url(base_url)
  url$query$access_token <- credentials$access_token
  build_url(url)
}

parse_query <- function(query) {
  params <- vapply(strsplit(query, "&")[[1]], str_split_fixed, "=", 2,
    FUN.VALUE = character(2)
  )

  values <- as.list(curl::curl_unescape(params[2, ]))
  names(values) <- curl::curl_unescape(params[1, ])
  values
}

compose_query <- function(elements) {
  if (length(elements) == 0) {
    return("")
  }

  if (!all(has_name(elements))) {
    stop("All components of query must be named", call. = FALSE)
  }

  stopifnot(is.list(elements))
  elements <- compact(elements)

  names <- curl::curl_escape(names(elements))

  encode <- function(x) {
    if (inherits(x, "AsIs")) return(x)
    curl::curl_escape(x)
  }
  values <- vapply(elements, encode, character(1))

  paste0(names, "=", values, collapse = "&")
}

#' Parse and print http dates.
#'
#' @description
#' As defined in RFC2616,
#' <https://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.3>, there are
#' three valid formats:
#'
#' * Sun, 06 Nov 1994 08:49:37 GMT  ; RFC 822, updated by RFC 1123
#' * Sunday, 06-Nov-94 08:49:37 GMT ; RFC 850, obsoleted by RFC 1036
#' * Sun Nov  6 08:49:37 1994       ; ANSI C's asctime() format
#'
#' @param x For `parse_http_date`, a character vector of strings to parse.
#'   All elements must be of the same type.
#'
#'   For `http_date`, a `POSIXt` vector.
#' @param failure What to return on failure?
#' @export
#' @return A POSIXct object if succesful, otherwise `failure`
#' @examples
#' parse_http_date("Sun, 06 Nov 1994 08:49:37 GMT")
#' parse_http_date("Sunday, 06-Nov-94 08:49:37 GMT")
#' parse_http_date("Sun Nov  6 08:49:37 1994")
#'
#' http_date(Sys.time())
parse_http_date <- function(x, failure = structure(NA_real_, class = "Date")) {
  if (length(x) == 0) return(NULL)

  fmts <- c(
    "%a, %d %b %Y %H:%M:%S",
    "%A, %d-%b-%y %H:%M:%S",
    "%a %b %d %H:%M:%S %Y"
  )

  # Need to set C locale to ensure that months/days are parsed in English

  for (fmt in fmts) {
    parsed <- c_time(as.POSIXct(strptime(x, fmt, tz = "GMT")))
    if (all(!is.na(parsed))) return(parsed)
  }

  rep(failure, length(x))
}

#' @export
#' @rdname parse_http_date
http_date <- function(x) {
  if (is.null(x)) return(NULL)
  stopifnot(inherits(x, "POSIXt"))
  c_time(strftime(x, "%a, %d %b %Y %H:%M:%S", tz = "GMT", usetz = TRUE))
}

c_time <- function(code) {
  old <- Sys.getlocale(category = "LC_TIME")
  Sys.setlocale(category = "LC_TIME", locale = "C")
  on.exit(Sys.setlocale(category = "LC_TIME", locale = old))

  force(code)
}

#' Guess the media type of a path from its extension.
#'
#' DEPRECATED: please use `mime::guess_type` instead.
#'
#' @param x path to file
#' @keywords internal
#' @export
guess_media <- function(x) {
  .Deprecated("mime::guess_type")
  mime::guess_type(x, empty = NULL)
}

#' Give verbose output.
#'
#' A verbose connection provides much more information about the flow of
#' information between the client and server.
#'
#' @section Prefixes:
#'
#' `verbose()` uses the following prefixes to distinguish between
#' different components of the http messages:
#'
#' * `*` informative curl messages
#' * `->` headers sent (out)
#' * `>>` data sent (out)
#' * `*>` ssl data sent (out)
#' * `<-` headers received (in)
#' * `<<` data received (in)
#' * `<*` ssl data received (in)
#'
#' @family config
#' @param data_out Show data sent to the server.
#' @param info Show informational text from curl. This is mainly useful
#'   for debugging https and auth problems, so is disabled by default.
#' @param data_in Show data recieved from the server.
#' @param ssl Show even data sent/recieved over SSL connections?
#' @seealso [with_verbose()] makes it easier to use verbose mode
#'  even when the requests are buried inside another function call.
#' @export
#' @examples
#' \dontrun{
#' GET("http://httpbin.org", verbose())
#' GET("http://httpbin.org", verbose(info = TRUE))
#'
#' f <- function() {
#'   GET("http://httpbin.org")
#' }
#' with_verbose(f())
#' with_verbose(f(), info = TRUE)
#'
#' # verbose() makes it easy to see exactly what POST requests send
#' POST_verbose <- function(body, ...) {
#'   POST("https://httpbin.org/post", body = body, verbose(), ...)
#'   invisible()
#' }
#' POST_verbose(list(x = "a", y = "b"))
#' POST_verbose(list(x = "a", y = "b"), encode = "form")
#' POST_verbose(FALSE)
#' POST_verbose(NULL)
#' POST_verbose("")
#' POST_verbose("xyz")
#' }
verbose <- function(data_out = TRUE, data_in = FALSE, info = FALSE, ssl = FALSE) {
  debug <- function(type, msg) {
    switch(type + 1,
      text =       if (info)            prefix_message("*  ", msg),
      headerIn =                        prefix_message("<- ", msg),
      headerOut =                       prefix_message("-> ", msg),
      dataIn =     if (data_in)         prefix_message("<<  ", msg, TRUE),
      dataOut =    if (data_out)        prefix_message(">> ", msg, TRUE),
      sslDataIn =  if (ssl && data_in)  prefix_message("*< ", msg, TRUE),
      sslDataOut = if (ssl && data_out) prefix_message("*> ", msg, TRUE)
    )
  }
  config(debugfunction = debug, verbose = TRUE)
}

prefix_message <- function(prefix, x, blank_line = FALSE) {
  x <- readBin(x, character())

  lines <- unlist(strsplit(x, "\n", fixed = TRUE, useBytes = TRUE))
  out <- paste0(prefix, lines, collapse = "\n")
  message(out)
  if (blank_line) cat("\n")
}

#' Extract content from a request.
#'
#' There are currently three ways to retrieve the contents of a request:
#' as a raw object (`as = "raw"`), as a character vector,
#' (`as = "text"`), and as parsed into an R object where possible,
#' (`as = "parsed"`). If `as` is not specified, `content`
#' does its best to guess which output is most appropriate.
#'
#' `content` currently knows about the following mime types:
#'
#' * `text/html`: [xml2::read_html()]
#' * `text/xml`: [xml2::read_xml()]
#' * `text/csv`: [readr::read_csv()]
#' * `text/tab-separated-values`: [readr::read_tsv()]
#' * `application/json`: [jsonlite::fromJSON()]
#' * `application/x-www-form-urlencoded`: `parse_query`
#' * `image/jpeg`: [jpeg::readJPEG()]
#' * `image/png`: [png::readPNG()]
#'
#' `as = "parsed"` is provided as a convenience only: if the type you
#' are trying to parse is not available, use `as = "text"` and parse
#' yourself.
#'
#' @section WARNING:
#'
#' When using `content()` in a package, DO NOT use on `as = "parsed"`.
#' Instead, check the mime-type is what you expect, and then parse yourself.
#' This is safer, as you will fail informatively if the API changes, and
#' you will protect yourself against changes to httr.
#'
#' @param x request object
#' @param as desired type of output: `raw`, `text` or
#'   `parsed`. `content` attempts to automatically figure out
#'   which one is most appropriate, based on the content-type.
#' @param type MIME type (aka internet media type) used to override
#'   the content type returned by the server. See
#'   <http://en.wikipedia.org/wiki/Internet_media_type> for a list of
#'   common types.
#' @param encoding For text, overrides the charset or the Latin1 (ISO-8859-1)
#'   default, if you know that the server is returning the incorrect encoding
#'   as the charset in the content-type. Use for text and parsed outputs.
#' @param ... Other parameters parsed on to the parsing functions, if
#'  `as = "parsed"`
#' @family response methods
#' @return
#' For "raw", a raw vector.
#'
#' For "text", a character vector of length 1. The character vector is always
#' re-encoded to UTF-8. If this encoding fails (usually because the page
#' declares an incorrect encoding), `content()` will return `NA`.
#'
#' For "auto", a parsed R object.
#' @export
#' @examples
#' r <- POST("http://httpbin.org/post", body = list(a = 1, b = 2))
#' content(r) # automatically parses JSON
#' cat(content(r, "text"), "\n") # text content
#' content(r, "raw") # raw bytes from server
#' 
#' rlogo <- content(GET("http://cran.r-project.org/Rlogo.jpg"))
#' plot(0:1, 0:1, type = "n")
#' rasterImage(rlogo, 0, 0, 1, 1)
#' @aliases text_content parsed_content
content <- function(x, as = NULL, type = NULL, encoding = NULL, ...) {
  stopifnot(is.response(x))

  type <- type %||% x$headers[["Content-Type"]] %||%
    mime::guess_type(x$url, empty = "application/octet-stream")

  as <- as %||% parseability(type)
  as <- match.arg(as, c("raw", "text", "parsed"))

  if (is.path(x$content)) {
    raw <- readBin(x$content, "raw", file.info(x$content)$size)
  } else {
    raw <- x$content
  }

  switch(as,
    raw = raw,
    text = parse_text(raw, type, encoding),
    parsed = parse_auto(raw, type, encoding, ...)
  )
}

#' @export
text_content <- function(x) {
  message("text_content() deprecated. Use content(x, as = 'text')")
  content(x, as = "text")
}

#' @export
parsed_content <- function(x, ...) {
  message("text_content() deprecated. Use content(x, as = 'parsed')")
  content(x, as = "parsed", ...)
}

#' Does the request have content associated with it?
#'
#' @keywords internal
#' @export
#' @examples
#' has_content(POST("http://httpbin.org/post", body = FALSE))
#' has_content(HEAD("http://httpbin.org/headers"))
has_content <- function(x) {
  length(x$content) > 0
}

context("Content")

test_that("parse content is null if no body", {
  skip_httpbin()
  out <- content(HEAD("http://httpbin.org/headers"))
  expect_equal(out, NULL)
})


# has_content -------------------------------------------------------------

test_that("POST result with empty body doesn't have content", {
  skip_httpbin()
  r <- POST("http://httpbin.org/post", body = FALSE)
  expect_false(has_content(r))
})

test_that("HEAD requests don't have body", {
  skip_httpbin()
  r <- HEAD("http://httpbin.org/headers")
  expect_false(has_content(r))
})

test_that("regular POST does have content", {
  skip_httpbin()
  r <- POST("http://httpbin.org/post")
  expect_true(has_content(r))
})

#' Describe an OAuth endpoint.
#'
#' See [oauth_endpoints()] for a list of popular OAuth endpoints
#' baked into httr.
#'
#' @param request url used to request initial (unauthenticated) token.
#'   If using OAuth2.0, leave as `NULL`.
#' @param authorize url to send client to for authorisation. Set to `NULL`
#'   if not needed
#' @param access url used to exchange unauthenticated for authenticated token.
#' @param ... other additional endpoints.
#' @param base_url option url to use as base for `request`,
#'   `authorize` and `access` urls.
#' @family OAuth
#' @export
#' @examples
#' linkedin <- oauth_endpoint("requestToken", "authorize", "accessToken",
#'   base_url = "https://api.linkedin.com/uas/oauth"
#' )
#' github <- oauth_endpoint(NULL, "authorize", "access_token",
#'   base_url = "https://github.com/login/oauth"
#' )
#' facebook <- oauth_endpoint(
#'   authorize = "https://www.facebook.com/dialog/oauth",
#'   access = "https://graph.facebook.com/oauth/access_token"
#' )
#'
#' oauth_endpoints
oauth_endpoint <- function(request = NULL, authorize, access, ...,
                           base_url = NULL) {
  urls <- list(request = request, authorize = authorize, access = access, ...)

  if (is.null(base_url)) {
    return(do.call(endpoint, urls))
  }

  # If base_url provided, add it as a prefix
  path <- parse_url(base_url)$path
  add_base_url <- function(x) {
    if (is.null(x)) return(x)
    if (substr(x, 1, 4) == "http") return(x)
    modify_url(base_url, path = file.path(path, x))
  }
  urls <- lapply(urls, add_base_url)
  do.call(endpoint, urls)
}
endpoint <- function(request, authorize, access, ...) {
  structure(list(request = request, authorize = authorize, access = access, ...),
    class = "oauth_endpoint"
  )
}

is.oauth_endpoint <- function(x) inherits(x, "oauth_endpoint")

#' @export
print.oauth_endpoint <- function(x, ...) {
  x <- compact(x)
  cat("<oauth_endpoint>\n")
  cat(paste0(" ", format(paste0(names(x), ": ")), unlist(x), collapse = "\n"))
  cat("\n")
  invisible(x)
}

#' Popular oauth endpoints.
#'
#' Provides some common OAuth endpoints.
#'
#' @param name One of the following endpoints: linkedin, twitter,
#'   vimeo, google, facebook, github, azure.
#' @export
#' @examples
#' oauth_endpoints("twitter")
oauth_endpoints <- function(name) {
  switch(name,
    linkedin = oauth_endpoint(
      base_url = "https://www.linkedin.com/uas/oauth2",
      authorize = "authorization",
      access = "accessToken"
    ),
    twitter = oauth_endpoint(
      base_url = "https://api.twitter.com/oauth",
      request = "request_token",
      authorize = "authenticate",
      access = "access_token"
    ),
    vimeo = oauth_endpoint(
      base_url = "https://api.vimeo.com/oauth",
      request = "request_token",
      authorize = "authorize",
      access = "access_token"
    ),
    yahoo = oauth_endpoint(
      base_url = "https://api.login.yahoo.com/oauth2",
      request = "get_request_token",
      authorize = "request_auth",
      access = "get_token"
    ),
    google = oauth_endpoint(
      base_url = "https://accounts.google.com/o/oauth2",
      authorize = "auth",
      access = "token",
      validate = "https://www.googleapis.com/oauth2/v1/tokeninfo",
      revoke = "revoke"
    ),
    tumblr = oauth_endpoint(
      base_url = "http://www.tumblr.com/oauth/",
      request = "request_token",
      authorize = "authorize",
      access = "access_token"
    ),
    facebook = oauth_endpoint(
      authorize = "https://www.facebook.com/dialog/oauth",
      access = "https://graph.facebook.com/oauth/access_token"
    ),
    github = oauth_endpoint(
      base_url = "https://github.com/login/oauth",
      request = NULL,
      authorize = "authorize",
      access = "access_token"
    ),
    azure = oauth_endpoint(
      base_url = "https://login.windows.net/common/oauth2",
      authorize = "authorize",
      access = "token"
    ),
    meetup = oauth_endpoint(
      base_url = "https://secure.meetup.com/oauth2",
      authorize = "authorize",
      access = "access"
    ),
    stop("Unknown endpoint", call. = FALSE)
  )
}

#' Send PUT request to server.
#'
#' @inheritParams POST
#' @family http methods
#' @export
#' @examples
#' POST("http://httpbin.org/put")
#' PUT("http://httpbin.org/put")
#' 
#' b2 <- "http://httpbin.org/put"
#' PUT(b2, body = "A simple text string")
#' PUT(b2, body = list(x = "A simple text string"))
#' PUT(b2, body = list(y = upload_file(system.file("CITATION"))))
#' PUT(b2, body = list(x = "A simple text string"), encode = "json")
PUT <- function(url = NULL, config = list(), ...,
                body = NULL, encode = c("multipart", "form", "json", "raw"),
                handle = NULL) {
  hu <- handle_url(handle, url, ...)
  req <- request_build("PUT", hu$url, body_config(body, match.arg(encode)), config, ...)
  request_perform(req, hu$handle$handle)
}

#' Create a webserver to listen for OAuth callback.
#'
#' This opens a web browser pointing to `request_url`, and opens a
#' webserver on port 1410 to listen to the reponse.  The redirect url should
#' either be set previously (during the OAuth authentication dance) or
#' supplied as a parameter to the url.  See [oauth1.0_token()] and
#' [oauth2.0_token()] for examples of both techniques.
#'
#' This function should not normally be called directly by the user.
#'
#' @param request_url the url to send the browser to
#' @param is_interactive DEPRECATED
#' @param host ip address for the listener
#' @param port for the listener
#' @export
#' @keywords internal
oauth_listener <- function(request_url, is_interactive = interactive()) {
  if (!is_installed("httpuv")) {
    stop("httpuv package required to capture OAuth credentials.")
  }

  info <- NULL
  listen <- function(env) {
    if (!identical(env$PATH_INFO, "/")) {
      return(list(
        status = 404L,
        headers = list("Content-Type" = "text/plain"),
        body = "Not found"
      ))
    }

    query <- env$QUERY_STRING
    if (!is.character(query) || identical(query, "")) {
      info <<- NA
    } else {
      info <<- parse_query(gsub("^\\?", "", query))
    }

    list(
      status = 200L,
      headers = list("Content-Type" = "text/plain"),
      body = "Authentication complete. Please close this page and return to R."
    )
  }
  use <- listener_endpoint()
  server <- httpuv::startServer(use$host, use$port, list(call = listen))
  on.exit(httpuv::stopServer(server))

  message("Waiting for authentication in browser...")
  message("Press Esc/Ctrl + C to abort")
  BROWSE(request_url)
  while (is.null(info)) {
    httpuv::service()
    Sys.sleep(0.001)
  }
  httpuv::service() # to send text back to browser

  if (identical(info, NA)) {
    stop("Authentication failed.", call. = FALSE)
  }

  message("Authentication complete.")
  info
}

#' The oauth callback url.
#'
#' The url that [oauth_listener()] expects that the client be
#' referred to.
#'
#' @keywords internal
#' @export
oauth_callback <- function() {
  paste0(
    "http://",
    Sys.getenv("HTTR_SERVER", "localhost"),
    ":",
    Sys.getenv("HTTR_SERVER_PORT", "1410"),
    "/"
  )
}

listener_endpoint <- function() {
  list(
    host = Sys.getenv("HTTR_LOCALHOST", "127.0.0.1"),
    port = as.integer(Sys.getenv("HTTR_PORT", "1410"))
  )
}

---
title: "Best practices for API packages"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Best practices for API packages}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
library(httr)
knitr::opts_chunk$set(comment = "#>", collapse = TRUE)
```

httr is no longer recommended for new code: we recommend [httr2](https://httr2.r-lib.org).

---
title: "Getting started with httr"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with httr}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo = FALSE}
library(httr)
knitr::opts_chunk$set(comment = "#>", collapse = TRUE, eval = FALSE)
```

# httr quickstart guide

httr is no longer recommended for new code: we recommend [httr2](https://httr2.r-lib.org).

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-endpoint.r
\name{oauth_endpoint}
\alias{oauth_endpoint}
\title{Describe an OAuth endpoint.}
\usage{
oauth_endpoint(request = NULL, authorize, access, ..., base_url = NULL)
}
\arguments{
\item{request}{url used to request initial (unauthenticated) token.
If using OAuth2.0, leave as \code{NULL}.}

\item{authorize}{url to send client to for authorisation. Set to \code{NULL}
if not needed}

\item{access}{url used to exchange unauthenticated for authenticated token.}

\item{...}{other additional endpoints.}

\item{base_url}{option url to use as base for \code{request},
\code{authorize} and \code{access} urls.}
}
\description{
See \code{\link[=oauth_endpoints]{oauth_endpoints()}} for a list of popular OAuth endpoints
baked into httr.
}
\examples{
linkedin <- oauth_endpoint("requestToken", "authorize", "accessToken",
  base_url = "https://api.linkedin.com/uas/oauth"
)
github <- oauth_endpoint(NULL, "authorize", "access_token",
  base_url = "https://github.com/login/oauth"
)
facebook <- oauth_endpoint(
  authorize = "https://www.facebook.com/dialog/oauth",
  access = "https://graph.facebook.com/oauth/access_token"
)

oauth_endpoints
}
\seealso{
Other OAuth: 
\code{\link{oauth1.0_token}()},
\code{\link{oauth2.0_token}()},
\code{\link{oauth_app}()},
\code{\link{oauth_service_token}()}
}
\concept{OAuth}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callback.R
\name{get_callback}
\alias{get_callback}
\alias{set_callback}
\title{Install or uninstall a callback function}
\usage{
get_callback(name)

set_callback(name, new_callback = NULL)
}
\arguments{
\item{name}{Character scalar, name of the callback to query or set.}

\item{new_callback}{The callback function to install, a function object;
or \code{NULL} to remove the currently installed callback (if any).}
}
\value{
\code{get_callback} returns the currently installed
callback, or \code{NULL} if none is installed.

\code{set_callback} returns the previously installed callback,
or \code{NULL} if none was installed.
}
\description{
Supported callback functions: \describe{
\item{\sQuote{request}}{This callback is called before an HTTP request
is performed, with the \code{request} object as an argument.
If the callback returns a value other than \code{NULL}, the HTTP
request is not performed at all, and the return value of the callback
is returned. This mechanism can be used to replay previously
recorded HTTP responses.
}
\item{\sQuote{response}}{This callback is called after an HTTP request
is performed. The callback is called with two arguments: the
\code{request} object and the \code{response} object of the HTTP
request. If this callback returns a value other than \code{NULL},
then this value is returned by \code{httr}.}
}
}
\details{
Note that it is not possible to install multiple callbacks of the same
type. The installed callback overwrites the previously intalled one.
To uninstall a callback function, set it to \code{NULL} with
\code{set_callback()}.

See the \code{httrmock} package for a proper example that uses
callbacks.
}
\examples{
\dontrun{
## Log all HTTP requests to the screeen
req_logger <- function(req) {
  cat("HTTP request to", sQuote(req$url), "\n")
}

old <- set_callback("request", req_logger)
g1 <- GET("https://httpbin.org")
g2 <- GET("https://httpbin.org/ip")
set_callback("request", old)

## Log all HTTP requests and response status codes as well
req_logger2 <- function(req) {
  cat("HTTP request to", sQuote(req$url), "... ")
}
res_logger <- function(req, res) {
  cat(res$status_code, "\n")
}

old_req <- set_callback("request", req_logger2)
old_res <- set_callback("response", res_logger)
g3 <- GET("https://httpbin.org")
g4 <- GET("https://httpbin.org/ip")
set_callback("request", old_req)
set_callback("response", old_res)

## Return a recorded response, without performing the HTTP request
replay <- function(req) {
  if (req$url == "https://httpbin.org") g3
}
old_req <- set_callback("request", replay)
grec <- GET("https://httpbin.org")
grec$date == g3$date
set_callback("request", old_req)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insensitive.r
\name{insensitive}
\alias{insensitive}
\title{Create a vector with case insensitive name matching.}
\usage{
insensitive(x)
}
\arguments{
\item{x}{vector to modify}
}
\description{
Create a vector with case insensitive name matching.
}
\examples{
x <- c("abc" = 1, "def" = 2)
x["ABC"]
y <- insensitive(x)
y["ABC"]
y[["ABC"]]
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-token.r
\docType{class}
\name{Token-class}
\alias{Token-class}
\alias{Token}
\alias{Token1.0}
\alias{Token2.0}
\alias{TokenServiceAccount}
\title{OAuth token objects.}
\format{
An R6 class object.
}
\description{
These objects represent the complete set of data needed for OAuth access:
an app, an endpoint, cached credentials and parameters. They should be
created through their constructor functions \code{\link[=oauth1.0_token]{oauth1.0_token()}}
and \code{\link[=oauth2.0_token]{oauth2.0_token()}}.
}
\section{Methods}{

\itemize{
\item \code{cache()}: caches token to disk
\item \code{sign(method, url)}: returns list of url and config
\item \code{refresh()}: refresh access token (if possible)
\item \code{validate()}: TRUE if the token is still valid, FALSE otherwise
}
}

\section{Caching}{

OAuth tokens are cached on disk in a file called \code{.httr-oauth}
saved in the current working directory.  Caching is enabled if:
\itemize{
\item The session is interactive, and the user agrees to it, OR
\item The \code{.httr-oauth} file is already present, OR
\item \code{getOption("httr_oauth_cache")} is \code{TRUE}
}

You can suppress caching by setting the \code{httr_oauth_cache} option to
\code{FALSE}.

Tokens are cached based on their endpoint and parameters.

The cache file should not be included in source code control or R packages
(because it contains private information), so httr will automatically add
the appropriate entries to \code{.gitignore} and \code{.Rbuildignore} if needed.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/retry.R
\name{RETRY}
\alias{RETRY}
\title{Retry a request until it succeeds.}
\usage{
RETRY(
  verb,
  url = NULL,
  config = list(),
  ...,
  body = NULL,
  encode = c("multipart", "form", "json", "raw"),
  times = 3,
  pause_base = 1,
  pause_cap = 60,
  pause_min = 1,
  handle = NULL,
  quiet = FALSE,
  terminate_on = NULL,
  terminate_on_success = TRUE
)
}
\arguments{
\item{verb}{Name of verb to use.}

\item{url}{the url of the page to retrieve}

\item{config}{Additional configuration settings such as http
authentication (\code{\link[=authenticate]{authenticate()}}), additional headers
(\code{\link[=add_headers]{add_headers()}}), cookies (\code{\link[=set_cookies]{set_cookies()}}) etc.
See \code{\link[=config]{config()}} for full details and list of helpers.}

\item{...}{Further named parameters, such as \code{query}, \code{path}, etc,
passed on to \code{\link[=modify_url]{modify_url()}}. Unnamed parameters will be combined
with \code{\link[=config]{config()}}.}

\item{body}{One of the following:
\itemize{
\item \code{FALSE}: No body. This is typically not used with \code{POST},
\code{PUT}, or \code{PATCH}, but can be useful if you need to send a
bodyless request (like \code{GET}) with \code{VERB()}.
\item \code{NULL}: An empty body
\item \code{""}: A length 0 body
\item \code{upload_file("path/")}: The contents of a file.  The mime
type will be guessed from the extension, or can be supplied explicitly
as the second argument to \code{upload_file()}
\item A character or raw vector: sent as is in body. Use
\code{\link[=content_type]{content_type()}} to tell the server what sort of data
you are sending.
\item A named list: See details for encode.
}}

\item{encode}{If the body is a named list, how should it be encoded? Can be
one of form (application/x-www-form-urlencoded), multipart,
(multipart/form-data), or json (application/json).

For "multipart", list elements can be strings or objects created by
\code{\link[=upload_file]{upload_file()}}. For "form", elements are coerced to strings
and escaped, use \code{I()} to prevent double-escaping. For "json",
parameters are automatically "unboxed" (i.e. length 1 vectors are
converted to scalars). To preserve a length 1 vector as a vector,
wrap in \code{I()}. For "raw", either a character or raw vector. You'll
need to make sure to set the \code{\link[=content_type]{content_type()}} yourself.}

\item{times}{Maximum number of requests to attempt.}

\item{pause_base, pause_cap}{This method uses exponential back-off with
full jitter - this means that each request will randomly wait between 0
and \code{pause_base * 2 ^ attempt} seconds, up to a maximum of
\code{pause_cap} seconds.}

\item{pause_min}{Minimum time to wait in the backoff; generally
only necessary if you need pauses less than one second (which may
not be kind to the server, use with caution!).}

\item{handle}{The handle to use with this request. If not
supplied, will be retrieved and reused from the \code{\link[=handle_pool]{handle_pool()}}
based on the scheme, hostname and port of the url. By default \pkg{httr}
requests to the same scheme/host/port combo. This substantially reduces
connection time, and ensures that cookies are maintained over multiple
requests to the same host. See \code{\link[=handle_pool]{handle_pool()}} for more
details.}

\item{quiet}{If \code{FALSE}, will print a message displaying how long
until the next request.}

\item{terminate_on}{Optional vector of numeric HTTP status codes that if found
on the response will terminate the retry process. If \code{NULL}, will keep
retrying while \code{\link[=http_error]{http_error()}} is \code{TRUE} for the response.}

\item{terminate_on_success}{If \code{TRUE}, the default, this will
automatically terminate when the request is successful, regardless of the
value of \code{terminate_on}.}
}
\value{
The last response. Note that if the request doesn't succeed after
\code{times} times this will be a failed request, i.e. you still need
to use \code{\link[=stop_for_status]{stop_for_status()}}.
}
\description{
Safely retry a request until it succeeds, as defined by the \code{terminate_on}
parameter, which by default means a response for which \code{\link[=http_error]{http_error()}}
is \code{FALSE}. Will also retry on error conditions raised by the underlying curl code,
but if the last retry still raises one, \code{RETRY} will raise it again with
\code{\link[=stop]{stop()}}.
It is designed to be kind to the server: after each failure
randomly waits up to twice as long. (Technically it uses exponential
backoff with jitter, using the approach outlined in
\url{https://www.awsarchitectureblog.com/2015/03/backoff.html}.)
If the server returns status code 429 and specifies a \code{retry-after} value, that
value will be used instead, unless it's smaller than \code{pause_min}.
}
\examples{
# Succeeds straight away
RETRY("GET", "http://httpbin.org/status/200")
# Never succeeds
RETRY("GET", "http://httpbin.org/status/500")
\dontrun{
# Invalid hostname generates curl error condition and is retried but eventually
# raises an error condition.
RETRY("GET", "http://invalidhostname/")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/http-browse.r
\name{BROWSE}
\alias{BROWSE}
\title{Open specified url in browser.}
\usage{
BROWSE(url = NULL, config = list(), ..., handle = NULL)
}
\arguments{
\item{url}{the url of the page to retrieve}

\item{config}{All configuration options are ignored because the request
is handled by the browser, not \pkg{RCurl}.}

\item{...}{Further named parameters, such as \code{query}, \code{path}, etc,
passed on to \code{\link[=modify_url]{modify_url()}}. Unnamed parameters will be combined
with \code{\link[=config]{config()}}.}

\item{handle}{The handle to use with this request. If not
supplied, will be retrieved and reused from the \code{\link[=handle_pool]{handle_pool()}}
based on the scheme, hostname and port of the url. By default \pkg{httr}
requests to the same scheme/host/port combo. This substantially reduces
connection time, and ensures that cookies are maintained over multiple
requests to the same host. See \code{\link[=handle_pool]{handle_pool()}} for more
details.}
}
\value{
A \code{\link[=response]{response()}} object.
}
\description{
(This isn't really a http verb, but it seems to follow the same format).
}
\details{
Only works in interactive sessions.
}
\examples{
BROWSE("http://google.com")
BROWSE("http://had.co.nz")
}
\seealso{
Other http methods: 
\code{\link{DELETE}()},
\code{\link{GET}()},
\code{\link{HEAD}()},
\code{\link{PATCH}()},
\code{\link{POST}()},
\code{\link{PUT}()},
\code{\link{VERB}()}
}
\concept{http methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response-type.R
\name{http_type}
\alias{http_type}
\title{Extract the content type of a response}
\usage{
http_type(x)
}
\arguments{
\item{x}{A response}
}
\value{
A string giving the complete mime type, with all parameters
stripped off.
}
\description{
Extract the content type of a response
}
\examples{
r1 <- GET("http://httpbin.org/image/png")
http_type(r1)
headers(r1)[["Content-Type"]]

r2 <- GET("http://httpbin.org/ip")
http_type(r2)
headers(r2)[["Content-Type"]]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response-status.r
\name{status_code}
\alias{status_code}
\title{Extract status code from response.}
\usage{
status_code(x)
}
\arguments{
\item{x}{A response}
}
\description{
Extract status code from response.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.r
\name{config}
\alias{config}
\title{Set curl options.}
\usage{
config(..., token = NULL)
}
\arguments{
\item{...}{named Curl options.}

\item{token}{An OAuth token (1.0 or 2.0)}
}
\description{
Generally you should only need to use this function to set CURL options
directly if there isn't already a helpful wrapper function, like
\code{\link[=set_cookies]{set_cookies()}}, \code{\link[=add_headers]{add_headers()}} or
\code{\link[=authenticate]{authenticate()}}. To use this function effectively requires
some knowledge of CURL, and CURL options. Use \code{\link[=httr_options]{httr_options()}} to
see a complete list of available options. To see the libcurl documentation
for a given option, use \code{\link[=curl_docs]{curl_docs()}}.
}
\details{
Unlike Curl (and RCurl), all configuration options are per request, not
per handle.
}
\examples{
# There are a number of ways to modify the configuration of a request
# * you can add directly to a request
HEAD("https://www.google.com", verbose())

# * you can wrap with with_config()
with_config(verbose(), HEAD("https://www.google.com"))

# * you can set global with set_config()
old <- set_config(verbose())
HEAD("https://www.google.com")
# and re-establish the previous settings with
set_config(old, override = TRUE)
HEAD("https://www.google.com")
# or
reset_config()
HEAD("https://www.google.com")

# If available, you should use a friendly httr wrapper over RCurl
# options. But you can pass Curl options (as listed in httr_options())
# in config
HEAD("https://www.google.com/", config(verbose = TRUE))
}
\seealso{
\code{\link[=set_config]{set_config()}} to set global config defaults, and
\code{\link[=with_config]{with_config()}} to temporarily run code with set options.

All known available options are listed in \code{\link[=httr_options]{httr_options()}}

Other config: 
\code{\link{add_headers}()},
\code{\link{authenticate}()},
\code{\link{set_cookies}()},
\code{\link{timeout}()},
\code{\link{use_proxy}()},
\code{\link{user_agent}()},
\code{\link{verbose}()}

Other ways to set configuration: 
\code{\link{set_config}()},
\code{\link{with_config}()}
}
\concept{config}
\concept{ways to set configuration}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.r
\name{use_proxy}
\alias{use_proxy}
\title{Use a proxy to connect to the internet.}
\usage{
use_proxy(url, port = NULL, username = NULL, password = NULL, auth = "basic")
}
\arguments{
\item{url, port}{location of proxy}

\item{username, password}{login details for proxy, if needed}

\item{auth}{type of HTTP authentication to use. Should be one of the
following: basic, digest, digest_ie, gssnegotiate, ntlm, any.}
}
\description{
Use a proxy to connect to the internet.
}
\examples{
# See http://www.hidemyass.com/proxy-list for a list of public proxies
# to test with
# GET("http://had.co.nz", use_proxy("64.251.21.73", 8080), verbose())
}
\seealso{
Other config: 
\code{\link{add_headers}()},
\code{\link{authenticate}()},
\code{\link{config}()},
\code{\link{set_cookies}()},
\code{\link{timeout}()},
\code{\link{user_agent}()},
\code{\link{verbose}()}
}
\concept{config}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response-status.r
\name{stop_for_status}
\alias{stop_for_status}
\alias{warn_for_status}
\alias{message_for_status}
\title{Take action on http error.}
\usage{
stop_for_status(x, task = NULL)

warn_for_status(x, task = NULL)

message_for_status(x, task = NULL)
}
\arguments{
\item{x}{a response, or numeric http code (or other object with
\code{status_code} method)}

\item{task}{The text of the message: either \code{NULL} or a
character vector. If non-\code{NULL}, the error message will finish with
"Failed to \code{task}".}
}
\value{
If request was successful, the response (invisibly). Otherwise,
raised a classed http error or warning, as generated by
\code{\link[=http_condition]{http_condition()}}
}
\description{
Converts http errors to R errors or warnings - these should always
be used whenever you're creating requests inside a function, so
that the user knows why a request has failed.
}
\examples{
x <- GET("http://httpbin.org/status/200")
stop_for_status(x) # nothing happens
warn_for_status(x)
message_for_status(x)

x <- GET("http://httpbin.org/status/300")
\dontrun{
stop_for_status(x)
}
warn_for_status(x)
message_for_status(x)

x <- GET("http://httpbin.org/status/404")
\dontrun{
stop_for_status(x)
}
warn_for_status(x)
message_for_status(x)

# You can provide more information with the task argument
warn_for_status(x, "download spreadsheet")
message_for_status(x, "download spreadsheet")
}
\seealso{
\code{\link[=http_status]{http_status()}} and
\verb{http://en.wikipedia.org/wiki/Http_status_codes} for more information
on http status codes.

Other response methods: 
\code{\link{content}()},
\code{\link{http_error}()},
\code{\link{http_status}()},
\code{\link{response}()}
}
\concept{response methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-token.r
\name{oauth1.0_token}
\alias{oauth1.0_token}
\title{Generate an oauth1.0 token.}
\usage{
oauth1.0_token(
  endpoint,
  app,
  permission = NULL,
  as_header = TRUE,
  private_key = NULL,
  cache = getOption("httr_oauth_cache")
)
}
\arguments{
\item{endpoint}{An OAuth endpoint, created by \code{\link[=oauth_endpoint]{oauth_endpoint()}}}

\item{app}{An OAuth consumer application, created by
\code{\link[=oauth_app]{oauth_app()}}}

\item{permission}{optional, a string of permissions to ask for.}

\item{as_header}{If \code{TRUE}, the default, sends oauth in header.
If \code{FALSE}, adds as parameter to url.}

\item{private_key}{Optional, a key provided by \code{\link[openssl:read_key]{openssl::read_key()}}.
Used for signed OAuth 1.0.}

\item{cache}{A logical value or a string. \code{TRUE} means to cache
using the default cache file \code{.httr-oauth}, \code{FALSE} means
don't cache, and \code{NA} means to guess using some sensible heuristics.
A string means use the specified path as the cache file.}
}
\value{
A \code{Token1.0} reference class (RC) object.
}
\description{
This is the final object in the OAuth dance - it encapsulates the app,
the endpoint, other parameters and the received credentials.
}
\details{
See \code{\link[=Token]{Token()}} for full details about the token object, and the
caching policies used to store credentials across sessions.
}
\seealso{
Other OAuth: 
\code{\link{oauth2.0_token}()},
\code{\link{oauth_app}()},
\code{\link{oauth_endpoint}()},
\code{\link{oauth_service_token}()}
}
\concept{OAuth}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response.r
\name{response}
\alias{response}
\title{The response object.}
\description{
The response object captures all information from a request.  It includes
fields:
\itemize{
\item \code{url} the url the request was actually sent to (after redirects)
\item \code{handle} the handle associated with the url
\item \code{status_code} the http status code
\item \code{header} a named list of headers returned by the server
\item \code{cookies} a named list of cookies returned by the server
\item \code{content} the body of the response, as raw vector. See \code{\link[=content]{content()}} for various ways to access the content.
\item \code{time} request timing information
\item \code{config} configuration for the request
}
}
\details{
For non-http(s) responses, some parts including the status and
header may not be interpretable the same way as http responses.
}
\seealso{
Other response methods: 
\code{\link{content}()},
\code{\link{http_error}()},
\code{\link{http_status}()},
\code{\link{stop_for_status}()}
}
\concept{response methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbose.r
\name{verbose}
\alias{verbose}
\title{Give verbose output.}
\usage{
verbose(data_out = TRUE, data_in = FALSE, info = FALSE, ssl = FALSE)
}
\arguments{
\item{data_out}{Show data sent to the server.}

\item{data_in}{Show data recieved from the server.}

\item{info}{Show informational text from curl. This is mainly useful
for debugging https and auth problems, so is disabled by default.}

\item{ssl}{Show even data sent/recieved over SSL connections?}
}
\description{
A verbose connection provides much more information about the flow of
information between the client and server.
}
\section{Prefixes}{


\code{verbose()} uses the following prefixes to distinguish between
different components of the http messages:
\itemize{
\item \code{*} informative curl messages
\item \verb{->} headers sent (out)
\item \verb{>>} data sent (out)
\item \verb{*>} ssl data sent (out)
\item \verb{<-} headers received (in)
\item \verb{<<} data received (in)
\item \verb{<*} ssl data received (in)
}
}

\examples{
GET("http://httpbin.org", verbose())
GET("http://httpbin.org", verbose(info = TRUE))

f <- function() {
  GET("http://httpbin.org")
}
with_verbose(f())
with_verbose(f(), info = TRUE)

# verbose() makes it easy to see exactly what POST requests send
POST_verbose <- function(body, ...) {
  POST("https://httpbin.org/post", body = body, verbose(), ...)
  invisible()
}
POST_verbose(list(x = "a", y = "b"))
POST_verbose(list(x = "a", y = "b"), encode = "form")
POST_verbose(FALSE)
POST_verbose(NULL)
POST_verbose("")
POST_verbose("xyz")
}
\seealso{
\code{\link[=with_verbose]{with_verbose()}} makes it easier to use verbose mode
even when the requests are buried inside another function call.

Other config: 
\code{\link{add_headers}()},
\code{\link{authenticate}()},
\code{\link{config}()},
\code{\link{set_cookies}()},
\code{\link{timeout}()},
\code{\link{use_proxy}()},
\code{\link{user_agent}()}
}
\concept{config}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/authenticate.r
\name{authenticate}
\alias{authenticate}
\title{Use http authentication.}
\usage{
authenticate(user, password, type = "basic")
}
\arguments{
\item{user}{user name}

\item{password}{password}

\item{type}{type of HTTP authentication.  Should be one of the following
types supported by Curl: basic, digest, digest_ie, gssnegotiate,
ntlm, any.  It defaults to "basic", the most common type.}
}
\description{
It's not obvious how to turn authentication off after using it, so
I recommend using custom handles with authentication.
}
\examples{
\dontrun{
GET("http://httpbin.org/basic-auth/user/passwd")
GET(
  "http://httpbin.org/basic-auth/user/passwd",
  authenticate("user", "passwd")
)
}
}
\seealso{
Other config: 
\code{\link{add_headers}()},
\code{\link{config}()},
\code{\link{set_cookies}()},
\code{\link{timeout}()},
\code{\link{use_proxy}()},
\code{\link{user_agent}()},
\code{\link{verbose}()}
}
\concept{config}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-exchanger.r
\name{oauth_exchanger}
\alias{oauth_exchanger}
\title{Walk the user through the OAuth2 dance without a local webserver.}
\usage{
oauth_exchanger(request_url)
}
\arguments{
\item{request_url}{the url to provide to the user}
}
\description{
This performs a similar function to \code{\link[=oauth_listener]{oauth_listener()}},
but without running a local webserver.  This manual process can be useful
in situations where the user is remotely accessing the machine outside a
browser (say via ssh) or when it's not possible to successfully receive a
callback (such as when behind a firewall).
}
\details{
This function should generally not be called directly by the user.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/headers.r
\name{add_headers}
\alias{add_headers}
\title{Add additional headers to a request.}
\usage{
add_headers(..., .headers = character())
}
\arguments{
\item{...}{named header values.  To stop an existing header from being
set, pass an empty string: \code{""}.}

\item{.headers}{a named character vector}
}
\description{
Wikipedia provides a useful list of common http headers:
\url{https://en.wikipedia.org/wiki/List_of_HTTP_header_fields}.
}
\examples{
add_headers(a = 1, b = 2)
add_headers(.headers = c(a = "1", b = "2"))

\dontrun{
GET("http://httpbin.org/headers")

# Add arbitrary headers
GET(
  "http://httpbin.org/headers",
  add_headers(version = version$version.string)
)

# Override default headers with empty strings
GET("http://httpbin.org/headers", add_headers(Accept = ""))
}
}
\seealso{
\code{\link[=accept]{accept()}} and \code{\link[=content_type]{content_type()}} for
convenience functions for setting accept and content-type headers.

Other config: 
\code{\link{authenticate}()},
\code{\link{config}()},
\code{\link{set_cookies}()},
\code{\link{timeout}()},
\code{\link{use_proxy}()},
\code{\link{user_agent}()},
\code{\link{verbose}()}
}
\concept{config}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write-function.R
\name{write_function}
\alias{write_function}
\title{S3 object to define response writer.}
\usage{
write_function(subclass, ...)
}
\arguments{
\item{subclass, ...}{Class name and fields. Used in class constructors.}

\item{x}{A \code{write_function} object to process.}
}
\description{
This S3 object allows you to control how the response body is saved.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.r
\name{content}
\alias{content}
\alias{text_content}
\alias{parsed_content}
\title{Extract content from a request.}
\usage{
content(x, as = NULL, type = NULL, encoding = NULL, ...)
}
\arguments{
\item{x}{request object}

\item{as}{desired type of output: \code{raw}, \code{text} or
\code{parsed}. \code{content} attempts to automatically figure out
which one is most appropriate, based on the content-type.}

\item{type}{MIME type (aka internet media type) used to override
the content type returned by the server. See
\url{http://en.wikipedia.org/wiki/Internet_media_type} for a list of
common types.}

\item{encoding}{For text, overrides the charset or the Latin1 (ISO-8859-1)
default, if you know that the server is returning the incorrect encoding
as the charset in the content-type. Use for text and parsed outputs.}

\item{...}{Other parameters parsed on to the parsing functions, if
\code{as = "parsed"}}
}
\value{
For "raw", a raw vector.

For "text", a character vector of length 1. The character vector is always
re-encoded to UTF-8. If this encoding fails (usually because the page
declares an incorrect encoding), \code{content()} will return \code{NA}.

For "auto", a parsed R object.
}
\description{
There are currently three ways to retrieve the contents of a request:
as a raw object (\code{as = "raw"}), as a character vector,
(\code{as = "text"}), and as parsed into an R object where possible,
(\code{as = "parsed"}). If \code{as} is not specified, \code{content}
does its best to guess which output is most appropriate.
}
\details{
\code{content} currently knows about the following mime types:
\itemize{
\item \code{text/html}: \code{\link[xml2:read_xml]{xml2::read_html()}}
\item \code{text/xml}: \code{\link[xml2:read_xml]{xml2::read_xml()}}
\item \code{text/csv}: \code{\link[readr:read_delim]{readr::read_csv()}}
\item \code{text/tab-separated-values}: \code{\link[readr:read_delim]{readr::read_tsv()}}
\item \code{application/json}: \code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}}
\item \code{application/x-www-form-urlencoded}: \code{parse_query}
\item \code{image/jpeg}: \code{\link[jpeg:readJPEG]{jpeg::readJPEG()}}
\item \code{image/png}: \code{\link[png:readPNG]{png::readPNG()}}
}

\code{as = "parsed"} is provided as a convenience only: if the type you
are trying to parse is not available, use \code{as = "text"} and parse
yourself.
}
\section{WARNING}{


When using \code{content()} in a package, DO NOT use on \code{as = "parsed"}.
Instead, check the mime-type is what you expect, and then parse yourself.
This is safer, as you will fail informatively if the API changes, and
you will protect yourself against changes to httr.
}

\examples{
r <- POST("http://httpbin.org/post", body = list(a = 1, b = 2))
content(r) # automatically parses JSON
cat(content(r, "text"), "\n") # text content
content(r, "raw") # raw bytes from server

rlogo <- content(GET("http://cran.r-project.org/Rlogo.jpg"))
plot(0:1, 0:1, type = "n")
rasterImage(rlogo, 0, 0, 1, 1)
}
\seealso{
Other response methods: 
\code{\link{http_error}()},
\code{\link{http_status}()},
\code{\link{response}()},
\code{\link{stop_for_status}()}
}
\concept{response methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sha1.r
\name{sha1_hash}
\alias{sha1_hash}
\title{SHA1 hash}
\usage{
sha1_hash(key, string, method = "HMAC-SHA1")
}
\arguments{
\item{key}{The key to create the hash with}

\item{string}{data to securely hash}

\item{method}{The method to use, either HMAC-SHA1 or RSA-SHA1}
}
\description{
Creates a SHA1 hash of data using either HMAC or RSA.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/doctor.R
\name{httr_dr}
\alias{httr_dr}
\title{Diagnose common configuration problems}
\usage{
httr_dr()
}
\description{
Currently one check: that curl uses nss.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-listener.r
\name{oauth_callback}
\alias{oauth_callback}
\title{The oauth callback url.}
\usage{
oauth_callback()
}
\description{
The url that \code{\link[=oauth_listener]{oauth_listener()}} expects that the client be
referred to.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-server-side.R
\name{jwt_signature}
\alias{jwt_signature}
\title{Generate a JWT signature given credentials.}
\usage{
jwt_signature(
  credentials,
  scope,
  aud,
  sub = NULL,
  iat = as.integer(Sys.time()),
  exp = iat + duration,
  duration = 60L * 60L
)
}
\arguments{
\item{credentials}{Parsed contents of the credentials file.}

\item{scope}{A space-delimited list of the permissions that the application
requests.}

\item{aud}{A descriptor of the intended target of the assertion. This
typically comes from the service auth file.}

\item{sub}{The email address of the user for which the application is
requesting delegated access.}

\item{iat}{The time the assertion was issued, measured in seconds since
00:00:00 UTC, January 1, 1970.}

\item{exp}{The expiration time of the assertion, measured in seconds since
00:00:00 UTC, January 1, 1970. This value has a maximum of 1 hour from
the issued time.}

\item{duration}{Duration of token, in seconds.}

\item{iss}{Email address of the client_id of the application making the
access token request.}
}
\description{
As described in
\url{https://developers.google.com/accounts/docs/OAuth2ServiceAccount}
}
\examples{
\dontrun{
cred <- jsonlite::fromJSON("~/Desktop/httrtest-45693cbfac92.json")
jwt_signature(cred, "https://www.googleapis.com/auth/userinfo.profile")
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response-status.r
\name{http_condition}
\alias{http_condition}
\title{Generate a classed http condition.}
\usage{
http_condition(x, type, task = NULL, call = sys.call(-1))
}
\arguments{
\item{x}{a response, or numeric http code (or other object with
\code{status_code} method)}

\item{type}{type of condition to generate. Must be one of error,
warning or message.}

\item{task}{The text of the message: either \code{NULL} or a
character vector. If non-\code{NULL}, the error message will finish with
"Failed to \code{task}".}

\item{call}{The call stored in the condition object.}
}
\value{
An S3 object that inherits from (e.g.) condition, \code{type},
http_error, http_400 and http_404.
}
\description{
This function generate S3 condition objects which are passed to
\code{\link[=stop]{stop()}} or \code{\link[=warning]{warning()}} to generate classes warnings
and error. These can be used in conjunction with \code{\link[=tryCatch]{tryCatch()}}
to respond differently to different type of failure.
}
\examples{
# You can use tryCatch to take different actions based on the type
# of error. Note that tryCatch will call the first handler that
# matches any classes of the condition, not the best matching, so
# always list handlers from most specific to least specific
f <- function(url) {
  tryCatch(stop_for_status(GET(url)),
    http_404 = function(c) "That url doesn't exist",
    http_403 = function(c) "You need to authenticate!",
    http_400 = function(c) "You made a mistake!",
    http_500 = function(c) "The server screwed up"
  )
}
f("http://httpbin.org/status/404")
f("http://httpbin.org/status/403")
f("http://httpbin.org/status/505")
}
\seealso{
\url{http://adv-r.had.co.nz/Exceptions-Debugging.html#condition-handling}
for more details about R's condition handling model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.r
\name{httr_options}
\alias{httr_options}
\alias{curl_docs}
\title{List available options.}
\usage{
httr_options(matches)

curl_docs(x)
}
\arguments{
\item{matches}{If not missing, this restricts the output so that either
the httr or curl option matches this regular expression.}

\item{x}{An option name (either short or full).}
}
\value{
A data frame with three columns:
\item{httr}{The short name used in httr}
\item{libcurl}{The full name used by libcurl}
\item{type}{The type of R object that the option accepts}
}
\description{
This function lists all available options for \code{\link[=config]{config()}}.
It provides both the short R name which you use with httr, and the longer
Curl name, which is useful when searching the documentation. \code{curl_doc}
opens a link to the libcurl documentation for an option in your browser.
}
\details{
RCurl and httr use slightly different names to libcurl: the initial
\code{CURLOPT_} is removed, all underscores are converted to periods and
the option is given in lower case.  Thus "CURLOPT_SSLENGINE_DEFAULT"
becomes "sslengine.default".
}
\examples{
httr_options()
httr_options("post")

# Use curl_docs to read the curl documentation for each option.
# You can use either the httr or curl option name.
curl_docs("userpwd")
curl_docs("CURLOPT_USERPWD")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-token.r
\name{oauth2.0_token}
\alias{oauth2.0_token}
\title{Generate an oauth2.0 token.}
\usage{
oauth2.0_token(
  endpoint,
  app,
  scope = NULL,
  user_params = NULL,
  type = NULL,
  use_oob = getOption("httr_oob_default"),
  oob_value = NULL,
  as_header = TRUE,
  use_basic_auth = FALSE,
  cache = getOption("httr_oauth_cache"),
  config_init = list(),
  client_credentials = FALSE,
  credentials = NULL,
  query_authorize_extra = list()
)
}
\arguments{
\item{endpoint}{An OAuth endpoint, created by \code{\link[=oauth_endpoint]{oauth_endpoint()}}}

\item{app}{An OAuth consumer application, created by
\code{\link[=oauth_app]{oauth_app()}}}

\item{scope}{a character vector of scopes to request.}

\item{user_params}{Named list holding endpoint specific parameters to pass to
the server when posting the request for obtaining or refreshing the
access token.}

\item{type}{content type used to override incorrect server response}

\item{use_oob}{if FALSE, use a local webserver for the OAuth dance.
Otherwise, provide a URL to the user and prompt for a validation
code. Defaults to the of the \code{"httr_oob_default"} default,
or \code{TRUE} if \code{httpuv} is not installed.}

\item{oob_value}{if provided, specifies the value to use for the redirect_uri
parameter when retrieving an authorization URL. Defaults to "urn:ietf:wg:oauth:2.0:oob".
Requires \code{use_oob = TRUE}.}

\item{as_header}{If \code{TRUE}, the default, configures the token to add
itself to the bearer header of subsequent requests. If \code{FALSE},
configures the token to add itself as a url parameter of subsequent
requests.}

\item{use_basic_auth}{if \code{TRUE} use http basic authentication to
retrieve the token. Some authorization servers require this.
If \code{FALSE}, the default, retrieve the token by including the
app key and secret in the request body.}

\item{cache}{A logical value or a string. \code{TRUE} means to cache
using the default cache file \code{.httr-oauth}, \code{FALSE} means
don't cache, and \code{NA} means to guess using some sensible heuristics.
A string means use the specified path as the cache file.}

\item{config_init}{Additional configuration settings sent to
\code{\link[=POST]{POST()}}, e.g. \code{\link[=user_agent]{user_agent()}}.}

\item{client_credentials}{Default to \code{FALSE}. Set to \code{TRUE} to use
\emph{Client Credentials Grant} instead of \emph{Authorization
Code Grant}. See \url{https://www.rfc-editor.org/rfc/rfc6749#section-4.4}.}

\item{credentials}{Advanced use only: allows you to completely customise
token generation.}

\item{query_authorize_extra}{Default to \code{list()}. Set to named list
holding query parameters to append to initial auth page query. Useful for
some APIs.}
}
\value{
A \code{Token2.0} reference class (RC) object.
}
\description{
This is the final object in the OAuth dance - it encapsulates the app,
the endpoint, other parameters and the received credentials. It is a
reference class so that it can be seamlessly updated (e.g. using
\verb{$refresh()}) when access expires.
}
\details{
See \code{\link[=Token]{Token()}} for full details about the token object, and the
caching policies used to store credentials across sessions.
}
\seealso{
Other OAuth: 
\code{\link{oauth1.0_token}()},
\code{\link{oauth_app}()},
\code{\link{oauth_endpoint}()},
\code{\link{oauth_service_token}()}
}
\concept{OAuth}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{cache_info}
\alias{cache_info}
\alias{rerequest}
\title{Compute caching information for a response.}
\usage{
cache_info(r)

rerequest(r)
}
\arguments{
\item{r}{A response}
}
\description{
\code{cache_info()} gives details of cacheability of a response,
\code{rerequest()} re-performs the original request doing as little work
as possible (if not expired, returns response as is, or performs
revalidation if Etag or Last-Modified headers are present).
}
\examples{
# Never cached, always causes redownload
r1 <- GET("https://www.google.com")
cache_info(r1)
r1$date
rerequest(r1)$date

# Expires in a year
r2 <- GET("https://www.google.com/images/srpr/logo11w.png")
cache_info(r2)
r2$date
rerequest(r2)$date

\dontrun{
# Has last-modified and etag, so does revalidation
r3 <- GET("http://httpbin.org/cache")
cache_info(r3)
r3$date
rerequest(r3)$date

# Expires after 5 seconds
r4 <- GET("http://httpbin.org/cache/5")
cache_info(r4)
r4$date
rerequest(r4)$date
Sys.sleep(5)
cache_info(r4)
rerequest(r4)$date
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-init.R
\name{init_oauth2.0}
\alias{init_oauth2.0}
\alias{oauth2.0_authorize_url}
\alias{oauth2.0_access_token}
\title{Retrieve OAuth 2.0 access token.}
\usage{
init_oauth2.0(
  endpoint,
  app,
  scope = NULL,
  user_params = NULL,
  type = NULL,
  use_oob = getOption("httr_oob_default"),
  oob_value = NULL,
  is_interactive = interactive(),
  use_basic_auth = FALSE,
  config_init = list(),
  client_credentials = FALSE,
  query_authorize_extra = list()
)

oauth2.0_authorize_url(
  endpoint,
  app,
  scope,
  redirect_uri = app$redirect_uri,
  state = nonce(),
  query_extra = list()
)

oauth2.0_access_token(
  endpoint,
  app,
  code,
  user_params = NULL,
  type = NULL,
  use_basic_auth = FALSE,
  redirect_uri = app$redirect_uri,
  client_credentials = FALSE,
  config = list()
)
}
\arguments{
\item{endpoint}{An OAuth endpoint, created by \code{\link[=oauth_endpoint]{oauth_endpoint()}}}

\item{app}{An OAuth consumer application, created by
\code{\link[=oauth_app]{oauth_app()}}}

\item{scope}{a character vector of scopes to request.}

\item{user_params}{Named list holding endpoint specific parameters to pass to
the server when posting the request for obtaining or refreshing the
access token.}

\item{type}{content type used to override incorrect server response}

\item{use_oob}{if FALSE, use a local webserver for the OAuth dance.
Otherwise, provide a URL to the user and prompt for a validation
code. Defaults to the of the \code{"httr_oob_default"} default,
or \code{TRUE} if \code{httpuv} is not installed.}

\item{oob_value}{if provided, specifies the value to use for the redirect_uri
parameter when retrieving an authorization URL. Defaults to "urn:ietf:wg:oauth:2.0:oob".
Requires \code{use_oob = TRUE}.}

\item{is_interactive}{DEPRECATED}

\item{use_basic_auth}{if \code{TRUE} use http basic authentication to
retrieve the token. Some authorization servers require this.
If \code{FALSE}, the default, retrieve the token by including the
app key and secret in the request body.}

\item{config_init}{Additional configuration settings sent to
\code{\link[=POST]{POST()}}, e.g. \code{\link[=user_agent]{user_agent()}}.}

\item{client_credentials}{Default to \code{FALSE}. Set to \code{TRUE} to use
\emph{Client Credentials Grant} instead of \emph{Authorization
Code Grant}. See \url{https://www.rfc-editor.org/rfc/rfc6749#section-4.4}.}

\item{query_authorize_extra}{Default to \code{list()}. Set to named list
holding query parameters to append to initial auth page query. Useful for
some APIs.}

\item{query_extra}{See \code{query_authorize_extra}}
}
\description{
See demos for use.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write-function.R
\name{write_stream}
\alias{write_stream}
\title{Process output in a streaming manner.}
\usage{
write_stream(f)
}
\arguments{
\item{f}{Callback function. It should have a single argument, a raw
vector containing the bytes recieved from the server. This will usually
be 16k or less. The return value of the function is ignored.}
}
\description{
This is the most general way of processing the response from the server -
you receive the raw bytes as they come in, and you can do whatever you want
with them.
}
\examples{
GET(
  "https://github.com/jeroen/data/raw/gh-pages/diamonds.json",
  write_stream(function(x) {
    print(length(x))
    length(x)
  })
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-endpoint.r
\name{oauth_endpoints}
\alias{oauth_endpoints}
\title{Popular oauth endpoints.}
\usage{
oauth_endpoints(name)
}
\arguments{
\item{name}{One of the following endpoints: linkedin, twitter,
vimeo, google, facebook, github, azure.}
}
\description{
Provides some common OAuth endpoints.
}
\examples{
oauth_endpoints("twitter")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-signature.r
\name{sign_oauth}
\alias{sign_oauth}
\alias{sign_oauth1.0}
\alias{sign_oauth2.0}
\title{Sign an OAuth request}
\usage{
sign_oauth1.0(app, token = NULL, token_secret = NULL, as_header = TRUE, ...)

sign_oauth2.0(access_token, as_header = TRUE)
}
\description{
Deprecated. Instead create a config object directly using
\code{config(token = my_token)}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.r
\name{with_config}
\alias{with_config}
\alias{with_verbose}
\title{Execute code with configuration set.}
\usage{
with_config(config = config(), expr, override = FALSE)

with_verbose(expr, ...)
}
\arguments{
\item{config}{Settings as generated by \code{\link[=add_headers]{add_headers()}},
\code{\link[=set_cookies]{set_cookies()}} or \code{\link[=authenticate]{authenticate()}}.}

\item{expr}{code to execute under specified configuration}

\item{override}{if \code{TRUE}, ignore existing settings, if \code{FALSE},
combine new config with old.}

\item{...}{Other arguments passed on to \code{\link[=verbose]{verbose()}}}
}
\description{
Execute code with configuration set.
}
\examples{
with_config(verbose(), {
  GET("http://had.co.nz")
  GET("http://google.com")
})

# Or even easier:
with_verbose(GET("http://google.com"))
}
\seealso{
Other ways to set configuration: 
\code{\link{config}()},
\code{\link{set_config}()}
}
\concept{ways to set configuration}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/http-get.r
\name{GET}
\alias{GET}
\title{GET a url.}
\usage{
GET(url = NULL, config = list(), ..., handle = NULL)
}
\arguments{
\item{url}{the url of the page to retrieve}

\item{config}{Additional configuration settings such as http
authentication (\code{\link[=authenticate]{authenticate()}}), additional headers
(\code{\link[=add_headers]{add_headers()}}), cookies (\code{\link[=set_cookies]{set_cookies()}}) etc.
See \code{\link[=config]{config()}} for full details and list of helpers.}

\item{...}{Further named parameters, such as \code{query}, \code{path}, etc,
passed on to \code{\link[=modify_url]{modify_url()}}. Unnamed parameters will be combined
with \code{\link[=config]{config()}}.}

\item{handle}{The handle to use with this request. If not
supplied, will be retrieved and reused from the \code{\link[=handle_pool]{handle_pool()}}
based on the scheme, hostname and port of the url. By default \pkg{httr}
requests to the same scheme/host/port combo. This substantially reduces
connection time, and ensures that cookies are maintained over multiple
requests to the same host. See \code{\link[=handle_pool]{handle_pool()}} for more
details.}
}
\value{
A \code{\link[=response]{response()}} object.
}
\description{
GET a url.
}
\section{RFC2616}{

The GET method means retrieve whatever information (in the form of an
entity) is identified by the Request-URI. If the Request-URI refers to a
data-producing process, it is the produced data which shall be returned as
the entity in the response and not the source text of the process, unless
that text happens to be the output of the process.

The semantics of the GET method change to a "conditional GET" if the
request message includes an If-Modified-Since, If-Unmodified-Since,
If-Match, If-None-Match, or If-Range header field. A conditional GET method
requests that the entity be transferred only under the circumstances
described by the conditional header field(s). The conditional GET method is
intended to reduce unnecessary network usage by allowing cached entities to
be refreshed without requiring multiple requests or transferring data
already held by the client.

The semantics of the GET method change to a "partial GET" if the request
message includes a Range header field. A partial GET requests that only
part of the entity be transferred, as described in \url{https://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.35}
The partial GET method is intended to reduce unnecessary network usage by
allowing partially-retrieved entities to be completed without transferring
data already held by the client.
}

\examples{
GET("http://google.com/")
\dontrun{
GET("http://google.com/", path = "search")
GET("http://google.com/", path = "search", query = list(q = "ham"))
}

# See what GET is doing with httpbin.org
\dontrun{
url <- "http://httpbin.org/get"
GET(url)
GET(url, add_headers(a = 1, b = 2))
GET(url, set_cookies(a = 1, b = 2))
GET(url, add_headers(a = 1, b = 2), set_cookies(a = 1, b = 2))
GET(url, authenticate("username", "password"))
GET(url, verbose())
}

# You might want to manually specify the handle so you can have multiple
# independent logins to the same website.
\dontrun{
google <- handle("http://google.com")
GET(handle = google, path = "/")
GET(handle = google, path = "search")
}
}
\seealso{
Other http methods: 
\code{\link{BROWSE}()},
\code{\link{DELETE}()},
\code{\link{HEAD}()},
\code{\link{PATCH}()},
\code{\link{POST}()},
\code{\link{PUT}()},
\code{\link{VERB}()}
}
\concept{http methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response-status.r
\name{http_status}
\alias{http_status}
\title{Give information on the status of a request.}
\usage{
http_status(x)
}
\arguments{
\item{x}{a request object or a number.}
}
\value{
If the status code does not match a known status, an error.
Otherwise, a list with components
\item{category}{the broad category of the status}
\item{message}{the meaning of the status code}
}
\description{
Extract the http status code and convert it into a human readable message.
}
\details{
http servers send a status code with the response to each request. This code
gives information regarding the outcome of the execution of the request
on the server. Roughly speaking, codes in the 100s and 200s mean the request
was successfully executed; codes in the 300s mean the page was redirected;
codes in the 400s mean there was a mistake in the way the client sent the
request; codes in the 500s mean the server failed to fulfill
an apparently valid request. More details on the codes can be found at
\verb{http://en.wikipedia.org/wiki/Http_error_codes}.
}
\examples{
http_status(100)
http_status(404)

\dontrun{
x <- GET("http://httpbin.org/status/200")
http_status(x)

http_status(GET("http://httpbin.org/status/300"))
http_status(GET("http://httpbin.org/status/301"))
http_status(GET("http://httpbin.org/status/404"))

# errors out on unknown status
http_status(GET("http://httpbin.org/status/320"))
}
}
\seealso{
Other response methods: 
\code{\link{content}()},
\code{\link{http_error}()},
\code{\link{response}()},
\code{\link{stop_for_status}()}
}
\concept{response methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.r
\name{set_config}
\alias{set_config}
\alias{reset_config}
\title{Set (and reset) global httr configuration.}
\usage{
set_config(config, override = FALSE)

reset_config()
}
\arguments{
\item{config}{Settings as generated by \code{\link[=add_headers]{add_headers()}},
\code{\link[=set_cookies]{set_cookies()}} or \code{\link[=authenticate]{authenticate()}}.}

\item{override}{if \code{TRUE}, ignore existing settings, if \code{FALSE},
combine new config with old.}
}
\value{
invisibility, the old global config.
}
\description{
Set (and reset) global httr configuration.
}
\examples{
GET("http://google.com")
set_config(verbose())
GET("http://google.com")
reset_config()
GET("http://google.com")
}
\seealso{
Other ways to set configuration: 
\code{\link{config}()},
\code{\link{with_config}()}
}
\concept{ways to set configuration}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sha1.r
\name{hmac_sha1}
\alias{hmac_sha1}
\title{HMAC SHA1}
\usage{
hmac_sha1(key, string)
}
\arguments{
\item{key}{secret key}

\item{string}{data to securely hash}
}
\description{
As described in \url{http://datatracker.ietf.org/doc/rfc2104/}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/httr.r
\docType{package}
\name{httr-package}
\alias{httr}
\alias{httr-package}
\title{\pkg{httr} makes http easy.}
\description{
\code{httr} is organised around the six most common http verbs:
\code{\link[=GET]{GET()}}, \code{\link[=PATCH]{PATCH()}},
\code{\link[=POST]{POST()}}, \code{\link[=HEAD]{HEAD()}},
\code{\link[=PUT]{PUT()}}, and \code{\link[=DELETE]{DELETE()}}.
}
\details{
Each request returns a \code{\link[=response]{response()}} object which provides
easy access to status code, cookies, headers, timings, and other useful
info.  The content of the request is available as a raw vector
(\code{\link[=content]{content()}}), character vector (\code{\link[=text_content]{text_content()}}),
or parsed into an R object (\code{\link[=parsed_content]{parsed_content()}}), currently for
html, xml, json, png and jpeg).

Requests can be modified by various config options like
\code{\link[=set_cookies]{set_cookies()}}, \code{\link[=add_headers]{add_headers()}},
\code{\link[=authenticate]{authenticate()}}, \code{\link[=use_proxy]{use_proxy()}},
\code{\link[=verbose]{verbose()}}, and \code{\link[=timeout]{timeout()}}

httr supports OAuth 1.0 and 2.0. Use \code{\link[=oauth1.0_token]{oauth1.0_token()}} and
\code{\link[=oauth2.0_token]{oauth2.0_token()}} to get user tokens, and
\code{\link[=sign_oauth1.0]{sign_oauth1.0()}} and \code{\link[=sign_oauth2.0]{sign_oauth2.0()}} to sign
requests. The demos directory has twelve demos of using OAuth: four for 1.0
(linkedin, twitter, vimeo, and yahoo) and eight for 2.0 (azure, facebook,
github, google, linkedin, reddit, yahoo, and yelp).
}
\seealso{
Useful links:
\itemize{
  \item \url{https://httr.r-lib.org/}
  \item \url{https://github.com/r-lib/httr}
  \item Report bugs at \url{https://github.com/r-lib/httr/issues}
}

}
\author{
\strong{Maintainer}: Hadley Wickham \email{hadley@posit.co}

Other contributors:
\itemize{
  \item Posit, PBC [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-signature.r
\name{oauth_signature}
\alias{oauth_signature}
\alias{oauth_header}
\title{Generate oauth signature.}
\usage{
oauth_signature(
  url,
  method = "GET",
  app,
  token = NULL,
  token_secret = NULL,
  private_key = NULL,
  other_params = NULL
)

oauth_header(info)
}
\arguments{
\item{url, method}{Url and http method of request.}

\item{app}{\code{\link[=oauth_app]{oauth_app()}} object representing application.}

\item{token, token_secret}{OAuth token and secret.}

\item{other_params}{Named argument providing additional parameters
(e.g. \code{oauth_callback} or \code{oauth_body_hash}).}
}
\value{
A list of oauth parameters.
}
\description{
For advanced use only. Occassionally needed for sites that use some
components of the OAuth spec, but not all of them (e.g. 2-legged oauth)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-init.R
\name{init_oauth1.0}
\alias{init_oauth1.0}
\title{Retrieve OAuth 1.0 access token.}
\usage{
init_oauth1.0(
  endpoint,
  app,
  permission = NULL,
  is_interactive = interactive(),
  private_key = NULL
)
}
\arguments{
\item{endpoint}{An OAuth endpoint, created by \code{\link[=oauth_endpoint]{oauth_endpoint()}}}

\item{app}{An OAuth consumer application, created by
\code{\link[=oauth_app]{oauth_app()}}}

\item{permission}{optional, a string of permissions to ask for.}

\item{is_interactive}{DEPRECATED}

\item{private_key}{Optional, a key provided by \code{\link[openssl:read_key]{openssl::read_key()}}.
Used for signed OAuth 1.0.}
}
\description{
See demos for use.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-app.r
\name{oauth_app}
\alias{oauth_app}
\title{Create an OAuth application.}
\usage{
oauth_app(appname, key, secret = NULL, redirect_uri = oauth_callback())
}
\arguments{
\item{appname}{name of the application.  This is not used for OAuth, but is
used to make it easier to identify different applications.}

\item{key}{consumer key, also sometimes called the client ID}

\item{secret}{consumer secret, also sometimes called the client secret.
Despite its name, this does not necessarily need to be protected like a
password, i.e. the user still has to authenticate themselves and grant the
app permission to access resources on their behalf. For example, see
Google's docs for
\href{https://developers.google.com/identity/protocols/OAuth2#installed}{OAuth2 for installed applications}.}

\item{redirect_uri}{The URL that user will be redirected to after
authorisation is complete. You should generally leave this as the default
unless you're using a non-standard auth flow (like with shiny).}
}
\description{
See the demos for instructions on how to create an OAuth app for linkedin,
twitter, vimeo, facebook, github and google. When wrapping an API from a
package, the author may want to include a default app to facilitate early and
casual use and then provide a method for heavy or advanced users to supply
their own app or key and secret.
}
\examples{
\dontrun{
google_app <- oauth_app(
  "google",
  key = "123456789.apps.googleusercontent.com",
  secret = "abcdefghijklmnopqrstuvwxyz"
)
}
}
\seealso{
Other OAuth: 
\code{\link{oauth1.0_token}()},
\code{\link{oauth2.0_token}()},
\code{\link{oauth_endpoint}()},
\code{\link{oauth_service_token}()}
}
\concept{OAuth}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/http-put.r
\name{PUT}
\alias{PUT}
\title{Send PUT request to server.}
\usage{
PUT(
  url = NULL,
  config = list(),
  ...,
  body = NULL,
  encode = c("multipart", "form", "json", "raw"),
  handle = NULL
)
}
\arguments{
\item{url}{the url of the page to retrieve}

\item{config}{Additional configuration settings such as http
authentication (\code{\link[=authenticate]{authenticate()}}), additional headers
(\code{\link[=add_headers]{add_headers()}}), cookies (\code{\link[=set_cookies]{set_cookies()}}) etc.
See \code{\link[=config]{config()}} for full details and list of helpers.}

\item{...}{Further named parameters, such as \code{query}, \code{path}, etc,
passed on to \code{\link[=modify_url]{modify_url()}}. Unnamed parameters will be combined
with \code{\link[=config]{config()}}.}

\item{body}{One of the following:
\itemize{
\item \code{FALSE}: No body. This is typically not used with \code{POST},
\code{PUT}, or \code{PATCH}, but can be useful if you need to send a
bodyless request (like \code{GET}) with \code{VERB()}.
\item \code{NULL}: An empty body
\item \code{""}: A length 0 body
\item \code{upload_file("path/")}: The contents of a file.  The mime
type will be guessed from the extension, or can be supplied explicitly
as the second argument to \code{upload_file()}
\item A character or raw vector: sent as is in body. Use
\code{\link[=content_type]{content_type()}} to tell the server what sort of data
you are sending.
\item A named list: See details for encode.
}}

\item{encode}{If the body is a named list, how should it be encoded? Can be
one of form (application/x-www-form-urlencoded), multipart,
(multipart/form-data), or json (application/json).

For "multipart", list elements can be strings or objects created by
\code{\link[=upload_file]{upload_file()}}. For "form", elements are coerced to strings
and escaped, use \code{I()} to prevent double-escaping. For "json",
parameters are automatically "unboxed" (i.e. length 1 vectors are
converted to scalars). To preserve a length 1 vector as a vector,
wrap in \code{I()}. For "raw", either a character or raw vector. You'll
need to make sure to set the \code{\link[=content_type]{content_type()}} yourself.}

\item{handle}{The handle to use with this request. If not
supplied, will be retrieved and reused from the \code{\link[=handle_pool]{handle_pool()}}
based on the scheme, hostname and port of the url. By default \pkg{httr}
requests to the same scheme/host/port combo. This substantially reduces
connection time, and ensures that cookies are maintained over multiple
requests to the same host. See \code{\link[=handle_pool]{handle_pool()}} for more
details.}
}
\description{
Send PUT request to server.
}
\examples{
POST("http://httpbin.org/put")
PUT("http://httpbin.org/put")

b2 <- "http://httpbin.org/put"
PUT(b2, body = "A simple text string")
PUT(b2, body = list(x = "A simple text string"))
PUT(b2, body = list(y = upload_file(system.file("CITATION"))))
PUT(b2, body = list(x = "A simple text string"), encode = "json")
}
\seealso{
Other http methods: 
\code{\link{BROWSE}()},
\code{\link{DELETE}()},
\code{\link{GET}()},
\code{\link{HEAD}()},
\code{\link{PATCH}()},
\code{\link{POST}()},
\code{\link{VERB}()}
}
\concept{http methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress.R
\name{progress}
\alias{progress}
\title{Add a progress bar.}
\usage{
progress(type = c("down", "up"), con = stdout())
}
\arguments{
\item{type}{Type of progress to display: either number of bytes uploaded
or downloaded.}

\item{con}{Connection to send output too. Usually \code{stdout()} or
\code{stderr}.}
}
\description{
Add a progress bar.
}
\examples{
cap_speed <- config(max_recv_speed_large = 10000)
\donttest{
# If file size is known, you get a progress bar:
x <- GET("http://httpbin.org/bytes/102400", progress(), cap_speed)
# Otherwise you get the number of bytes downloaded:
x <- GET("http://httpbin.org/stream-bytes/102400", progress(), cap_speed)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/upload-file.r
\name{upload_file}
\alias{upload_file}
\title{Upload a file with \code{\link[=POST]{POST()}} or \code{\link[=PUT]{PUT()}}.}
\usage{
upload_file(path, type = NULL)
}
\arguments{
\item{path}{path to file}

\item{type}{mime type of path. If not supplied, will be guess by
\code{\link[mime:guess_type]{mime::guess_type()}} when needed.}
}
\description{
Upload a file with \code{\link[=POST]{POST()}} or \code{\link[=PUT]{PUT()}}.
}
\examples{
citation <- upload_file(system.file("CITATION"))
POST("http://httpbin.org/post", body = citation)
POST("http://httpbin.org/post", body = list(y = citation))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content.r
\name{has_content}
\alias{has_content}
\title{Does the request have content associated with it?}
\usage{
has_content(x)
}
\description{
Does the request have content associated with it?
}
\examples{
has_content(POST("http://httpbin.org/post", body = FALSE))
has_content(HEAD("http://httpbin.org/headers"))
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timeout.r
\name{timeout}
\alias{timeout}
\title{Set maximum request time.}
\usage{
timeout(seconds)
}
\arguments{
\item{seconds}{number of seconds to wait for a response until giving up.
Can not be less than 1 ms.}
}
\description{
Set maximum request time.
}
\details{
This timeout is passed on to \code{\link[curl:handle]{curl::handle_setopt()}}.
See there and \code{\link[curl:curl_options]{curl::curl_options()}} for more details.
}
\examples{
\dontrun{
GET("http://httpbin.org/delay/3", timeout(1))
GET("http://httpbin.org/delay/1", timeout(2))
}
}
\seealso{
Other config: 
\code{\link{add_headers}()},
\code{\link{authenticate}()},
\code{\link{config}()},
\code{\link{set_cookies}()},
\code{\link{use_proxy}()},
\code{\link{user_agent}()},
\code{\link{verbose}()}
}
\concept{config}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/url.r
\name{parse_url}
\alias{parse_url}
\alias{build_url}
\title{Parse and build urls according to RFC3986.}
\usage{
parse_url(url)

build_url(url)
}
\arguments{
\item{url}{For \code{parse_url} a character vector (of length 1) to parse
into components; for \code{build_url} a list of components to turn back
into a string.}
}
\value{
a list containing:
\itemize{
\item scheme
\item hostname
\item port
\item path
\item params
\item fragment
\item query, a list
\item username
\item password
}
}
\description{
See \url{https://www.rfc-editor.org/rfc/rfc3986} for details of parsing
algorithm.
}
\examples{
parse_url("http://google.com/")
parse_url("http://google.com:80/")
parse_url("http://google.com:80/?a=1&b=2")

url <- parse_url("http://google.com/")
url$scheme <- "https"
url$query <- list(q = "hello")
build_url(url)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/http-delete.r
\name{DELETE}
\alias{DELETE}
\title{Send a DELETE request.}
\usage{
DELETE(
  url = NULL,
  config = list(),
  ...,
  body = NULL,
  encode = c("multipart", "form", "json", "raw"),
  handle = NULL
)
}
\arguments{
\item{url}{the url of the page to retrieve}

\item{config}{Additional configuration settings such as http
authentication (\code{\link[=authenticate]{authenticate()}}), additional headers
(\code{\link[=add_headers]{add_headers()}}), cookies (\code{\link[=set_cookies]{set_cookies()}}) etc.
See \code{\link[=config]{config()}} for full details and list of helpers.}

\item{...}{Further named parameters, such as \code{query}, \code{path}, etc,
passed on to \code{\link[=modify_url]{modify_url()}}. Unnamed parameters will be combined
with \code{\link[=config]{config()}}.}

\item{body}{One of the following:
\itemize{
\item \code{FALSE}: No body. This is typically not used with \code{POST},
\code{PUT}, or \code{PATCH}, but can be useful if you need to send a
bodyless request (like \code{GET}) with \code{VERB()}.
\item \code{NULL}: An empty body
\item \code{""}: A length 0 body
\item \code{upload_file("path/")}: The contents of a file.  The mime
type will be guessed from the extension, or can be supplied explicitly
as the second argument to \code{upload_file()}
\item A character or raw vector: sent as is in body. Use
\code{\link[=content_type]{content_type()}} to tell the server what sort of data
you are sending.
\item A named list: See details for encode.
}}

\item{encode}{If the body is a named list, how should it be encoded? Can be
one of form (application/x-www-form-urlencoded), multipart,
(multipart/form-data), or json (application/json).

For "multipart", list elements can be strings or objects created by
\code{\link[=upload_file]{upload_file()}}. For "form", elements are coerced to strings
and escaped, use \code{I()} to prevent double-escaping. For "json",
parameters are automatically "unboxed" (i.e. length 1 vectors are
converted to scalars). To preserve a length 1 vector as a vector,
wrap in \code{I()}. For "raw", either a character or raw vector. You'll
need to make sure to set the \code{\link[=content_type]{content_type()}} yourself.}

\item{handle}{The handle to use with this request. If not
supplied, will be retrieved and reused from the \code{\link[=handle_pool]{handle_pool()}}
based on the scheme, hostname and port of the url. By default \pkg{httr}
requests to the same scheme/host/port combo. This substantially reduces
connection time, and ensures that cookies are maintained over multiple
requests to the same host. See \code{\link[=handle_pool]{handle_pool()}} for more
details.}
}
\value{
A \code{\link[=response]{response()}} object.
}
\description{
Send a DELETE request.
}
\section{RFC2616}{


The DELETE method requests that the origin server delete the resource
identified by the Request-URI. This method MAY be overridden by human
intervention (or other means) on the origin server. The client cannot be
guaranteed that the operation has been carried out, even if the status code
returned from the origin server indicates that the action has been
completed successfully. However, the server SHOULD NOT indicate success
unless, at the time the response is given, it intends to delete the
resource or move it to an inaccessible location.

A successful response SHOULD be 200 (OK) if the response includes an entity
describing the status, 202 (Accepted) if the action has not yet been
enacted, or 204 (No Content) if the action has been enacted but the
response does not include an entity.

If the request passes through a cache and the Request-URI identifies one or
more currently cached entities, those entries SHOULD be treated as stale.
Responses to this method are not cacheable.
}

\examples{
\dontrun{
DELETE("http://httpbin.org/delete")
POST("http://httpbin.org/delete")
}
}
\seealso{
Other http methods: 
\code{\link{BROWSE}()},
\code{\link{GET}()},
\code{\link{HEAD}()},
\code{\link{PATCH}()},
\code{\link{POST}()},
\code{\link{PUT}()},
\code{\link{VERB}()}
}
\concept{http methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/http-head.r
\name{HEAD}
\alias{HEAD}
\title{Get url HEADers.}
\usage{
HEAD(url = NULL, config = list(), ..., handle = NULL)
}
\arguments{
\item{url}{the url of the page to retrieve}

\item{config}{Additional configuration settings such as http
authentication (\code{\link[=authenticate]{authenticate()}}), additional headers
(\code{\link[=add_headers]{add_headers()}}), cookies (\code{\link[=set_cookies]{set_cookies()}}) etc.
See \code{\link[=config]{config()}} for full details and list of helpers.}

\item{...}{Further named parameters, such as \code{query}, \code{path}, etc,
passed on to \code{\link[=modify_url]{modify_url()}}. Unnamed parameters will be combined
with \code{\link[=config]{config()}}.}

\item{handle}{The handle to use with this request. If not
supplied, will be retrieved and reused from the \code{\link[=handle_pool]{handle_pool()}}
based on the scheme, hostname and port of the url. By default \pkg{httr}
requests to the same scheme/host/port combo. This substantially reduces
connection time, and ensures that cookies are maintained over multiple
requests to the same host. See \code{\link[=handle_pool]{handle_pool()}} for more
details.}
}
\value{
A \code{\link[=response]{response()}} object.
}
\description{
Get url HEADers.
}
\section{RFC2616}{

The HEAD method is identical to GET except that the server MUST NOT return
a message-body in the response. The metainformation contained in the HTTP
headers in response to a HEAD request SHOULD be identical to the
information sent in response to a GET request. This method can be used for
obtaining metainformation about the entity implied by the request without
transferring the entity-body itself. This method is often used for testing
hypertext links for validity, accessibility, and recent modification.

The response to a HEAD request MAY be cacheable in the sense that the
information contained in the response MAY be used to update a previously
cached entity from that resource. If the new field values indicate that the
cached entity differs from the current entity (as would be indicated by a
change in Content-Length, Content-MD5, ETag or Last-Modified), then the
cache MUST treat the cache entry as stale.
}

\examples{
HEAD("http://google.com")
headers(HEAD("http://google.com"))
}
\seealso{
Other http methods: 
\code{\link{BROWSE}()},
\code{\link{DELETE}()},
\code{\link{GET}()},
\code{\link{PATCH}()},
\code{\link{POST}()},
\code{\link{PUT}()},
\code{\link{VERB}()}
}
\concept{http methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/user-agent.r
\name{user_agent}
\alias{user_agent}
\title{Set user agent.}
\usage{
user_agent(agent)
}
\arguments{
\item{agent}{string giving user agent}
}
\description{
Override the default RCurl user agent of \code{NULL}
}
\examples{
\dontrun{
GET("http://httpbin.org/user-agent")
GET("http://httpbin.org/user-agent", user_agent("httr"))
}
}
\seealso{
Other config: 
\code{\link{add_headers}()},
\code{\link{authenticate}()},
\code{\link{config}()},
\code{\link{set_cookies}()},
\code{\link{timeout}()},
\code{\link{use_proxy}()},
\code{\link{verbose}()}
}
\concept{config}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date.R
\name{parse_http_date}
\alias{parse_http_date}
\alias{http_date}
\title{Parse and print http dates.}
\usage{
parse_http_date(x, failure = structure(NA_real_, class = "Date"))

http_date(x)
}
\arguments{
\item{x}{For \code{parse_http_date}, a character vector of strings to parse.
All elements must be of the same type.

For \code{http_date}, a \code{POSIXt} vector.}

\item{failure}{What to return on failure?}
}
\value{
A POSIXct object if succesful, otherwise \code{failure}
}
\description{
As defined in RFC2616,
\url{http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.3}, there are
three valid formats:
\itemize{
\item Sun, 06 Nov 1994 08:49:37 GMT  ; RFC 822, updated by RFC 1123
\item Sunday, 06-Nov-94 08:49:37 GMT ; RFC 850, obsoleted by RFC 1036
\item Sun Nov  6 08:49:37 1994       ; ANSI C's asctime() format
}
}
\examples{
parse_http_date("Sun, 06 Nov 1994 08:49:37 GMT")
parse_http_date("Sunday, 06-Nov-94 08:49:37 GMT")
parse_http_date("Sun Nov  6 08:49:37 1994")

http_date(Sys.time())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/headers.r
\name{headers}
\alias{headers}
\title{Extract the headers from a response}
\usage{
headers(x)
}
\arguments{
\item{x}{A request object}
}
\description{
Extract the headers from a response
}
\examples{
\dontrun{
r <- GET("http://httpbin.org/get")
headers(r)
}
}
\seealso{
\code{\link[=add_headers]{add_headers()}} to send additional headers in a
request
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-token-utils.R
\name{revoke_all}
\alias{revoke_all}
\title{Revoke all OAuth tokens in the cache.}
\usage{
revoke_all(cache_path = NA)
}
\arguments{
\item{cache_path}{Path to cache file. Defaults to \code{.httr-oauth} in
current directory.}
}
\description{
Use this function if you think that your token may have been compromised,
e.g. you accidentally uploaded the cache file to github. It's not possible
to automatically revoke all tokens - this function will warn when it can't.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-listener.r
\name{oauth_listener}
\alias{oauth_listener}
\title{Create a webserver to listen for OAuth callback.}
\usage{
oauth_listener(request_url, is_interactive = interactive())
}
\arguments{
\item{request_url}{the url to send the browser to}

\item{is_interactive}{DEPRECATED}
}
\description{
This opens a web browser pointing to \code{request_url}, and opens a
webserver on port 1410 to listen to the reponse.  The redirect url should
either be set previously (during the OAuth authentication dance) or
supplied as a parameter to the url.  See \code{\link[=oauth1.0_token]{oauth1.0_token()}} and
\code{\link[=oauth2.0_token]{oauth2.0_token()}} for examples of both techniques.
}
\details{
This function should not normally be called directly by the user.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/headers.r
\name{content_type}
\alias{content_type}
\alias{content_type_json}
\alias{content_type_xml}
\alias{accept}
\alias{accept_json}
\alias{accept_xml}
\title{Set content-type and accept headers.}
\usage{
content_type(type)

content_type_json()

content_type_xml()

accept(type)

accept_json()

accept_xml()
}
\arguments{
\item{type}{A mime type or a file extension. If a file extension (i.e. starts
with \code{.}) will guess the mime type using \code{\link[mime:guess_type]{mime::guess_type()}}.}
}
\description{
These are convenient wrappers aroud \code{\link[=add_headers]{add_headers()}}.
}
\details{
\code{accept_json}/\code{accept_xml} and
\code{content_type_json}/\code{content_type_xml} are useful shortcuts to
ask for json or xml responses or tell the server you are sending json/xml.
}
\examples{
\dontrun{
GET("http://httpbin.org/headers")

GET("http://httpbin.org/headers", accept_json())
GET("http://httpbin.org/headers", accept("text/csv"))
GET("http://httpbin.org/headers", accept(".doc"))

GET("http://httpbin.org/headers", content_type_xml())
GET("http://httpbin.org/headers", content_type("text/csv"))
GET("http://httpbin.org/headers", content_type(".xml"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth-token.r
\name{oauth_service_token}
\alias{oauth_service_token}
\title{Generate OAuth token for service accounts.}
\usage{
oauth_service_token(endpoint, secrets, scope = NULL, sub = NULL)
}
\arguments{
\item{endpoint}{An OAuth endpoint, created by \code{\link[=oauth_endpoint]{oauth_endpoint()}}}

\item{secrets}{Secrets loaded from JSON file, downloaded from console.}

\item{scope}{a character vector of scopes to request.}

\item{sub}{The email address of the user for which the application is
requesting delegated access.}
}
\description{
Service accounts provide a way of using OAuth2 without user intervention.
They instead assume that the server has access to a private key used
to sign requests. The OAuth app is not needed for service accounts:
that information is embedded in the account itself.
}
\examples{
\dontrun{
endpoint <- oauth_endpoints("google")
secrets <- jsonlite::fromJSON("~/Desktop/httrtest-45693cbfac92.json")
scope <- "https://www.googleapis.com/auth/bigquery.readonly"

token <- oauth_service_token(endpoint, secrets, scope)
}
}
\seealso{
Other OAuth: 
\code{\link{oauth1.0_token}()},
\code{\link{oauth2.0_token}()},
\code{\link{oauth_app}()},
\code{\link{oauth_endpoint}()}
}
\concept{OAuth}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write-function.R
\name{write_disk}
\alias{write_disk}
\alias{write_memory}
\title{Control where the response body is written.}
\usage{
write_disk(path, overwrite = FALSE)

write_memory()
}
\arguments{
\item{path}{Path to content to.}

\item{overwrite}{Will only overwrite existing \code{path} if TRUE.}
}
\description{
The default behaviour is to use \code{write_memory()}, which caches
the response locally in memory. This is useful when talking to APIs as
it avoids a round-trip to disk. If you want to save a file that's bigger
than memory, use \code{write_disk()} to save it to a known path.
}
\examples{
tmp <- tempfile()
r1 <- GET("https://www.google.com", write_disk(tmp))
readLines(tmp)

# The default
r2 <- GET("https://www.google.com", write_memory())

# Save a very large file
\dontrun{
GET(
  "http://www2.census.gov/acs2011_5yr/pums/csv_pus.zip",
  write_disk("csv_pus.zip"), progress()
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response-status.r
\name{http_error}
\alias{http_error}
\alias{url_success}
\alias{url_ok}
\title{Check for an http error.}
\usage{
http_error(x, ...)
}
\arguments{
\item{x}{Object to check. Default methods are provided for strings
(which perform an \code{\link[=HEAD]{HEAD()}} request), responses, and
integer status codes.}

\item{...}{Other arguments passed on to methods.}
}
\value{
\code{TRUE} if the request fails (status code 400 or above),
otherwise \code{FALSE}.
}
\description{
Check for an http error.
}
\examples{
# You can pass a url:
http_error("http://www.google.com")
http_error("http://httpbin.org/status/404")

# Or a request
r <- GET("http://httpbin.org/status/201")
http_error(r)

# Or an (integer) status code
http_error(200L)
http_error(404L)
}
\seealso{
Other response methods: 
\code{\link{content}()},
\code{\link{http_status}()},
\code{\link{response}()},
\code{\link{stop_for_status}()}
}
\concept{response methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handle.r
\name{handle}
\alias{handle}
\title{Create a handle tied to a particular host.}
\usage{
handle(url, cookies = TRUE)
}
\arguments{
\item{url}{full url to site}

\item{cookies}{DEPRECATED}
}
\description{
This handle preserves settings and cookies across multiple requests. It is
the foundation of all requests performed through the httr package, although
it will mostly be hidden from the user.
}
\note{
Because of the way argument dispatch works in R, using handle() in the
http methods (See \code{\link[=GET]{GET()}}) will cause problems when trying to
pass configuration arguments (See examples below). Directly specifying the
handle when using http methods is not recommended in general, since the
selection of the correct handle is taken care of when the user passes an url
(See \code{\link[=handle_pool]{handle_pool()}}).
}
\examples{
handle("http://google.com")
handle("https://google.com")

h <- handle("http://google.com")
GET(handle = h)
# Should see cookies sent back to server
GET(handle = h, config = verbose())

h <- handle("http://google.com", cookies = FALSE)
GET(handle = h)$cookies
\dontrun{
# Using the preferred way of configuring the http methods
# will not work when using handle():
GET(handle = h, timeout(10))
# Passing named arguments will work properly:
GET(handle = h, config = list(timeout(10), add_headers(Accept = "")))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/url.r
\name{modify_url}
\alias{modify_url}
\title{Modify a url.}
\usage{
modify_url(
  url,
  scheme = NULL,
  hostname = NULL,
  port = NULL,
  path = NULL,
  query = NULL,
  params = NULL,
  fragment = NULL,
  username = NULL,
  password = NULL
)
}
\arguments{
\item{url}{the url to modify}

\item{scheme, hostname, port, path, query, params, fragment, username, password}{components of the url to change}
}
\description{
Modify a url by first parsing it and then replacing components with
the non-NULL arguments of this function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/handle-pool.r
\docType{data}
\name{handle_pool}
\alias{handle_pool}
\alias{handle_find}
\alias{handle_reset}
\title{Maintain a pool of handles.}
\format{
An environment.
}
\usage{
handle_pool

handle_find(url)

handle_reset(url)
}
\description{
The handle pool is used to automatically reuse Curl handles for the same
scheme/host/port combination. This ensures that the http session is
automatically reused, and cookies are maintained across requests to a site
without user intervention.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cookies.r
\name{cookies}
\alias{cookies}
\title{Access cookies in a response.}
\usage{
cookies(x)
}
\arguments{
\item{x}{A response.}
}
\description{
Access cookies in a response.
}
\examples{
r <- GET("http://httpbin.org/cookies/set", query = list(a = 1, b = 2))
cookies(r)
}
\seealso{
\code{\link[=set_cookies]{set_cookies()}} to send cookies in request.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/media-guess.r
\name{guess_media}
\alias{guess_media}
\title{Guess the media type of a path from its extension.}
\usage{
guess_media(x)
}
\arguments{
\item{x}{path to file}
}
\description{
DEPRECATED: please use \code{mime::guess_type} instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/media-parse.r
\name{parse_media}
\alias{parse_media}
\title{Parse a media type.}
\usage{
parse_media(x)
}
\arguments{
\item{x}{String to parse}
}
\description{
Parsed according to RFC 2616, as at
\url{https://www.rfc-editor.org/rfc/rfc2616#section-3.7}.
}
\details{
A simplified minimal EBNF is:
\itemize{
\item media-type     = \verb{type "/" subtype *( ";" parameter )}
\item type           = \code{token}
\item subtype        = \code{token}
\item parameter      = \verb{attribute "=" value}
\item attribute      = \code{token}
\item value          = \code{token | quoted-string}
\item token          = \verb{1*<any CHAR except CTLs or ()<>@,;:\\"/[]?=\\\{\\\}}
}
}
\examples{
parse_media("text/plain")
parse_media("text/plain; charset=utf-8")
parse_media("text/plain; charset=\"utf-8\"")
parse_media("text/plain; randomparam=\";=;=\"")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cookies.r
\name{set_cookies}
\alias{set_cookies}
\title{Set cookies.}
\usage{
set_cookies(..., .cookies = character(0))
}
\arguments{
\item{...}{a named cookie values}

\item{.cookies}{a named character vector}
}
\description{
Set cookies.
}
\examples{
set_cookies(a = 1, b = 2)
set_cookies(.cookies = c(a = "1", b = "2"))

GET("http://httpbin.org/cookies")
GET("http://httpbin.org/cookies", set_cookies(a = 1, b = 2))
}
\seealso{
\code{\link[=cookies]{cookies()}} to see cookies in response.

Other config: 
\code{\link{add_headers}()},
\code{\link{authenticate}()},
\code{\link{config}()},
\code{\link{timeout}()},
\code{\link{use_proxy}()},
\code{\link{user_agent}()},
\code{\link{verbose}()}
}
\concept{config}
